package reasoner.grounding;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import java.sql.ResultSet;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;

import reasoner.ReasonerConfig;
import reasoner.io.dataset.Relation;
import reasoner.lineage.Lineage;
import reasoner.lineage.LineageArray;
import reasoner.lineage.LineageNot;
import reasoner.lineage.providers.LineageProvider;
import reasoner.query.compiled.CompiledConstraint;
import reasoner.query.compiled.Query;
import reasoner.query.compiled.QueryStep;
import reasoner.query.compiled.CompiledDeductionRule;

public class Grounder {
	
	private LineageArray lineages;
	private LineageProvider lineageProvider;
	private IntervalLogic intervalLogic;
	private IntervalProvider intervalProvider;
	
	// statistics
	public int numberConstraintClauses;
	
	public Grounder() throws Exception {
		lineages = LineageArray.getInstance();
		lineageProvider = LineageProvider.getInstance();
		intervalLogic = new IntervalLogic();
		intervalProvider = IntervalProvider.getInstance();
	}
	
	/**
	 * top level method for query processing
	 * 
	 * initiates each query level
	 */
	public void queryScheduler(Query query) throws Exception {
		for (int i = 0; i < query.steps.length; i++) {
			if (ReasonerConfig.debugMode) {
				System.out.println("Entering Query Level " + i);
			}
			this.executeStep(query.steps[i]);
		}
		lineages.finalizeGrounding();
		System.out.println("done");
	}
	
	/**
	 * delegates each rule to `executeStep'
	 * creates the logical or of the lineage
	 * inserts the data into database
	 */
  private void executeStep(QueryStep step) throws Exception {
		step.headRelation.createTable();
		
		int numArgs = step.headRelation.numNonTempArgs;
		HashMap<String, LinkedList<Interval>> args2intervals = new HashMap<String, LinkedList<Interval>>();
		for (CompiledDeductionRule rule : step.rules){
			executeRule(rule, args2intervals, numArgs);
		}

		// insert into DB
		for (Entry<String,LinkedList<Interval>> e : args2intervals.entrySet()) {
			
		  LinkedList<Interval> intervals = e.getValue();
		  if (intervals.isEmpty()) {
		  	// no interval could be deduced
		  	continue;
		  }
		  
			// manage lineage
			for (Interval interval : intervals) {
			  if (ReasonerConfig.debugMode && interval.lineageArrayIds.isEmpty()) {
          throw new Exception("no lineage available in interval!");
        }
			  
			  // create lineage
			  int arrayId;
			  if (interval.lineageArrayIds.size() == 1) {
			  	arrayId = interval.lineageArrayIds.toArray()[0];
        }else{
          Lineage[] children = new Lineage[interval.lineageArrayIds.size()];
          TIntIterator it = interval.lineageArrayIds.iterator();
          for (int i=0; i<interval.lineageArrayIds.size(); i++){
            int id = it.next();
            children[i] = lineages.getLineage(id);
          }
          Lineage lineage = lineageProvider.getLineageOr(children);
          arrayId = lineages.addLineage(lineage);
        }
			  
			  // delete old lineage ids and add arrayId
			  interval.lineageArrayIds.clear();
			  interval.lineageArrayIds.add(arrayId);
			}
			
			// get args
			String[] args = e.getKey().split("\t");
      if (args.length != numArgs) {
        throw new Exception("Tab in an argument: " + e.getKey());
      }
			
			step.headRelation.addFactToBatch(args, -1.0, intervals.toArray(new Interval[intervals.size()]));
		}
		step.headRelation.executeBatch();
	}
	
	/**
	 * handles a single rule during grounding
	 * saves the results in @args2intervals : fact -> List<Interval>
	 * the intervals are non-overlapping, and merged
	 * the lineageIds of the interval point to logical ands
	 * logical ands are created here
	 */
	private void executeRule(CompiledDeductionRule rule, HashMap<String, LinkedList<Interval>> args2intervals, int numArgs) throws Exception {
		if (ReasonerConfig.debugMode) {
			System.out.println("SQL: " + rule.stmt);
		}
		
		ResultSet rset = rule.stmt.executeQuery();
		
		LinkedList<Interval> currentIntervals = new LinkedList<Interval>();
		String key = null;
		TIntSet negativeArrayIds = new TIntHashSet();
		int[] positiveArrayIds = new int[rule.positiveArrayIdsInRset.length];
		positiveArrayIds[0] = -1; // ArrayIds >= 0. So this marks uninitialized.
		Interval interval = null;
		while (rset.next()) {
			// manage key
			String newKey = rule.getNonTemporalArguments(rset);
			
			// is the key new?
			if (key == null || !newKey.equals(key)){
				if (!args2intervals.containsKey(newKey)){
					args2intervals.put(newKey, new LinkedList<Interval>());
				}
				if (key != null) {
					if (interval != null) {
						attachLineage(positiveArrayIds, negativeArrayIds, interval);
						currentIntervals.add(interval);
						interval = null;
						positiveArrayIds[0] = -1;
					}
					if (ReasonerConfig.debugMode) {
						checkCurrentIntervals(currentIntervals);
					}
				  args2intervals.put(key, intervalLogic.mergeIntervals(args2intervals.get(key), currentIntervals));
				}
				key = newKey;
		    currentIntervals.clear();
			}
			
			int[] newPositiveArrayIds = new int[rule.positiveArrayIdsInRset.length];
			for (int i = 0; i < rule.positiveArrayIdsInRset.length; i++) {
			  newPositiveArrayIds[i] = rset.getInt(rule.positiveArrayIdsInRset[i]);
			}
			
			if (!Arrays.equals(newPositiveArrayIds, positiveArrayIds)) {
				// another instance of the rule body
				// let's create the lineage
				if (interval != null) {
					attachLineage(positiveArrayIds, negativeArrayIds, interval);
					currentIntervals.add(interval);
					interval = null;
				}
				if (ReasonerConfig.debugMode) {
					checkCurrentIntervals(currentIntervals);
				}
				interval = rule.getDeducedInterval(rset, intervalProvider);
				positiveArrayIds = newPositiveArrayIds;
				negativeArrayIds.clear();
			}
			
			for (int i = 0; i < rule.negativeArrayIdsInRset.length; i++) {
				int negativeArrayId = rset.getInt(rule.negativeArrayIdsInRset[i]);
				if (!rset.wasNull()) {
					negativeArrayIds.add(negativeArrayId);
				}
			}
		}
		if (key != null) {
			if (interval != null) {
				attachLineage(positiveArrayIds, negativeArrayIds, interval);
				currentIntervals.add(interval);
				interval = null;
			}
			if (ReasonerConfig.debugMode) {
				checkCurrentIntervals(currentIntervals);
			}
		  args2intervals.put(key, intervalLogic.mergeIntervals(args2intervals.get(key), currentIntervals));
		}
		
		rset.close();
	}
	
	private void checkCurrentIntervals(LinkedList<Interval> intervals) throws Exception {
		for (Interval interval : intervals) {
			if (interval.begin >= interval.end) {
				throw new Exception("Invalid interval: " + interval);
			}
			if (interval.lineageArrayIds.contains(-1)) {
				throw new Exception("Invalid array id in interval " + interval);
			}
		}
	}
	
	/**
	 * Creates lineage for one grounded rule body and attaches it to @interval.
	 * posLit AND posLit AND ... AND NOT (negLit OR negLit ... OR negLit)
	 */
	private void attachLineage(int[] positiveArrayIds, TIntSet negativeArrayIds, Interval interval) throws Exception {
		if (interval == null) {
			return;
		}
		int arrayId;
		if (negativeArrayIds.isEmpty() && positiveArrayIds.length == 1){
		  // no new lineage data structure necessary
			arrayId = positiveArrayIds[0];
			lineages.getLineage(arrayId);
		} else {
		  // a new logical and is necessary
		  int numChildrenLineage = positiveArrayIds.length;
		  Lineage negLineage = getNegativeLineage(negativeArrayIds);
		  if (negLineage != null) {
		  	numChildrenLineage += 1;
		  }
		  Lineage[] children = new Lineage[numChildrenLineage];
		  
		  // add positive lineage
			for (int i = 0; i < positiveArrayIds.length; i++) {
			  children[i] = lineages.getLineage(positiveArrayIds[i]);
			}
			// add negative lineage
			if (negLineage != null) {
			  children[numChildrenLineage - 1] = negLineage;
			}
			// create lineage
			Lineage and = lineageProvider.getLineageAnd(children);
			arrayId = lineages.addLineage(and);
		}
		interval.lineageArrayIds.clear();
		interval.lineageArrayIds.add(arrayId);
	}
	
	/**
	 * get negative lineage for a single part of positive lineage
	 * executes all negative sub queries
	 * negative lineage is of the form: NOT( OR (neg1, neg2, ... ))
	 * 
	 * returns null, if there is no negative literal to be added
	 */
	private Lineage getNegativeLineage(TIntSet negativeArrayIds) throws Exception {
	  int numNegLineageIds = negativeArrayIds.size();
	  if (numNegLineageIds == 0) {
	  	return null; // no lineage to be added
	  }
	  // create the negative lineage
	  Lineage[] negLineage = new Lineage[numNegLineageIds];
	  TIntIterator it = negativeArrayIds.iterator();
	  Lineage toNegate = null;
	  if (numNegLineageIds == 1) {
	    toNegate = lineages.getLineage(it.next());
	  } else {
	    for (int i = 0; i < numNegLineageIds; i++) {
        negLineage[i] = lineages.getLineage(it.next());
      }
      toNegate = lineageProvider.getLineageOr(negLineage);
	  }
	  LineageNot not = lineageProvider.getLineageNot(toNegate);
	  return not;
	}
	
	/**
	 * returns all facts in the output of the query
	 * without constraints
	 */
	public LinkedList<Tuple> output(Query query) throws Exception {
		return output(query, null);
	}
	
	/**
	 * returns all facts in the output of the query.
	 * constraint lineage is added to each interval.
	 */
	public LinkedList<Tuple> output(Query query, Lineage constraintLineage) throws Exception {
		LinkedList<Tuple> result = new LinkedList<Tuple>();
		for (Relation outputRelation : query.output){
		  result.addAll(outputRelation.getFacts());
		}
		// add constraint lineage to each interval
		if (constraintLineage != null) {
			for (Tuple fact : result) {
				LinkedList<Interval> newIntervals = new LinkedList<Interval>();
				for (Interval interval : fact.intervals) {
					Lineage intervalLineage = lineages.getLineage(interval.lineageArrayIds.toArray()[0]);
					Lineage lineageWithConstraint = lineageProvider.getLineageAnd(intervalLineage, constraintLineage);
					
					int arrayId = lineages.addLineage(lineageWithConstraint);
					newIntervals.add(intervalProvider.get(interval.begin, interval.end, arrayId));
				}
				
				// replace intervals in fact
				intervalProvider.handBack(fact.intervals);
				fact.intervals.clear();
				fact.intervals.addAll(newIntervals);
			}
		}
		return result;
	}
	
	/**
	 * grounds constraints
	 * 
	 * returns null, if no constraints were found
	 * returns a conjunction of grounded constraints, otherwise
	 */
	public Lineage groundConstraints(LinkedList<CompiledConstraint> cons) throws Exception{
	  LinkedList<Lineage> constraints = new LinkedList<Lineage>();
	  
	  for (CompiledConstraint constraint : cons) {
	  	if (ReasonerConfig.debugMode) {
	  		System.out.println(constraint.stmt);
	  	}
	    ResultSet rset = constraint.stmt.executeQuery();
	    while(rset.next()) {
	      List<Lineage> children = new LinkedList<Lineage>();
	      // positive literals
	      for (int i = 0; i < constraint.numPosArgs; i++) {
	      	Lineage lineage = lineages.getLineage(rset.getInt(i + 1)); 
	        children.add(lineage);
	      }
	      // negative literals
	      for (int i = 0; i < constraint.numNegArgs; i++) {
	      	int arrayId = rset.getInt(constraint.numPosArgs + i + 1);
	      	if (!rset.wasNull()) {
	      		Lineage lineage = lineages.getLineage(arrayId);
	      		lineage = lineageProvider.getLineageNot(lineage);
	      		children.add(lineage);
	      	}
	      }
	      Lineage and = lineageProvider.getLineageAnd(children.toArray(new Lineage[children.size()]));
	      if (constraint.negated) {
	      	constraints.add(lineageProvider.getLineageNot(and));
	      } else {
	      	constraints.add(and);
	      }
	    }
	    rset.close();
	  }
	  
	  // load the confidences from database
	  lineages.finalizeGrounding();
	  
	  Lineage[] constraintsArray = new Lineage[constraints.size()];
	  constraintsArray = constraints.toArray(constraintsArray);
	  numberConstraintClauses = constraintsArray.length;
	  if (constraintsArray.length == 0 ) {
	  	return null;
	  } if (constraintsArray.length == 1) {
	  	return constraintsArray[0];
	  } else {
	  	return lineageProvider.getLineageAnd((Lineage[])constraintsArray);
	  }
	}
	
	public void clear() throws Exception {
		numberConstraintClauses = 0;
		intervalLogic.clear();
	}
}
