package reasoner.grounding;

import util.provider.Providable;
import gnu.trove.set.hash.TIntHashSet;;

/**
 * A class capturing a time interval.
 * 
 * Many (non-overlapping) intervals will constitute a fact's time annotation.
 * Intervals are half open: [begin, end)
 * The arrayIds refer to lineage formulas making the fact at this interval true.
 * In other words, all lineage ids in the set are connected by a logical or.
 * 
 * @author mdylla@mpi-inf.mpg.de (Maximilian Dylla)
 */
public class Interval extends Providable implements Comparable<Interval>{
	// [begin,end)
  public int begin;
  public int end;
  // a logical OR between all lineage formulas deriving this interval
  public TIntHashSet lineageArrayIds;
  
  // for output and labels only
  public double confidence;
  
  public Interval(){
  	this.lineageArrayIds = new TIntHashSet();
  	this.confidence = -1.0;
  }
  
  public void clear() {
    begin = 0;
    end = 0;
    lineageArrayIds.clear();
    confidence = -1.0;
  }
  
  /**
   * calculates the overlap with @other in number of time points
   */
  public int getOverlap(Interval other) {
    if (end <= other.begin || other.end <= begin) {
      return 0;  // the intervals are disjoint
    }
    int commonBegin = begin;
    if (other.begin > begin) {
      commonBegin = other.begin;
    }
    int commonEnd = end;
    if (other.end < end) {
      commonEnd = other.end;
    }
    return commonEnd - commonBegin;
  }
  
  public boolean isContainedIn(Interval other) {
  	return other.begin <= this.begin && this.end <= other.end;
  }
  
  /**
   * calculates the length of the interval in number of time points
   */
  public int getLength() throws Exception {
  	int length = end - begin; 
  	if (length < 0) {
  		throw new Exception("Invalid length " + length + " of interval: " + this);
  	}
    return length;
  }
  
  public int compareTo(Interval other){
		if (other == this)
			return 0;
		Interval otherInt = (Interval)other;
		if (otherInt.begin < this.begin)
			return -1;
		else if (otherInt.begin > this.begin)
			return 1;
		else{
			// compare end
			if (otherInt.end < this.end){
				return -1;
			}else if (otherInt.end > this.end){
				return 1;
			}else{
				return -1;
			}
		}
  }
  
  public String toString(){
  	return "["+this.begin+";"+this.end+"@ids:"+this.lineageArrayIds+"@c:"+this.confidence+"]";
  }
}
