package reasoner.grounding;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.list.linked.TIntLinkedList;
import gnu.trove.map.hash.TIntIntHashMap;
import gnu.trove.set.hash.TIntHashSet;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;

import reasoner.ReasonerConfig;
import util.time.Timer;

/**
 * Encodes the interval logic of deduplicating the intervals of facts.
 * Is used by @Grounder.
 */
public class IntervalLogic {
	
	private Timer timer;
	private IntervalProvider intervalProvider;
	
	public IntervalLogic() throws Exception {
		timer = Timer.getInstance();
		timer.reset(IntervalLogic.class.getName());
		intervalProvider = IntervalProvider.getInstance();
	}

	/**
	 * @param oldIntervals
	 * @param newIntervals will be made non-overlapping and then merged with oldIntervals
	 * @return merged intervals
	 */
	public LinkedList<Interval> mergeIntervals(final LinkedList<Interval> oldIntervals, LinkedList<Interval> newIntervals) throws Exception {
		timer.startTimer(IntervalLogic.class.getName());
		
		newIntervals = this.makeNonOverlapping(newIntervals);
		
		LinkedList<Interval> result = new LinkedList<Interval>();
		Iterator<Interval> itOld = oldIntervals.iterator();
		Iterator<Interval> itNew = newIntervals.iterator();
		
		Interval intOld = this.next(itOld);
		Interval intNew = this.next(itNew);
		if (intOld!=null && intNew!=null){
			// both lists are non-empty => merging necessary
			
		  // merge sort like iteration
			while (intOld!=null && intNew!=null){
				if (intOld.end <= intNew.begin){
					// no overlap
					result.add(intOld);
					intOld = this.next(itOld);
					continue;
				}else if (intNew.end <= intOld.begin){
					// no overlap
					result.add(intNew);
					intNew = this.next(itNew);
					continue;
				}else if (intOld.begin < intNew.begin){
					// add the part of the old one alone
					Interval interval = intervalProvider.get(intOld.begin, intNew.begin, intOld.lineageArrayIds);
					intOld.begin = intNew.begin;
					result.add(interval);
					continue;
				}else if (intNew.begin < intOld.begin){
				  // add the part of the new one alone
				  Interval interval = intervalProvider.get(intNew.begin, intOld.begin, intNew.lineageArrayIds);
					intNew.begin = intOld.begin;
					result.add(interval);
					continue;
				}else if (intNew.begin == intOld.begin){
					if (intNew.end == intOld.end){
						// the intervals are identical
						intNew.lineageArrayIds.addAll(intOld.lineageArrayIds);
						result.add(intNew);
						intNew = this.next(itNew);
						intervalProvider.handBack(intOld);
						intOld = this.next(itOld);
						continue;
					}else if (intNew.end < intOld.end){
						intNew.lineageArrayIds.addAll(intOld.lineageArrayIds);
						result.add(intNew);
						intOld.begin = intNew.end;
						intNew = this.next(itNew);
						continue;
					}else if (intOld.end < intNew.end){
						intOld.lineageArrayIds.addAll(intNew.lineageArrayIds);
						result.add(intOld);
						intNew.begin = intOld.end;
						intOld = this.next(itOld);
					}else System.err.println("This case should never happen");
				}else System.err.println("this case should never happen");
			}
		}
		
		// use up the tail
		while (intOld!=null){
			result.add(intOld);
			intOld = this.next(itOld);
		}
		while (intNew!=null){
			result.add(intNew);
			intNew = this.next(itNew);
		}
		
		result = mergeAdjacentIntervals(result);
		
		timer.stopTimer(IntervalLogic.class.getName());
		return result;
	}
	
	/**
	 * Helper function to guard the iterator exception
	 */
	private Interval next(Iterator<Interval> it){
		if (it.hasNext()){
			return it.next();
		}else{
			return null;
		}
	}
	
	/**
	 * Merges two intervals, if they have the same lineage ids and are adjacent.
	 * Assumes that the input is ordered and non-overlapping.
	 * O(|input|)
	 * 
	 * protected for testing
	 */
	protected LinkedList<Interval> mergeAdjacentIntervals(LinkedList<Interval> input) throws Exception {
	  LinkedList<Interval> result = new LinkedList<Interval>();
	  if (input.isEmpty()) {
	    return result;
	  }
	  
	  // linear scan over the list of intervals
	  Iterator<Interval> it = input.iterator();
	  Interval lastInterval = it.next();
	  int lastBegin = lastInterval.begin;
	  while (it.hasNext()) {
	    Interval currentInterval = it.next();
	    
	    if (ReasonerConfig.debugMode) {
	      if (lastInterval.end > currentInterval.begin) {
	        System.err.println("Input intervals were overlapping: "+lastInterval+" "+currentInterval);
	        return null;
	      }
	    }
	    if (currentInterval.begin != lastInterval.end // they are not adjacent
	        || currentInterval.lineageArrayIds.size() != lastInterval.lineageArrayIds.size() // sets of lineage ids are not equal
	        || !currentInterval.lineageArrayIds.containsAll(lastInterval.lineageArrayIds)) { // sets of lineage ids are not equal
	      mergeAdjacentIntervalsHelper(lastInterval, lastBegin, result);
	      lastBegin = currentInterval.begin;
	    }
	    lastInterval = currentInterval;
	  }
	  mergeAdjacentIntervalsHelper(lastInterval, lastBegin, result);
	  return result;
	}
	
	/**
	 * merge two intervals or return the old one, if no changes occurred
	 */
	private void mergeAdjacentIntervalsHelper(
	    Interval lastInterval,
	    Integer lastBegin,
	    LinkedList<Interval> result) throws Exception {
    // we need to create a new interval
    if (lastBegin == lastInterval.begin) {
      // no changes, emit the old interval
      result.add(lastInterval);
    } else {
      Interval newInterval = intervalProvider.get(lastBegin, lastInterval.end, lastInterval.lineageArrayIds);
      result.add(newInterval);
    }
	}
	
	/**
	 * protected for testing
	 */
	protected LinkedList<Interval> makeNonOverlapping(LinkedList<Interval> intervals) throws Exception {
		if (intervals.isEmpty()) {
			return intervals;
		}
		//linear scan through all intervals to fill datastructures
		TIntHashSet limits = new TIntHashSet(); // all different time points
		HashMap<Integer,LinkedList<Interval>> begins = new HashMap<Integer,LinkedList<Interval>>(); // timepoint -> list of intervals beginning there
		HashMap<Integer,LinkedList<Interval>> ends = new HashMap<Integer,LinkedList<Interval>>(); // timepoint -> list of intervals ending there 
		for (Interval interval : intervals){
			limits.add(interval.begin);
			limits.add(interval.end);
			if (begins.get(interval.begin)==null)
				begins.put(interval.begin, new LinkedList<Interval>());
			if (ends.get(interval.end)==null)
				ends.put(interval.end, new LinkedList<Interval>());
			begins.get(interval.begin).add(interval);
			ends.get(interval.end).add(interval);
		}
		
		// create new non-overlapping intervals
		LinkedList<Interval> result = new LinkedList<Interval>();
		TIntIntHashMap arrayId2multiplicity = new TIntIntHashMap();
		TIntLinkedList list = new TIntLinkedList();
		list.addAll(limits);
		list.sort();
		TIntIterator it = list.iterator();
		
		// initialize with first limit
		int lastLimit = it.next();
		for (Interval interval : begins.get(lastLimit)){
			TIntIterator itArrayIds = interval.lineageArrayIds.iterator();
			for (int j=0; j<interval.lineageArrayIds.size(); j++){
				int arrayId = itArrayIds.next();
				boolean containedId = arrayId2multiplicity.increment(arrayId);
				if (!containedId)
					arrayId2multiplicity.put(arrayId, 1);
			}
		}
		TIntHashSet previousArrayIds = new TIntHashSet();
		
		// iterate over remaining limits
		for (int i=1; i<list.size(); i++){
			int limit = it.next();
			
      previousArrayIds.clear();
      previousArrayIds.addAll(arrayId2multiplicity.keys());
			
			// manage active lineage
			if (ends.get(limit)!=null){
				LinkedList<Interval> inters = ends.get(limit);
				for (Interval inter : inters){
					TIntIterator itArrayIds = inter.lineageArrayIds.iterator();
					for (int j=0; j<inter.lineageArrayIds.size(); j++){
						int arrayId = itArrayIds.next();
						arrayId2multiplicity.adjustValue(arrayId, -1);
						if (arrayId2multiplicity.get(arrayId)==0){
							arrayId2multiplicity.remove(arrayId);
						}
					}
				}
			}
			if (begins.get(limit)!=null){
				LinkedList<Interval> inters = begins.get(limit);
				for (Interval inter : inters){
					TIntIterator itArrayIds = inter.lineageArrayIds.iterator();
					for (int j=0; j<inter.lineageArrayIds.size(); j++){
						int arrayId = itArrayIds.next();
						if (!arrayId2multiplicity.increment(arrayId)){
							arrayId2multiplicity.put(arrayId, 1);
						}
					}
				}
			}
			
	    if (previousArrayIds.size() != arrayId2multiplicity.size()
	        || !previousArrayIds.containsAll(arrayId2multiplicity.keys())) {
	      // something changed -> add interval
	      if (!previousArrayIds.isEmpty()){
	        Interval interval = IntervalProvider.getInstance().get(lastLimit, limit, previousArrayIds);
	        result.add(interval);
	      }
	      
	      lastLimit = limit;
	    }      
		}
		intervalProvider.handBack(intervals);
		
		if (ReasonerConfig.debugMode && !arrayId2multiplicity.isEmpty()) {
			throw new Exception("bug in makeNonOverlapping");
		}
		return result;
	}
	
	public void clear() throws Exception {
		timer.reset(IntervalLogic.class.getName());
	}
}
