package reasoner.grounding;

import java.util.LinkedList;
import java.util.List;

import reasoner.ReasonerConfig;
import reasoner.io.Time;
import reasoner.io.dataset.Relation;
import reasoner.lineage.LineageArray;
import reasoner.lineage.confidence.ConfidenceCalculator;

/**
 * This class encapsulates a fact.
 * A fact is an instance of a @Relation.
 * 
 * Confidence calculations are triggered here.
 */
public class Tuple {
  public final Relation relation;
	public final String[] nonTempArgs;
	public final List<Interval> intervals;
	public final double prior;
	
	public Tuple(Relation relation, double prior, String... nonTempArgs) throws Exception {
	  this.relation = relation;
		intervals = new LinkedList<Interval>();
		this.nonTempArgs = nonTempArgs;
		this.prior = prior;
		if (ReasonerConfig.debugMode) {
			if (!relation.isExtensional && prior >= 0.0 && prior <= 1.0) {
				throw new Exception("only extensional relations should use the prior! " + relation);
			}
		  if (nonTempArgs.length != relation.numNonTempArgs) {
		    throw new Exception("Invalid number of arguments: "+this);
		  }
		}
	}
	
	public Tuple(Relation relation, String... nonTempArgs) throws Exception {
	  this(relation, -1.0, nonTempArgs);
	}
	
	public void calcConfExact(LineageArray lineageArray, ConfidenceCalculator calculator, double constraintConf) throws Exception {
		for (Interval interval : this.intervals){
			int lineageId = interval.lineageArrayIds.toArray()[0];
			double conf = calculator.confidence(lineageArray.getLineage(lineageId));
			interval.confidence = conf / constraintConf;
		}
	}
	
	public void calcConfExact(LineageArray lineageArray, ConfidenceCalculator calculator) throws Exception {
		calcConfExact(lineageArray, calculator, 1.0);
	}
	
	public void normalizeWorlds(int numWorlds){
		for (Interval interval : this.intervals){
			interval.confidence /= 1.0 * numWorlds;
		}
	}
	
	/**
	 * returns only the atemporal fact as string.
	 * e.g. isMarriedTo(David_Beckham,Victoria_Beckham)
	 */
	public String factString() {
	  StringBuffer buffer = new StringBuffer();
    buffer.append(relation.name);
    buffer.append("(");
    for (int i = 0; i < nonTempArgs.length; i++) {
      if (i > 0) {
        buffer.append(", ");
      }
      buffer.append("\"" + nonTempArgs[i] + "\"");
    }
    buffer.append(")");
    return buffer.toString();
	}
	
	public String toString(){
		if (!relation.isTemporal) {
			return factString() + " " + intervals.get(0).confidence;
		} else {
			Time time = Time.getInstance();
			String factString = factString();
			factString = factString.substring(0, factString.length() - 1);
			factString += ", ";
			StringBuffer buffer = new StringBuffer();
			int i = 0;
			for (Interval interval : intervals) {
				if (i > 0) {
					buffer.append("\n");
				}
				buffer.append(factString);
				buffer.append("'" + time.int2string(interval.begin) + "'");
				buffer.append(", ");
				buffer.append("'" + time.int2string(interval.end) + "'");
				buffer.append(") ");
				buffer.append(interval.confidence);
				i += 1;
			}
			return buffer.toString();
		}
	}
}
