package reasoner.io;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.set.hash.TIntHashSet;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.HashSet;

import reasoner.io.dataset.Dataset;
import reasoner.io.dataset.Relation;
import reasoner.lineage.LineageArray;

/**
 * manages the data sets
 * 
 * is a Singelton
 * 
 */
public class DatabaseManager {
	
	private Connection con;
	private static DatabaseManager instance;
	
	private Dataset currentDataset;
	
	private DatabaseManager() throws Exception {
    con = DriverManager.getConnection("jdbc:postgresql://" + DatabaseConfig.host + ":" + DatabaseConfig.port + "/" + DatabaseConfig.databaseName, DatabaseConfig.user, DatabaseConfig.password);
		currentDataset = null;
		init();
	}
	
	public static DatabaseManager getInstance() throws Exception {
		if (instance == null) {
			instance = new DatabaseManager();
		}
		return instance;
	}
	
	public Connection getConnection() {
	  return con;
	}
	
	private void init() throws Exception {
		PreparedStatement stmt = con.prepareStatement("SELECT tablename FROM pg_tables");
		
		boolean hasDatasetTable = false;
		ResultSet rset = stmt.executeQuery();
		while (rset.next()) {
			if (rset.getString(1).equals("datasets")) {
				hasDatasetTable = true;
				break;
			}
		}
		rset.close();
		stmt.close();
		
		if (!hasDatasetTable) {
			stmt = con.prepareStatement("CREATE TABLE datasets (dataset_id integer NOT NULL, dataset_name character varying, max_array_id integer, max_time integer, min_time integer, CONSTRAINT datasets_primary PRIMARY KEY (dataset_id)) WITH (OIDS=FALSE);");
			stmt.execute();
			stmt.close();
			stmt = con.prepareStatement("CREATE TABLE dataset_tables (dataset_id integer, table_name character varying, non_temp_args integer, is_temporal boolean, is_uncertain boolean) WITH (OIDS=FALSE);");
			stmt.execute();
			stmt.close();
		}
	}
	
	// === BEGIN NEW DATASET ===
	/**
	 * the functions to create a dataset have to be called in this order:
	 * createDataset()
	 * addNewRelation()^*
	 * finalizeDataset()
	 * 
	 * @param name : name of new dataset
	 * @return dataset id
	 */
	public int createDataset(String name) throws Exception {
		name = name.toLowerCase();
	  if (currentDataset != null) {
	    throw new Exception("unload old dataset first: " + currentDataset.name);
	  }
	  boolean existedAlready = true;
    try {
      loadDatasetPropertiesByName(name);
    } catch (Exception e) {
      existedAlready = false;
    }
    if (existedAlready) {
      throw new Exception("A dataset with the name " + name + " existed already");
    }
	  
	  // determine dataset id
    int datasetId = 0;
	  PreparedStatement stmt = con.prepareStatement("SELECT MAX(dataset_id) FROM datasets");
    ResultSet rset = stmt.executeQuery();
    if (rset.next()) {
      datasetId = rset.getInt(1) + 1;
    }
    rset.close();
    stmt.close();
    
    stmt = con.prepareStatement("CREATE TABLE " + name + "_confidences (array_id integer, confidence double precision);");
    stmt.execute();
    
    DatasetProperties properties = new DatasetProperties();
    properties.name = name;
    properties.datasetId = datasetId;
    properties.maxArrayId = 0;
    currentDataset = new Dataset(properties);
    
    Time.getInstance().minTime = Integer.MAX_VALUE;
    Time.getInstance().maxTime = Integer.MIN_VALUE;
    
    return datasetId;
	}
	
	public Relation newRelation(int datasetId,
	                     String relationName,
	                     int numNonTempArgs,
	                     boolean hasTemporalArg,
	                     boolean isExtensional,
	                     boolean isUncertain) throws Exception {
	  if (currentDataset == null
	      || currentDataset.id != datasetId) {
	    throw new Exception("could not add relation " + relationName + " to dataset");
	  }
	  if (currentDataset.relations.containsKey(relationName)) {
	    // it was already added
	    return currentDataset.relations.get(relationName);
	  }
	  HashSet<String> allTableNames = getAllTableNames();
	  if (allTableNames.contains(relationName)) {
	    throw new Exception("Relation existed already: "+relationName);
	  }
	  currentDataset.createRelation(relationName, numNonTempArgs, hasTemporalArg, isExtensional, isUncertain, con);
	  return currentDataset.relations.get(relationName);
	}
	
	/**
	 * the dataset which is created right now is written to database
	 * and unloaded
	 */
	public void finalizeDataset(int datasetId, int maxArrayId) throws Exception {
	  if (currentDataset == null
	      || currentDataset.id != datasetId) {
	    throw new Exception("Could not finalize dataset "+datasetId);
	  }
	  if (Time.getInstance().maxTime == Integer.MIN_VALUE) {
	  	Time.getInstance().maxTime = 100000;
	  }
	  if (Time.getInstance().minTime == Integer.MAX_VALUE) {
	  	Time.getInstance().minTime = -100000;
	  }
	  shrinkMinMaxTime();
	  currentDatasetToDatabase(maxArrayId);
	  currentDataset = null;
	}
	
	private void shrinkMinMaxTime() throws Exception {
		PreparedStatement stmt;
		for (Relation relation : currentDataset.relations.values()) {
			if (!tableExists(relation.name)) {
				throw new Exception("Table " + relation.name + " did not exist. Was the table created?");
			}
			
    	System.out.println("Updating time intervals of " + relation.name);
  		stmt = con.prepareStatement("UPDATE " + relation.name + " SET time_begin=? WHERE time_begin=?");
  		stmt.setInt(1, Time.getInstance().minTime);
  		stmt.setInt(2, Integer.MAX_VALUE);
  		stmt.execute();
  		stmt.close();

  		stmt = con.prepareStatement("UPDATE " + relation.name + " SET time_end=? WHERE time_end=?");
  		stmt.setInt(1, Time.getInstance().maxTime);
  		stmt.setInt(2, Integer.MIN_VALUE);
  		stmt.execute();
  		stmt.close();
    }
	}
	
	private boolean tableExists(String tableName) throws Exception {
		PreparedStatement stmt = con.prepareStatement("SELECT tablename FROM pg_tables WHERE tablename = ?");
		stmt.setString(1, tableName);
		ResultSet rset = stmt.executeQuery();
		boolean exists = rset.next();
		rset.close();
		stmt.close();
		return exists;
	}
	
	 /**
   * create a new dataset in database
   */
  private void currentDatasetToDatabase(int arrayId) throws Exception {
    PreparedStatement stmt = con.prepareStatement("INSERT INTO datasets (dataset_id, dataset_name, max_array_id, max_time, min_time) VALUES (?, ?, ?, ?, ?)");
    stmt.setInt(1, currentDataset.id);
    stmt.setString(2, currentDataset.name);
    stmt.setInt(3, arrayId);
    stmt.setInt(4, Time.getInstance().maxTime);
    stmt.setInt(5, Time.getInstance().minTime);
    stmt.execute();
    
    stmt = con.prepareStatement("INSERT INTO dataset_tables (dataset_id, table_name, non_temp_args, is_temporal, is_uncertain) VALUES (?, ?, ?, ?, ?)");
    // extensional relations
    for (Relation relation : currentDataset.relations.values()) {
      if (!relation.isExtensional) {
        throw new Exception("only extensional relations can be stored in database: " + relation.name);
      }
      stmt.setInt(1, currentDataset.id);
      stmt.setString(2, relation.name);
      stmt.setInt(3, relation.numNonTempArgs);
      stmt.setBoolean(4, relation.isTemporal);
      stmt.setBoolean(5, relation.isUncertain);
      stmt.addBatch();
    }
    // confidences table
    stmt.setInt(1, currentDataset.id);
    stmt.setString(2, currentDataset.name + "_confidences");
    stmt.setNull(3, java.sql.Types.NULL);
    stmt.setNull(4, java.sql.Types.NULL);
    stmt.setNull(5, java.sql.Types.NULL);
    stmt.addBatch();  // non_temp_args, is_temporal, is_uncertain is NULL
    
    stmt.executeBatch();
    
    stmt = con.prepareStatement("CREATE INDEX " + currentDataset.name + "_confidences_index ON " + currentDataset.name + "_confidences (array_id);");
    stmt.execute();
  }
  // === END NEW DATASET ===
	
	public void deleteDataset(String name) throws Exception {
		name = name.toLowerCase();
		DatasetProperties properties = loadDatasetPropertiesByName(name);
		Dataset dataset = loadDataset(properties);
		for (String tableName : dataset.relations.keySet()) {
			deleteTable(tableName);
		}
		
		PreparedStatement stmt = con.prepareStatement("DELETE FROM dataset_tables WHERE dataset_id=" + properties.datasetId);
		stmt.execute();
		stmt = con.prepareStatement("DELETE FROM datasets WHERE dataset_id=" + properties.datasetId);
		stmt.execute();
		stmt.close();
		clearDBall();
	}
	
	/**
	 * clears both intensional and extensional data
	 */
	public void clearDBall() throws Exception {
		System.out.println("Analyzing stored datasets");
		//	 get all dataset ids
    TIntHashSet datasetIds = new TIntHashSet();
    PreparedStatement stmt = con.prepareStatement("SELECT dataset_id FROM datasets;");
    ResultSet rset = stmt.executeQuery();
		while (rset.next()) {
			datasetIds.add(rset.getInt(1));
		}
		rset.close();
		stmt.close();
    
		// delete all dataset tables not belonging to a dataset.
    StringBuffer condition = new StringBuffer();
    if (!datasetIds.isEmpty()) {
      condition.append("WHERE ");
      TIntIterator it = datasetIds.iterator();
      for (int i = 0; i < datasetIds.size(); i++) {
        if (i > 0) {
          condition.append("AND ");
        }
        condition.append("dataset_id != ");
        condition.append(it.next());
        condition.append(" ");
      }
	  }
    condition.append(";");
    stmt = con.prepareStatement("DELETE FROM dataset_tables " + condition.toString());
    stmt.execute();
    stmt.close();
    
    clearDBintensional();
	}
	
	public void clearDBintensional() throws Exception {
		System.out.println("Deleting intensional tables");
	  // delete tables
		HashSet<String> tablesToKeep = getAllTableNames();
		PreparedStatement stmt = con.prepareStatement("SELECT tablename FROM pg_tables");
		ResultSet rset = stmt.executeQuery();
		while (rset.next()) {
			String tableName = rset.getString(1);
			
			if (tableName.startsWith("pg_")
					|| tableName.startsWith("sql_")) {
				continue;  // Postgres internal tables
			}
			
			if (tableName.equals("datasets")
					|| tableName.equals("dataset_tables")) {
				continue;  // TPDB internal tables
			}
			
			if (tablesToKeep.contains(tableName)) {
				continue;  // datasets' tables
			}
			
			deleteTable(tableName);
		}
		rset.close();
	}
	
	private void deleteTable(String tableName) throws Exception {
	  try {
	    con.prepareStatement("DROP TABLE " + tableName).execute();
	  } catch (Exception e) {
	    // if the table does not exist, we are fine
	  }
	}

	private DatasetProperties loadDatasetPropertiesByName(String name) throws Exception {
		name = name.toLowerCase();
		PreparedStatement stmt = con.prepareStatement("SELECT dataset_id, max_array_id, max_time, min_time FROM datasets WHERE dataset_name=?");
		stmt.setString(1, name);
		ResultSet rset = stmt.executeQuery();
		if (!rset.next()) {
			rset.close();
			throw new Exception("Dataset did not exist: " + name);
		}
		DatasetProperties properties = new DatasetProperties();
		properties.name = name;
		properties.datasetId = rset.getInt(1);
		properties.maxArrayId = rset.getInt(2);
		properties.maxTime = rset.getInt(3);
		properties.minTime = rset.getInt(4);
		rset.close();
		stmt.close();
		return properties;
	}

	public Dataset getDataset(String name) throws Exception {
	  DatasetProperties properties = loadDatasetPropertiesByName(name);
	  return loadDataset(properties);
	}
	
	/**
	 * for unit tests a dataset might be necessary
	 */
	public void loadDummyDataset() {
		DatasetProperties properties = new DatasetProperties();
		properties.datasetId = -1;
		properties.maxArrayId = 0;
		properties.name = "noDBdataset";
		Dataset dataset = new Dataset(properties);
		LineageArray.getInstance().setDataset(dataset);
	}
	
	/**
	 * load dataset tables, set min / max time, and return dataset
	 */
	private Dataset loadDataset(DatasetProperties properties) throws Exception {
		Time time = Time.getInstance();
		time.maxTime = properties.maxTime;
		time.minTime = properties.minTime;
		
	  Dataset dataset = new Dataset(properties);
		PreparedStatement stmt = con.prepareStatement("SELECT table_name, non_temp_args, is_temporal, is_uncertain FROM dataset_tables WHERE dataset_id=?");
		stmt.setInt(1, properties.datasetId);
		ResultSet rset = stmt.executeQuery();
		while (rset.next()) {
			if (rset.getString(1).startsWith(properties.name + "_confidences")) {
				continue;
			}
			dataset.createRelation(rset.getString(1),  // relation name
			                    rset.getInt(2),  // non_temp_args
			                    rset.getBoolean(3),  // is_temporal
			                    true, // is extensional
			                    rset.getBoolean(4), // is_uncertain
			                    con);
		}
		rset.close();
		stmt.close();
		
		LineageArray.getInstance().setDataset(dataset);
		return dataset;
	}
	
	/**
	 * @return all tables of all datasets
	 */
	private HashSet<String> getAllTableNames() throws Exception {
		HashSet<String> allTables = new HashSet<String>();
		PreparedStatement stmt = con.prepareStatement("SELECT table_name FROM dataset_tables");
		ResultSet rset = stmt.executeQuery();
		while (rset.next()) {
			allTables.add(rset.getString(1).toLowerCase());
		}
		rset.close();
		stmt.close();
		return allTables;
	}
	
	public class DatasetProperties {
		public String name;
		public int datasetId;
		public int maxArrayId;
		public int maxTime;
		public int minTime;
	}
}
