package reasoner.io;

import java.util.LinkedList;

import org.antlr.v4.runtime.ANTLRFileStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.tree.ParseTreeWalker;

import reasoner.io.antlr.DeductionLexer;
import reasoner.io.antlr.DeductionParser;
import reasoner.io.antlr.ParseTreeVisitor;
import reasoner.io.antlr.DeductionParser.ConstraintsContext;
import reasoner.io.antlr.DeductionParser.ProgramContext;
import reasoner.io.dataset.Dataset;
import reasoner.io.dataset.Relation;
import reasoner.query.compiled.CompiledConstraint;
import reasoner.query.compiled.Query;
import reasoner.query.parsed.ParsedConstraint;
import reasoner.query.parsed.ParsedQuery;
import util.io.FileScanner;

public class QueryParser {
	
	private Dataset dataset;
	
	public QueryParser(Dataset dataset) throws Exception {
		this.dataset = dataset;
	}
	
	/**
	 * Parses a list of constants (one per line) from a file.
	 * These constants should be inserted into the query plan.
	 */
	public void parseConstants(String path2file, Query query) throws Exception {
		System.out.println("Parsing " + path2file);
		FileScanner scanner = new FileScanner(path2file);
		while (scanner.hasNextLine()) {
			query.constantsToSet.add(scanner.nextLine());
		}
		scanner.close();
	}
	
	/**
	 * Parses and compiles a query plan from a file.
	 */
	public Query parseQueryPlan(String path2file) throws Exception {
		System.out.println("Parsing " + path2file);
		DeductionLexer lexer = new DeductionLexer(new ANTLRFileStream(path2file));
    DeductionParser parser = new DeductionParser(new CommonTokenStream(lexer));
    ProgramContext programContext = parser.program();
    ParseTreeWalker walker = new ParseTreeWalker();
    ParseTreeVisitor listener = new ParseTreeVisitor(dataset);
	  listener.parsedQuery = new ParsedQuery();
	  walker.walk(listener, programContext);
	  if (!listener.exceptions.isEmpty()) {
	  	throw listener.exceptions.get(0);
	  }
	  Query query = listener.parsedQuery.compileQuery(dataset);
		return query;
	}
	
	/**
	 * parses constraints from the file and adds their compiled version to the query
	 */
	public void parseConstraints(String path2file, Query query) throws Exception {
		System.out.println("Parsing " + path2file);
		DeductionLexer lexer = new DeductionLexer(new ANTLRFileStream(path2file));
    DeductionParser parser = new DeductionParser(new CommonTokenStream(lexer));
    ConstraintsContext constraintsContext = parser.constraints();
    ParseTreeWalker walker = new ParseTreeWalker();
    ParseTreeVisitor listener = new ParseTreeVisitor(dataset);
    listener.constraints = new LinkedList<ParsedConstraint>();
    walker.walk(listener, constraintsContext);
    if (!listener.exceptions.isEmpty()) {
	  	throw listener.exceptions.get(0);
	  }
    for (ParsedConstraint parsedConstraint : listener.constraints) {
    	CompiledConstraint compiledConstraint = parsedConstraint.compileConstraint(dataset);
    	query.constraints.add(compiledConstraint);
    }
	}
	
	public void parseOutputRelations(String path2file, Query query) throws Exception {
		System.out.println("Parsing " + path2file);
		FileScanner scanner = new FileScanner(path2file);
		while (scanner.hasNextLine()) {
			String line = scanner.nextLine();
			if (line.startsWith("//") || line.length() == 0) {
				continue;  // the line is commented
			}
			String relationName = line.toLowerCase();
			Relation relation = dataset.relations.get(relationName);
			if (relation == null) {
				throw new Exception("Query relation is unknown! " + relationName);
			}
			query.output.add(relation);
		}
		scanner.close();
	}
}
