package reasoner.io;

import reasoner.grounding.Interval;
import util.Integers;

public class Time {
  public int maxTime = 100000; // is set to different values by DatabaseManager
  public int minTime = -100000; // is set to different values by DatabaseManager
  
  private static Time instance;
  
	//every year has 372 = 12 * 31 days
  
  // parsing options
  private String delimiter = "-";
  private String placeholder = "X";
  private boolean forwards = false;
  private String beginToken = "time_begin";
  private String endToken = "time_end";
	
  public static Time getInstance() {
  	if (instance == null) {
  		instance = new Time();
  	}
  	return instance;
  }
  
  public void updateTimeBoundaries(Interval interval) {
  	if (interval.begin < minTime) {
  		minTime = interval.begin - 372;
  	}
  	if (interval.end > maxTime) {
  		maxTime = interval.end + 372;
  	}
  }
  
  /**
	 * delimiter examples: "\\.", "-"
	 * placeholder examples: "XX", "??"
	 * forwards: DD.MM.YYYY backwards: YYYY.MM.DD
	 * beginToken examples: "begin", "bigBang"
	 * endToken examples: "end", "now"
	 */
  public void setParsingOptions(String delimiter, String placeholder, boolean fowards, String beginToken, String endToken) {
  	this.delimiter = delimiter;
  	this.placeholder = placeholder;
  	this.forwards = fowards;
  	this.beginToken = beginToken;
  	this.endToken = endToken;
  }
  
  /**
   * supported input:
   * 02.09.1928
   * placeholder.09.2002
   * placeholder.placeholder.1200
   * 20.09.2008--placeholder.placeholder.2000
   * 20.01.2002--endToken
   * placeholder.placeholder.199placeholder
   * 
   * @return [begin, end)
   */
  public int[] string2int(String date) throws Exception {
  	String[] parts = date.split("--");
  	if (parts.length == 1) {
  		Date[] dates = parseDate(parts[0]);
  		return encodeInterval(dates[0], dates[dates.length - 1]);
  	} else {
  		Date[] beginDates = parseDate(parts[0]);
  		Date[] endDates = parseDate(parts[1]);
  		return encodeInterval(beginDates[0], endDates[endDates.length - 1]);
  	}
  }
  
	/**
	 * parses a date, e.g. 31-12-2012
	 * 
	 * Returns two dates, if the year contained a placeholder:
	 * for 199?-XX-XX it returns: [1990-XX-XX, 1999-XX-XX]
	 * If the year contains a placeholder, day and month are ignored.
   * 
   * @return Date : month / day are -1 if the placeholder was used.
   * 
	 */
	private Date[] parseDate(String toParse) throws Exception {
		if (toParse.equals(beginToken)) {
			Date[] dates = new Date[1];
			dates[0] = new BeginDate();
			return dates;
		}
		if (toParse.equals(endToken)) {
			Date[] dates = new Date[1];
			dates[0] = new EndDate();
			return dates;
		}
		
		String[] parts = toParse.split(delimiter);
		if (parts.length != 3){
			throw new Exception("wrong length " + parts.length + " of date " + toParse);
		}
		
		Date[] dates;
		int yearId;
		if (forwards) {
			yearId = 2;
		} else {
			yearId = 0;
		}
		if (parts[yearId].contains(placeholder)) {
			// e.g. 199X-XX-XX
			dates = new Date[2];
			for (int i = 0; i < 2; i++) {
				dates[i] = new Date();
				dates[i].month = -1;
				dates[i].day = -1;
			}
			if (parts[yearId].startsWith("-")) {
				// negative year
				dates[0].year = new Integer(parts[yearId].replace(placeholder, "9"));
				dates[1].year = new Integer(parts[yearId].replace(placeholder, "0"));
			} else {
				dates[0].year = new Integer(parts[yearId].replace(placeholder, "0"));
				dates[1].year = new Integer(parts[yearId].replace(placeholder, "9"));
			}
			return dates;
		} else {
			// year is complete, e.g. 1900-XX-XX
			dates = new Date[1];
			dates[0] = new Date();
			dates[0].year = new Integer(parts[yearId]);
		}
		
		
		if (parts[1].equals(placeholder + placeholder)) {
			dates[0].month = -1;
			dates[0].day = -1;
			return dates;
		} else {
			dates[0].month = new Integer(parts[1]);
		}
		
		String dayToParse;
		if (forwards) {
			dayToParse = parts[0];
		} else {
			dayToParse = parts[2];
		}
		if (dayToParse.equals(placeholder + placeholder)) {
			dates[0].day = -1;
			return dates;
		} else {
			dates[0].day = new Integer(dayToParse);
			return dates;
		}
	}
	
	/**
	 * @return [begin, end)
	 */
	private int[] encodeInterval(Date begin, Date end) throws Exception {
		int[] encoded = new int[2];
		encoded[0] = encodeDate(begin)[0];
		encoded[1] = encodeDate(end)[1];
		if (encoded[1] <= encoded[0]) {
			throw new Exception("Time encoding error: end<=begin: " + begin + " " + end);
		}
		return encoded;
	}
	
	/**
	 * encodes a date into an integer pair
	 * 
	 * if date.month == -1 a time span of a year is returned
	 * if date.day == -1 a time span of a month is returned
	 * otherwise the time span is a day
	 * 
	 * @return [begin, end)
	 */
	private int[] encodeDate(Date date) {
		int[] encoded = new int[2];
		if (date instanceof BeginDate) {
			encoded[0] = minTime;
			encoded[1] = minTime + 1;
			return encoded;
		}
		if (date instanceof EndDate) {
			encoded[1] = maxTime;
			encoded[0] = maxTime - 1;
			return encoded;
		}
		encoded[0] = date.year * 372;
		encoded[1] = date.year * 372;
		if (date.month != -1) {
			encoded[0] += (date.month - 1) * 31;
			encoded[1] += (date.month - 1) * 31;
		} else {
			encoded[1] += 372;
			return encoded;
		}
		if (date.day != -1) {
			encoded[0] += (date.day - 1);
			encoded[1] += (date.day - 1) + 1;
			return encoded;
		} else {
			encoded[1] += 31;
			return encoded;
		}
	}
	
	/**
	 * writes one according to the parsing parameters forwards, delimiter
	 */
	public String int2string(int day){
		if (day == maxTime) {
			return endToken;
		}
		if (day == minTime) {
			return beginToken;
		}
		Date date = decodeDate(day);
		StringBuffer buffer = new StringBuffer();
		if (forwards) {
		  buffer.append(Integers.getMinimumLength(date.day, 2));
		} else {
		  buffer.append(Integers.getMinimumLength(date.year, 4));
		}
		buffer.append(delimiter);
		buffer.append(Integers.getMinimumLength(date.month, 2));
		buffer.append(delimiter);
		if (forwards) {
		  buffer.append(Integers.getMinimumLength(date.year, 4));
    } else {
      buffer.append(Integers.getMinimumLength(date.day, 2));
    }
		return buffer.toString();
	}
	
	/**
	 * writes YYYY-MM-DD
	 * if the month or day is unknown, 'XX' is used
	 * 
	 * if the interval does not span a year / month / day
	 * the format
	 * YYYY-MM-DD--YYYY-MM-DD is used
	 */
	public String interval2string(Interval interval) {
		Date begin = decodeDate(interval.begin);
		Date end = decodeDate(interval.end);
		if (interval.end - interval.begin == 1) {
			// it is a day
			return Integers.getMinimumLength(begin.year, 4)
			       + "-" + Integers.getMinimumLength(begin.month, 2)
			       + "-" + Integers.getMinimumLength(begin.day, 2);
		} else if (interval.end - interval.begin == 31) {
			// it is a month
			return Integers.getMinimumLength(begin.year, 4)
      			 + "-" + Integers.getMinimumLength(begin.month, 2)
             + "-XX";
		} else if (interval.end - interval.begin == 372) {
			return Integers.getMinimumLength(begin.year, 4)
             + "-XX"
             + "-XX";
		} else {
			return Integers.getMinimumLength(begin.year, 4)
      			 + "-" + Integers.getMinimumLength(begin.month, 2)
             + "-" + Integers.getMinimumLength(begin.day, 2)
			       + "--"
			       + Integers.getMinimumLength(end.year, 4)
             + "-" + Integers.getMinimumLength(end.month, 2)
             + "-" + Integers.getMinimumLength(end.day, 2);
		}
	}
	
	private Date decodeDate(int timePoint) {
		Date date = new Date();
		date.year = timePoint / 372;
		timePoint %= 372;
		date.month = 1 + timePoint / 31;
		timePoint %= 31;
		date.day = 1 + timePoint;
		return date;
	}
	
	private class Date {
		public int day;  // -1 if unknown
		public int month;  // -1 if unknown
		public int year;
	}
	
	private class BeginDate extends Date {}
	private class EndDate extends Date{}
}
