package reasoner.io.antlr;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;

import org.antlr.v4.runtime.ParserRuleContext;
import org.antlr.v4.runtime.tree.ErrorNode;
import org.antlr.v4.runtime.tree.TerminalNode;

import reasoner.io.antlr.DeductionParser.ArgumentContext;
import reasoner.io.antlr.DeductionParser.ArithmeticLiteralContext;
import reasoner.io.antlr.DeductionParser.CommentContext;
import reasoner.io.antlr.DeductionParser.ConfidenceContext;
import reasoner.io.antlr.DeductionParser.ConjunctionContext;
import reasoner.io.antlr.DeductionParser.ConstantContext;
import reasoner.io.antlr.DeductionParser.ConstraintContext;
import reasoner.io.antlr.DeductionParser.ConstraintsContext;
import reasoner.io.antlr.DeductionParser.DateContext;
import reasoner.io.antlr.DeductionParser.DayContext;
import reasoner.io.antlr.DeductionParser.DeductionRuleContext;
import reasoner.io.antlr.DeductionParser.EmptyLineContext;
import reasoner.io.antlr.DeductionParser.GroundedLiteralContext;
import reasoner.io.antlr.DeductionParser.HeadContext;
import reasoner.io.antlr.DeductionParser.IdentifierContext;
import reasoner.io.antlr.DeductionParser.LabelContext;
import reasoner.io.antlr.DeductionParser.LabelsContext;
import reasoner.io.antlr.DeductionParser.LiteralContext;
import reasoner.io.antlr.DeductionParser.MonthContext;
import reasoner.io.antlr.DeductionParser.NegatedConstraintContext;
import reasoner.io.antlr.DeductionParser.NegatedLiteralContext;
import reasoner.io.antlr.DeductionParser.OperatorContext;
import reasoner.io.antlr.DeductionParser.ProgramContext;
import reasoner.io.antlr.DeductionParser.RelationNameContext;
import reasoner.io.antlr.DeductionParser.RelationalLiteralContext;
import reasoner.io.antlr.DeductionParser.VariableContext;
import reasoner.io.antlr.DeductionParser.YearContext;
import reasoner.io.dataset.Dataset;
import reasoner.io.weightlearning.ParsedLabel;
import reasoner.query.parsed.ArithmeticLiteral;
import reasoner.query.parsed.Constant;
import reasoner.query.parsed.ParsedConjunction;
import reasoner.query.parsed.ParsedConstraint;
import reasoner.query.parsed.ParsedDeductionRule;
import reasoner.query.parsed.ParsedQuery;
import reasoner.query.parsed.RecursionCheck;
import reasoner.query.parsed.RelationalLiteral;
import reasoner.query.parsed.Variable;

public class ParseTreeVisitor implements DeductionListener {
	
	public ParseTreeVisitor(Dataset dataset) {
		this.dataset = dataset;
		bound = new HashSet<Variable>();
		toBind = new HashSet<Variable>();
		name2variable = new HashMap<String, Variable>();
		exceptions = new LinkedList<Exception>();
		recursionCheck = new RecursionCheck();
	}
	
	private Dataset dataset;
	
	// State in the parse tree
	private boolean inDeductionRule;
	private boolean inHead;
	private boolean inConjunction;
	private boolean negated;
	
	// Variables
	// These attributes are cleared after each deduction rule is parsed
	private HashMap<String, Variable> name2variable;
	private HashSet<Variable> bound;
	private HashSet<Variable> toBind;
	
	// Literals, deduction rules, and constraints
	private RelationalLiteral relationalLiteral;
	private ArithmeticLiteral arithmeticLiteral;
	private ParsedConjunction conjunction;
	private ParsedDeductionRule deductionRule;
	private ParsedConstraint constraint;
	public LinkedList<ParsedConstraint> constraints;
	public ParsedQuery parsedQuery;
	
	// labels
	public LinkedList<ParsedLabel> labels;
	private ParsedLabel label;
	
	// Recursion
	private RecursionCheck recursionCheck;
	
	public List<Exception> exceptions;

	public void enterArgument(ArgumentContext ctx) {
		
	}

	public void enterArithmeticLiteral(ArithmeticLiteralContext ctx) {
		arithmeticLiteral = new ArithmeticLiteral(negated);
	}

	public void enterConstant(ConstantContext ctx) {
		String name = ctx.getText();
		name = name.substring(1, name.length() - 1);  // remove quotes
		boolean toSet = name.equals("constant_to_set");
		Constant constant = new Constant(name, toSet);
		if (relationalLiteral != null) {
			relationalLiteral.arguments.add(constant);
		} else if (arithmeticLiteral != null) {
			arithmeticLiteral.addArgument(constant);
		} else if (label != null) {
			if (toSet) {
				exceptions.add(new Exception("Cannot set constants in labels"));
			} else {
				label.arguments.add(constant);
			}
		} else {
			exceptions.add(new Exception("Could not assign constant: " + constant));
		}
	}

	public void enterDate(DateContext ctx) {
		String name = ctx.getText();
		name = name.substring(1, name.length() - 1);  // drop quotes
		Constant constant = new Constant(name, false);
		constant.isTemporal = true;
		if (relationalLiteral != null) {
			relationalLiteral.arguments.add(constant);
		} else if (arithmeticLiteral != null) {
			arithmeticLiteral.addArgument(constant);
		} else if (label != null) {
			label.temporalAnnotation.add(constant);
		} else {
			exceptions.add(new Exception("Could not assign date: " + constant));
		}
	}

	public void enterDay(DayContext ctx) {

	}

	public void enterDeductionRule(DeductionRuleContext ctx) {
		deductionRule = new ParsedDeductionRule();
		inDeductionRule = true;
	}

	public void enterIdentifier(IdentifierContext ctx) {

	}

	public void enterLiteral(LiteralContext ctx) {
		
	}

	public void enterMonth(MonthContext ctx) {

	}

	public void enterNegatedLiteral(NegatedLiteralContext ctx) {
		negated = true;
	}

	public void enterProgram(ProgramContext ctx) {

	}

	public void enterRelationName(RelationNameContext ctx) {
		String relationName = ctx.getText().toLowerCase();
		try {
			if (inConjunction && inDeductionRule) {
				recursionCheck.addBodyRelation(relationName);
			}
			if (inHead) {
				recursionCheck.newRule(relationName);
			}
		} catch (Exception e) {
			exceptions.add(e);
		}
		
		if (relationalLiteral != null) {
			if (dataset.relations.containsKey(relationName)) {
				// extensional
				if (inHead) {
					exceptions.add(new Exception("The relation " + relationName + " is extensional, but occurs in a head literal."));
				}
				relationalLiteral.relation = dataset.relations.get(relationName);
			} else {
				if (inDeductionRule) {
					recursionCheck.addRelationToCheck(relationName);
				}
				relationalLiteral.relationName = relationName;
			}
		}
		
		if (label != null) {
			if (dataset.relations.containsKey(relationName)) {
				label.relation = dataset.relations.get(relationName);
			} else {
				exceptions.add(new Exception("Unknown relation in label: " + relationName));
			}
		}
	}

	public void enterRelationalLiteral(RelationalLiteralContext ctx) {
		relationalLiteral = new RelationalLiteral();
	}

	public void enterVariable(VariableContext ctx) {
		String name = ctx.getText();
		if (!name2variable.containsKey(name)) {
			Variable var = new Variable(name);
			name2variable.put(name, var);
		}
		Variable var = name2variable.get(name);
		
		// Variables
		if (inHead || negated || arithmeticLiteral != null) {
			toBind.add(var);
		} else {
			bound.add(var);
		}
		
		// Deduction rules
		if (relationalLiteral != null) {
			relationalLiteral.arguments.add(var);
		} else {
			arithmeticLiteral.addArgument(var);
		}
	}

	public void enterYear(YearContext ctx) {

	}

	public void exitArgument(ArgumentContext ctx) {

	}

	public void exitArithmeticLiteral(ArithmeticLiteralContext ctx) {
		conjunction.arithmeticLiterals.add(arithmeticLiteral);
		arithmeticLiteral = null;
	}

	public void exitConstant(ConstantContext ctx) {

	}

	public void exitDate(DateContext ctx) {

	}

	public void exitDay(DayContext ctx) {

	}

	public void exitDeductionRule(DeductionRuleContext ctx) {
		checkBoundVariables(ctx.getText());
		
		// Recursion
		try {
			recursionCheck.exitingRule();
		} catch (Exception e) {
			exceptions.add(e);
		}
		
		// Deduction rules
		deductionRule.body = conjunction;
		conjunction = null;
		parsedQuery.rules.add(deductionRule);
		deductionRule = null;
		inDeductionRule = false;
	}

	public void exitIdentifier(IdentifierContext ctx) {

	}

	public void exitLiteral(LiteralContext ctx) {

	}

	public void exitMonth(MonthContext ctx) {

	}

	public void exitNegatedLiteral(NegatedLiteralContext ctx) {
		negated = false;
	}

	public void exitProgram(ProgramContext ctx) {
		try {
			recursionCheck.performCheck();
		} catch (Exception e) {
			exceptions.add(e);
		}
	}

	public void exitRelationName(RelationNameContext ctx) {

	}

	public void exitRelationalLiteral(RelationalLiteralContext ctx) {
		if (inHead) {
			deductionRule.head = relationalLiteral;
		} else {
			if (negated) {
				conjunction.negativeLiterals.add(relationalLiteral);
			} else {
				conjunction.positiveLiterals.add(relationalLiteral);
			}
		}
		
		if (relationalLiteral.relation != null
				&& relationalLiteral.relation.isTemporal) {
			int numArgs = relationalLiteral.arguments.size();
			if (numArgs < 3) {
				exceptions.add(new Exception("Too few arguments for relation: " + relationalLiteral));
			} else {
				relationalLiteral.arguments.get(numArgs - 1).isTemporal = true;
				relationalLiteral.arguments.get(numArgs - 2).isTemporal = true;
			}
		}
		
		relationalLiteral = null;
	}

	public void exitVariable(VariableContext ctx) {

	}

	public void exitYear(YearContext ctx) {

	}

	public void enterEveryRule(ParserRuleContext arg0) {

	}

	public void exitEveryRule(ParserRuleContext arg0) {

	}

	public void visitErrorNode(ErrorNode arg0) {

	}

	public void visitTerminal(TerminalNode arg0) {
		
	}

	public void enterHead(HeadContext ctx) {
		inHead = true;
	}
	
	public void exitHead(HeadContext ctx) {
		inHead = false;
	}

	public void enterComment(CommentContext ctx) {

	}

	public void enterEmptyLine(EmptyLineContext ctx) {

	}

	public void exitComment(CommentContext ctx) {

	}

	public void exitEmptyLine(EmptyLineContext ctx) {

	}

	public void enterOperator(OperatorContext ctx) {
		String name = ctx.getText();
		if (name.equals("=")) {
			arithmeticLiteral.operator = ArithmeticLiteral.Operator.EQUALS;
		} else if (name.equals("<")) {
			arithmeticLiteral.operator = ArithmeticLiteral.Operator.LESS;
		} else if (name.equals(">")) {
			arithmeticLiteral.operator = ArithmeticLiteral.Operator.GREATER;
		} else if (name.equals("<=")) {
			arithmeticLiteral.operator = ArithmeticLiteral.Operator.LESSEQUALS;
		} else if (name.equals(">=")) {
			arithmeticLiteral.operator = ArithmeticLiteral.Operator.GREATEREQUALS;
		} else {
			exceptions.add(new Exception("Unknown arithmetic operator: " + name));
		}
	}

	public void exitOperator(OperatorContext ctx) {

	}

	public void enterConjunction(ConjunctionContext ctx) {
		inConjunction = true;
		conjunction = new ParsedConjunction();
	}

	public void enterConstraint(ConstraintContext ctx) {
		constraint = new ParsedConstraint(false);
	}

	public void enterConstraints(ConstraintsContext ctx) {
		
	}

	public void enterNegatedConstraint(NegatedConstraintContext ctx) {
		constraint = new ParsedConstraint(true);
	}

	public void exitConjunction(ConjunctionContext ctx) {
		inConjunction = false;
		// the conjunction is removed either in the constraint or in the deduction rule
	}

	public void exitConstraint(ConstraintContext ctx) {
		checkBoundVariables(ctx.getText());
		constraint.conjunction = conjunction;
		conjunction = null;
		constraints.add(constraint);
		constraint = null;
	}

	public void exitConstraints(ConstraintsContext ctx) {

	}

	public void exitNegatedConstraint(NegatedConstraintContext ctx) {
		checkBoundVariables(ctx.getText());
		constraint.conjunction = conjunction;
		conjunction = null;
		constraints.add(constraint);
		constraint = null;
	}
	
	private void checkBoundVariables(String context) {
		toBind.removeAll(bound);
		if (!toBind.isEmpty()) {
			StringBuffer message = new StringBuffer();
			message.append("In ");
			message.append(context);
			message.append(" there were unbound variables:");
			for (Variable var : toBind) {
				message.append(" ");
				message.append(var.name);
			}
			Exception e = new Exception(message.toString());
			exceptions.add(e);
		}
		name2variable.clear();
		bound.clear();
		toBind.clear();
	}

	public void enterConfidence(ConfidenceContext ctx) {
		double confidence = new Double(ctx.getText());
		if (confidence > 1.0 || confidence < 0.0) {
			exceptions.add(new Exception("Invalid probability: " + ctx.getText()));
		}
		label.confidenceLabel = confidence;
	}

	public void enterGroundedLiteral(GroundedLiteralContext ctx) {

	}

	public void enterLabel(LabelContext ctx) {
		label = new ParsedLabel();
	}

	public void enterLabels(LabelsContext ctx) {
		
	}

	public void exitConfidence(ConfidenceContext ctx) {

	}

	public void exitGroundedLiteral(GroundedLiteralContext ctx) {

	}

	public void exitLabel(LabelContext ctx) {
		labels.add(label);
		label = null;
	}

	public void exitLabels(LabelsContext ctx) {
		
	}

}
