package reasoner.io.dataset;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.list.linked.TIntLinkedList;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.HashMap;

import reasoner.ReasonerConfig;
import reasoner.io.DatabaseManager.DatasetProperties;
import reasoner.lineage.LineageArray;
import reasoner.lineage.providers.LineageProvider;
import util.io.FileScanner;

public class Dataset {
  public final int id;
  public final String name;
  public HashMap<String, Relation> relations;
  public final int numExtensionalFacts;
  
  private Connection con;
	private TIntLinkedList arrayIdsToLoad;
  
  public Dataset(DatasetProperties properties) {
    name = properties.name.toLowerCase();
    id = properties.datasetId;
    relations = new HashMap<String, Relation>();
    numExtensionalFacts = properties.maxArrayId;
		arrayIdsToLoad = new TIntLinkedList();
  }
  
  /**
   * creates a relation
   * 
   * does not create the table in database!
   */
  public Relation createRelation(String name,
                          int numNonTempArgs,
                          boolean hasTempArg,
                          boolean isExtensional,
                          boolean isUncertain,
                          Connection con) throws Exception {
  	Relation relation = new Relation(this, name, numNonTempArgs, hasTempArg, isExtensional, isUncertain, con);
    relations.put(name, relation);
    this.con = con;
    return relation;
  }
  
  public void loadConfidenceLazy(int arrayId) throws Exception {
  	arrayIdsToLoad.add(arrayId);
  	if (arrayIdsToLoad.size() > 10000) {
  		eagerConfidenceLoading();
  	}
  }

  public void eagerConfidenceLoading() throws Exception {
  	if (arrayIdsToLoad.isEmpty()) {
  		return;
  	}
  	
  	StringBuffer sql = new StringBuffer();
		sql.append("SELECT array_id, confidence FROM ");
		sql.append(name.toLowerCase());
		sql.append("_confidences WHERE array_id IN (");
		TIntIterator it = arrayIdsToLoad.iterator();
		for (int i = 0; i < arrayIdsToLoad.size(); i++) {
			if (i > 0) {
				sql.append(", ");
			}
			int arrayId = it.next();
			sql.append(arrayId);
		}
		sql.append(");");
		arrayIdsToLoad.clear();
		
		System.out.println("Loading Leafs");
		if (ReasonerConfig.debugMode) {
			System.out.println(sql);
		}

		PreparedStatement stmt = con.prepareStatement(sql.toString());
		ResultSet rset = stmt.executeQuery();
		LineageArray array = LineageArray.getInstance();
		while (rset.next()) {
			array.setLeafConfidence(rset.getInt(1), rset.getDouble(2));
		}
		rset.close();
		stmt.close();
  }
  
  public void clear() throws Exception {
  	arrayIdsToLoad.clear();
    LineageArray.getInstance().clearQuery();
    LineageProvider.getInstance().clear();
  }
  
  /**
   * loads confidences from a tsv file and saves them in the database
   * File format:
   * array_id tab confidence [tab optional_other_data]
   */
  public void writeConfidencesToDatabase(String pathToFile) throws Exception {
  	PreparedStatement stmt = con.prepareStatement("UPDATE " + name + "_confidences SET confidence = ? WHERE array_id = ?;");
  	
		FileScanner scanner = new FileScanner(pathToFile);
		while (scanner.hasNextLine()) {
			String line = scanner.nextLine();
			String[] parts = line.split("\t");
			if (parts.length < 2) {
				throw new Exception("Invalid line. Not enought tabs? " + line);
			}
			
			int arrayId = new Integer(parts[0]);
			double confidence = new Double(parts[1]);
			stmt.setDouble(1, confidence);
			stmt.setInt(2, arrayId);
			stmt.addBatch();
		}
		scanner.close();
		
		System.out.println("Inserting into database");
		stmt.executeBatch();
  }
}
