package reasoner.io.dataset;

import gnu.trove.TIntCollection;
import gnu.trove.iterator.TIntIterator;
import gnu.trove.set.hash.TIntHashSet;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.LinkedList;
import java.util.List;

import reasoner.ReasonerConfig;
import reasoner.grounding.Tuple;
import reasoner.grounding.Interval;
import reasoner.grounding.IntervalProvider;
import reasoner.io.Time;
import reasoner.lineage.LineageArray;
import reasoner.lineage.LineageLeaf;
import util.Collections;

/**
 * Represents a relation both in the rules as well as a database table
 */
public class Relation {
  public Dataset dataset;  // the dataset this relation belongs to
  private IntervalProvider intervalProvider;
  
  // relation properties
  public final String name;
  public final int numNonTempArgs;
  public final boolean isTemporal;
  public final boolean isExtensional;
  public final boolean isUncertain;
  
  // SQL statements
  private PreparedStatement insertFact;
  private PreparedStatement queryTable;
  private PreparedStatement createTable;
  private PreparedStatement createIndex;
  private PreparedStatement queryArrayIds;
  private PreparedStatement insertConfidence;
  private PreparedStatement updateConfidence;
  
  public Relation(Dataset dataset,
                  String name,
                  int numNonTempArgs,
                  boolean hasTempArg,
                  boolean isExtensional,
                  boolean isUncertain,
                  Connection con) throws Exception {
    if (ReasonerConfig.debugMode) {
      if (numNonTempArgs < 1) {
        throw new Exception("Each relation must have at least one argument: "+name);
      }
      if (!isUncertain && !isExtensional) {
      	throw new Exception("Only extensional relations can be certain");
      }
    }
    this.dataset = dataset;
    intervalProvider = IntervalProvider.getInstance();
    this.name = name;
    this.numNonTempArgs = numNonTempArgs;
    this.isTemporal = hasTempArg;
    this.isExtensional = isExtensional;
    this.isUncertain = isUncertain;
    createSQLstatements(con);
  }
  
  /**
   * create table in database
   */
  public void createTable() throws Exception {
    createTable.execute();
  }
  
  /**
   * create indeces on the arguments
   * 
   * call after all tuples have been loaded
   */
  public void createIndices() throws Exception {
  	createIndex.execute();
  }
  
  /**
   * Use this function to insert a fact into the database.
   * Processing is batch oriented. First use this function to
   * fill the batch and then execute @executeBatch()
   * 
   * @param arguments : non-temporal arguments of the fact.
   * @param intervals : list of intervals. Is assumed to be sorted and non-overlapping.
   *                    Intervals will be handed back to the provider!
   * @param confidence : for extensional relations only. Set to -1.0, if relation is not extensional
   * @throws Exception
   */
  public void addFactToBatch(String[] arguments, double confidence, Interval... intervals) throws Exception {
    if (ReasonerConfig.debugMode) {
    	StringBuffer argumentsAsString = new StringBuffer();
    	for (String arg : arguments) {
    		argumentsAsString.append(arg);
    		argumentsAsString.append("\t");
    	}
      if (arguments.length != numNonTempArgs) {
        throw new Exception("Tried to add invalid number of arguments in " + name + ": " + argumentsAsString);
      }
      if (isExtensional) {
        if (isUncertain) {
	        if (confidence < 0.0 || confidence > 1.0) {
	          throw new Exception("Invalid confidence for extensional fact in " + name + ": " + confidence);
	        }
        } else {
        	if (confidence != 1.0) {
        		throw new Exception("Certain relations must have confidence 1.0");
        	}
        }
        if (intervals.length != 1) {
          throw new Exception("Extensional facts must have exactly one interval: " + name + " " + intervals);
        }
      } else {
        if (confidence >= 0.0 && confidence <= 1.0) {
          throw new Exception("Valid confidence for intensional fact in " + name + ": " + confidence);
        }
      }
      if (isTemporal) {
        if (intervals.length == 0) {
          throw new Exception("no interval when adding temporal fact: " + name + " " + argumentsAsString);
        }
        
        // are the intervals sorted and non-overlapping?
        for (int i = 1; i < intervals.length; i++) {
          if (intervals[i - 1].end > intervals[i].begin) {
            throw new Exception("Intervals were not sorted or overlapping in " + name + ": "+intervals);
          }
        }
        
        for (Interval interval : intervals) {
          if (isUncertain && interval.lineageArrayIds.size() != 1) {
            throw new Exception("Each interval must have exactly one array_id in " + name + ": "+interval);
          }
          if (!isUncertain && !interval.lineageArrayIds.isEmpty()) {
          	throw new Exception("Certain relations may not have array ids");
          }
        }
      } else {
        // non-temporal
        if (intervals.length != 1) {
          throw new Exception("non-temporal relations must have exactly one time-interval");
        }
        Interval interval = intervals[0];
        if (interval.begin < Time.getInstance().minTime
            || interval.end > Time.getInstance().maxTime) {
          throw new Exception("non-temporal relation with invalid time limits");
        }
        if (isUncertain && interval.lineageArrayIds.size() != 1) {
          throw new Exception("Each interval must have exactly one array_id in " + name + ": "+interval);
        }
        if (!isUncertain && !interval.lineageArrayIds.isEmpty()) {
        	throw new Exception("Certain relations may not have array ids");
        }
      }
    }
    
    // insert one row per interval
    for (Interval interval : intervals) {
	    int stmtId = 1;
	    for (String argument : arguments) {
	      insertFact.setString(stmtId++, argument);
	    }
	    insertFact.setInt(stmtId++, interval.begin);
	    insertFact.setInt(stmtId++, interval.end);
	    if (isUncertain) {
		    int arrayId = interval.lineageArrayIds.toArray()[0];
		    insertFact.setInt(stmtId++, arrayId);
	    }
	    insertFact.addBatch();
	    if (isExtensional && isUncertain) {
	    	int arrayId = interval.lineageArrayIds.toArray()[0];
	    	insertConfidence.setInt(1, arrayId);
	      insertConfidence.setDouble(2, confidence);
	      insertConfidence.addBatch();
	    }
	    intervalProvider.handBack(interval);
    }
  }
  
  /**
   * Write the batch of facts resulting from calls to @addFactToBatch() to database.
   */
  public void executeBatch() throws Exception {
  	try {
  		insertFact.executeBatch();
  	} catch (SQLException e) {
  		while (e != null) {
  			System.err.println(e.getMessage());
  			e = e.getNextException();
  		}
  	}
    if (isUncertain) {
    	insertConfidence.executeBatch();
    }
  }
  
  /**
   * create the SQL statements for this relation
   */
  private void createSQLstatements(Connection con) throws Exception {
    /*
     * CREATE TABLE @name
     * (
     *   arg0 character varying,
     *   arg1 ...
     *   time_begin integer,
     *   time_end integer,
     *   array_id integer  // only if uncertain
     * )
     * WITH (
     *   OIDS=FALSE
     * );
     */
    StringBuffer createTableString = new StringBuffer("CREATE TABLE ");
    createTableString.append(name);
    createTableString.append(" (");
    for (int i = 0; i < numNonTempArgs; i++) {
      createTableString.append("arg");
      createTableString.append(i);
      createTableString.append(" character varying, ");
    }
    createTableString.append("time_begin integer, ");
    createTableString.append("time_end integer");
    if (isUncertain) {
    	createTableString.append(", array_id integer");
    }
    createTableString.append(") WITH (OIDS=FALSE);");
    createTable = con.prepareStatement(createTableString.toString());
    
    /*
     * All permutations of indices:
     * CREATE INDEX @name_index ON (arg0, arg1, ..., time_begin, array_id);
     * ...
     * CREATE INDEX @name_indexN ON (..., arg1, arg0, time_begin, array_id);
     * 
     * And the additional index:
     * CREATE INDEX @name_index_array ON (array_id);
     * 
     * array_id is skipped if relation is certain
     */
    LinkedList<String> arguments = new LinkedList<String>();
    for (int i = 0; i < numNonTempArgs; i++) {
    	arguments.add("arg" + i);
    }
    List<List<String>> permutations = new Collections<String>().createAllPermutations(arguments);
    StringBuffer createIndexString = new StringBuffer();
    for (int i = 0; i < permutations.size(); i++) {
    	createIndexString.append("CREATE INDEX ");
    	createIndexString.append(name);
    	createIndexString.append("_index");
    	createIndexString.append(i);
    	createIndexString.append(" ON ");
    	createIndexString.append(name);
    	createIndexString.append(" (");
    	for (String arg : permutations.get(i)) {
    		createIndexString.append(arg);
    		createIndexString.append(", ");
    	}
    	createIndexString.append("time_begin");
    	if (isUncertain) {
    		createIndexString.append(", array_id");
    	}
    	createIndexString.append(");");
    }
    if (isUncertain) {
	    createIndexString.append("CREATE INDEX ");
	    createIndexString.append(name);
	  	createIndexString.append("_index_array ON ");
	  	createIndexString.append(name);
	  	createIndexString.append(" (array_id);");
    }
  	createIndex = con.prepareStatement(createIndexString.toString());
    
    /*
     * INSERT INTO @name
     * (arg0, arg1, ..., time_begin, time_end, array_id)  // array_id only if uncertain
     * VALUES (?, ?, ?, ..., ?)
     */
    StringBuffer insertFactString = new StringBuffer("INSERT INTO ");
    insertFactString.append(name);
    insertFactString.append(" (");
    StringBuffer values = new StringBuffer();
    for (int i = 0; i < numNonTempArgs; i++) {
      insertFactString.append("arg");
      insertFactString.append(i);
      insertFactString.append(", ");
      values.append("?, ");
    }
    insertFactString.append("time_begin, time_end");
    values.append("?, ?");
    if (isUncertain) {
    	insertFactString.append(", array_id");
    	values.append(", ?");
    }
    insertFactString.append(") VALUES (");
    insertFactString.append(values);
    insertFactString.append(");");
    insertFact = con.prepareStatement(insertFactString.toString());
    
    /*
     * SELECT
     *   arg0, arg1, ...,
     *   time_begin, time_end, array_id, [confidence --if extensional and uncertain]
     *  FROM @name [JOIN @dataset.name_confidences USING (array_id) --if extensional and uncertain]
     *  ORDER BY arg0, arg1, ..., time_begin;
     */
    StringBuffer queryTableString = new StringBuffer("SELECT ");
    for (int i = 0; i < numNonTempArgs; i++) {
      queryTableString.append("arg");
      queryTableString.append(i);
      queryTableString.append(", ");
    }
    queryTableString.append("time_begin, time_end");
    if (isUncertain) {
    	queryTableString.append(", array_id");
    }
    queryTableString.append(" FROM ");
    queryTableString.append(name);
    if (isExtensional && isUncertain) {
    	queryTableString.append(" JOIN ");
    	queryTableString.append(dataset.name);
    	queryTableString.append("_confidences USING (array_id)");
    }
    queryTableString.append(" ORDER BY ");
    for (int i = 0; i < numNonTempArgs; i++) {
    	 queryTableString.append("arg");
       queryTableString.append(i);
       queryTableString.append(", ");
    }
    queryTableString.append("time_begin;");
    queryTable = con.prepareStatement(queryTableString.toString());
    
    if (isUncertain) {
	    /*
	     * SELECT array_id
	     * FROM @name ;
	     */
	    StringBuffer queryArrayIdsString = new StringBuffer();
	    queryArrayIdsString.append("SELECT array_id");
	    queryArrayIdsString.append(" FROM ");
	    queryArrayIdsString.append(name);
	    queryArrayIdsString.append(";");
	    queryArrayIds = con.prepareStatement(queryArrayIdsString.toString());
    
	    /*
	     * INSERT INTO @dataset.name_confidences (array_id, confidence) VALUES (?, ?);
	     */
	    StringBuffer queryConfidenceString = new StringBuffer();
	    queryConfidenceString.append("INSERT INTO ");
	    queryConfidenceString.append(dataset.name);
	    queryConfidenceString.append("_confidences (array_id, confidence) VALUES (?, ?);");
	    insertConfidence = con.prepareStatement(queryConfidenceString.toString());
	    
	    /*
	     * UPDATE @dataset.name_confidences SET confidence = ? WHERE array_id = ?;
	     */
	    StringBuffer updateConfidenceString = new StringBuffer();
	    queryConfidenceString.append("UPDATE ");
	    queryConfidenceString.append(dataset.name);
	    queryConfidenceString.append("_confidences SET confidence = ? WHERE array_id = ?;");
	    updateConfidence = con.prepareStatement(updateConfidenceString.toString());
    }
  }
  
  /**
   * gets all facts from database by querying the table.
   * adds the constraint lineage to each interval.
   * if extensional relations are queried, the confidence is read from @LineageArray not from database.
   */
  public LinkedList<Tuple> getFacts() throws Exception {
  	LineageArray lineages = LineageArray.getInstance();
    LinkedList<Tuple> result = new LinkedList<Tuple>();
    ResultSet rset = queryTable.executeQuery();
    Tuple fact = null;
    String oldArgs = null;
    while (rset.next()) {
      StringBuffer newArgsBuffer = new StringBuffer();
      for (int i = 0; i < numNonTempArgs; i++) {
      	newArgsBuffer.append(rset.getString(i + 1));
      	newArgsBuffer.append("$$");
      }
      String newArgs = newArgsBuffer.toString();
      
      // is this a new fact?
      if (fact == null || !newArgs.equals(oldArgs)) {
        if (fact != null) {
          result.add(fact);
        }
        String[] nonTemporalArgs = new String[numNonTempArgs];
        for (int i = 0; i < numNonTempArgs; i++) {
          nonTemporalArgs[i] = rset.getString(i + 1);
        }
        if (isExtensional && isUncertain) {
        	int arrayId = rset.getInt(numNonTempArgs + 3);
        	double prior;
        	if (lineages.getLineage(arrayId) != null) {
        		prior = ((LineageLeaf)lineages.getLineage(arrayId)).conf;
        	} else {
        		prior = rset.getDouble(4 + numNonTempArgs);
        	}
        	fact = new Tuple(this, prior, nonTemporalArgs);
        } else {
        	fact = new Tuple(this, nonTemporalArgs);
        }
        oldArgs = newArgs;
      }
      
      // time interval
      int begin = rset.getInt(numNonTempArgs + 1);
      int end = rset.getInt(numNonTempArgs + 2);
      Interval interval = null;
      if (isUncertain) {
      	int arrayId = rset.getInt(numNonTempArgs + 3);
      	interval = intervalProvider.get(begin, end, arrayId);
      } else {
      	interval = intervalProvider.get(begin, end);
      }
      if (ReasonerConfig.debugMode) {
      	if (!fact.intervals.isEmpty() && fact.intervals.get(fact.intervals.size() - 1).end > begin) {
      		throw new Exception("Intervals were not ordered! Maybe you entered two tuples with identical non-temporal arguments and overlapping time intervals.");
      	}
      }
      fact.intervals.add(interval);
    }
    if (fact != null) {
      result.add(fact);
    }
    
    rset.close();
    return result;
  }
  
  /**
   * get all array ids of this relation from database
   */
  public TIntCollection getArrayIds() throws Exception {
  	if (!isUncertain) {
  		throw new Exception("Certain relations have no arrayIds");
  	}
  	TIntCollection arrayIds = new TIntHashSet();
  	ResultSet rset = queryArrayIds.executeQuery();
  	while (rset.next()) {
  		arrayIds.add(rset.getInt(1));
  	}
  	rset.close();
  	return arrayIds;
  }
  
  /**
   * Writes the confidence from LineageLeafs in @LineageArray to database.
   * Use this function to store the learned weights. 
   */
  public void writeConfidenceToDatabase() throws Exception {
  	if (!isExtensional) {
  		throw new Exception("Cannot update confidences of non-extensional relation!");
  	}
  	if (!isUncertain) {
  		throw new Exception("Certain relations have no confidence");
  	}
  	
  	LineageArray lineages = LineageArray.getInstance();
  	TIntCollection arrayIds = getArrayIds();
  	TIntIterator it = arrayIds.iterator();
  	for (int i = 0; i < arrayIds.size(); i++) {
  		int arrayId = it.next();
  		if (!lineages.wasLoaded(arrayId)) {
  			continue;
  		}
  		LineageLeaf leaf = (LineageLeaf)lineages.getLineage(arrayId);
  		updateConfidence.setDouble(1, leaf.conf);
  		updateConfidence.setInt(2, arrayId);
  		updateConfidence.addBatch();
  	}
  	updateConfidence.executeBatch();
  }
  
  public String toString() {
    return name;
  }
}
