package reasoner.io.loader;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import reasoner.grounding.Interval;
import reasoner.grounding.IntervalProvider;
import reasoner.io.DatabaseManager;
import reasoner.io.Time;
import reasoner.io.dataset.Relation;
import util.io.FileScanner;

/**
 * The tuple parser
 */
public class TupleParser {
  
  private HashMap<String, Relation> relations;
  private DatabaseManager manager;
  private IntervalProvider intervalProvider;
  private Time time;
  private int datasetId;
  private int arrayId;
  private int tuplesInBatch;
  private String datasetName;
  
  public TupleParser() throws Exception {
    relations = new HashMap<String, Relation>();
    manager = DatabaseManager.getInstance();
    intervalProvider = IntervalProvider.getInstance();
    time = Time.getInstance();
    time.setParsingOptions("-", "X", false, "begin", "end");
    datasetId = -1;
    arrayId = 0;
    tuplesInBatch = 0;
  }
  
  /**
   * call this function before starting to parse
   */
  public void createDataset(String datasetName) throws Exception {
  	manager.clearDBall();
    datasetId = manager.createDataset(datasetName);
    this.datasetName = datasetName;
  }

  /**
   * Call this function after creating the dataset
   * 
	 * relation definition file:
	 * relationName tab numberOfNonTemporalArguments tab isUncertain tab isTemporal
	 * Example:
	 * bornOnDate	1	false	true
	 * 
	 * @return relationNames for file reading
	 */
  public Set<String> parseRelations(String pathToFile) throws Exception {
  	Set<String> relationNames = new HashSet<String>();
  	
  	FileScanner scanner = new FileScanner(pathToFile);
  	while(scanner.hasNextLine()) {
  		String line = scanner.nextLine();
  		if (line.length() == 0 || line.startsWith("//")) {
  			continue;
  		}
  		
  		String[] parts = line.split("\t");
  		if (parts.length != 4) {
  			throw new Exception("Invalid Relation Definition: line");
  		}
  		
  		String relationName = parts[0].trim();
  		if (relationNames.contains(relationName)) {
  			throw new Exception("Duplicate relation name: " + relationName);
  		}
  		
  		relationNames.add(relationName);
  		int numNonTemporalArgs = new Integer(parts[1].trim());
  		boolean isUncertain = parseBoolean(parts[2]);
  		boolean isTemporal = parseBoolean(parts[3]);
  		Relation relation = manager.newRelation(datasetId, datasetName + "_" + relationName.toLowerCase(), numNonTemporalArgs, isTemporal, true, isUncertain);
  		relation.createTable();
  		relations.put(relation.name, relation);
  	}
  	scanner.close();
  	
  	return relationNames;
  }
  
  /**
   * parses a Boolean value from a string
   */
  private boolean parseBoolean(String s) throws Exception {
  	String normalized = s.trim().toLowerCase();
  	if (normalized.equals("true")) {
  		return true;
  	} else if (normalized.equals("false")) {
  		return false;
  	} else throw new Exception("Could not parse boolean parameter: " + s);
  }
  
  /**
   * for each file containing tuples call this function
   * 
   * @param pathToFile : the full path to the file to parse
   * @param separator : the character used to separate columns, e.g. \t
   * @param relationName : relation name 
   * 
   * File format:
	 * nonTempArg0 tab ... tab nonTempArgN [tab confidence] [tab beginTime tab endTime]
	 * confidences are a float or double
	 * times have the format: YYYY-MM-DD
   */
  public void parseTuplesForRelation(String pathToFile, String separator, String relationName) throws Exception {
  	Relation relation = relations.get(datasetName + "_" + relationName.toLowerCase());
  	if (relation == null) {
  		throw new Exception("Could not find relation: " + relationName);
  	}
  	int numEntries = getNumEntries(relation);
  	
    FileScanner scanner = new FileScanner(pathToFile);
    while (scanner.hasNextLine()) {
      String line = scanner.nextLine();
      line = line.trim();
      
      if (line.length() == 0 || line.startsWith("//")) {
      	continue;
      }
      
      String[] parts = line.split(separator);
      if (parts.length != numEntries) {
      	System.err.println("Invalid line: " + line);
      	continue;
      }
      
      for (int i = 0; i < parts.length; i++) {
      	parts[i] = parts[i].trim();
      }
      
      // non-temporal arguments
      String[] nonTempArgs = new String[relation.numNonTempArgs];
      for (int i = 0; i < nonTempArgs.length; i++) {
      	nonTempArgs[i] = parts[i];
      	if (nonTempArgs[i].length() > 2047) {
      		// cut arguments to maximum length of 2047
      		nonTempArgs[i] = nonTempArgs[i].substring(0, 2048);
      	}
      }
      
      int index = nonTempArgs.length;
      double conf = 1.0;
      if (relation.isUncertain) {
      	conf = new Double(parts[index]);
      	index += 1;
      }
      
      // temporal arg
      Interval interval = null;
      int begin = Integer.MAX_VALUE;
      int end = Integer.MIN_VALUE;
      if (relation.isTemporal) {
      	try {
      		begin = time.string2int(parts[index])[0];
      		index += 1;
      	} catch(Exception e) {
      		System.err.println("Error when parsing begin date: " + e.getMessage() + " " + line);
      		continue;
      	}
      	try {
      		if (parts[index].equals("time_end")) {
      			end = time.string2int(parts[index])[1];
      		} else {
      			end = time.string2int(parts[index])[0];
      		}
      		index += 1;
      	} catch(Exception e) {
      		System.err.println("Error when parsing end date: " + e.getMessage() + " " + line);
      		continue;
      	}
      }
      interval = intervalProvider.get(begin, end);
      time.updateTimeBoundaries(interval);
      
      // uncertain relations have an arrayId in the interval
      if (relation.isUncertain) {
      	interval.lineageArrayIds.add(arrayId);
      	arrayId += 1;
      }
      
      // insert tuple
      relation.addFactToBatch(nonTempArgs, conf, interval);
      tuplesInBatch++;
      
      if (tuplesInBatch >= 10000) {
        executeBatches();
      }
    }
    scanner.close();
    
    intervalProvider.allObjectsReturned();
  }
  
  private int getNumEntries(Relation relation) {
  	int entries = relation.numNonTempArgs;
  	if (relation.isUncertain) {
  		entries += 1;
  	}
  	if (relation.isTemporal) {
  		entries += 2;
  	}
  	return entries;
  }
  
  /**
   * call this function after all files have been parsed
   */
  public void parsingFinished() throws Exception {
    executeBatches();
    manager.finalizeDataset(datasetId, arrayId);
	  for (Relation relation : relations.values()) {
	  	System.out.println("Creating indices for " + relation.name);
	  	relation.createIndices();
	  }
  }
  
  /**
   * executes the batch statement for all relations
   */
  private void executeBatches() throws Exception {
  	tuplesInBatch = 0;
    for (Relation relation : relations.values()) {
      relation.executeBatch();
    }
  }
}
