package reasoner.io.weightlearning;

import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import java.util.HashMap;
import java.util.Map.Entry;

import reasoner.io.DatabaseManager;
import reasoner.lineage.Lineage;
import reasoner.lineage.LineageArray;
import reasoner.lineage.LineageLeaf;
import reasoner.lineage.providers.LineageProvider;
import reasoner.weightlearning.problem.LabeledFormula;
import reasoner.weightlearning.problem.ProblemInstance;
import reasoner.weightlearning.problem.TupleToLearn;
import util.io.FileScanner;

public class ProblemParser {
	
	/**
	 * Parses input files for the weight learning without grounding
	 * File structure:
	 * 
	 * tuples
	 * id
	 * 
	 * formulas
	 * a literal: number
	 * a negated literal: ~number
	 * a clause: literal space negated_literal space literal ...
	 * a formula: clause * clause * ...
	 * a line: label_confidence tab formula
	 * 
	 * *** will break here, if parseSolutions==false ***
	 * polynomials
	 * ...
	 * 
	 * solutions
	 * x1=0.1,x2=x3*x4
	 * x2=0.0
	 */
	public ProblemInstance parseProblemInstance(String pathToFile) throws Exception {
		DatabaseManager.getInstance().loadDummyDataset();
		
		FileScanner scanner = new FileScanner(pathToFile);
		TIntHashSet tuples = new TIntHashSet();
		scanner.nextLineContaining("tuples");
		while (scanner.hasNextLine()) {
			String line = scanner.nextLine();
			if (line.contains("formulas")) {
				break;
			}
			tuples.add(new Integer(line.trim()));
		}
		
		loadLineageLeafs(tuples);
		
		HashMap<Lineage, Double> labeledFormulas = new HashMap<Lineage, Double>();
		while (scanner.hasNextLine()) {
			String line = scanner.nextLine();
			if (line.contains("polynomials")) {
				break;
			}
			String[] formulaAndPrior = line.split("\t");
			if (formulaAndPrior.length != 2) {
				throw new Exception("invalid labeled formula line: " + line);
			}
			
			String[] parts = formulaAndPrior[1].split("\\*");
			Lineage[] clauses = new Lineage[parts.length];
			for (int i = 0; i < parts.length; i++) {
				clauses[i] = parseClause(parts[i], tuples.size());
			}
			Lineage formula = LineageProvider.getInstance().getLineageOr(clauses);
			labeledFormulas.put(formula, new Double(formulaAndPrior[0]));
		}
		scanner.close();
		
		return getProblemInstance(pathToFile, labeledFormulas, tuples);
	}
	
	private Lineage parseClause(String input, int maxTupleId) throws Exception {
		String[] parts = input.split(" ");
		Lineage[] literals = new Lineage[parts.length];
		for (int j = 0; j < literals.length; j++) {
			literals[j] = parseLiteral(parts[j], maxTupleId);
		}
		return LineageProvider.getInstance().getLineageAnd(literals);
	}
	
	private Lineage parseLiteral(String input, int maxTupleId) throws Exception {
		boolean isPositive = true;
		if (input.startsWith("~")) {
			isPositive = false;
			input = input.substring(1);
		}
		int arrayId = new Integer(input);
		if (arrayId >= maxTupleId) {
			throw new Exception("In a clause the tupleId " + arrayId + " is out of range <" + maxTupleId);
		}
		Lineage lineage = LineageArray.getInstance().getLineage(arrayId);
		if (!isPositive) {
			lineage = LineageProvider.getInstance().getLineageNot(lineage);
		}
		return lineage;
	}
	
	private ProblemInstance getProblemInstance(String pathToFile, HashMap<Lineage, Double> labeledFormulas, TIntSet tuples) throws Exception {
		LabeledFormula[] formulas = new LabeledFormula[labeledFormulas.size()];
		int i = 0;
		for (Entry<Lineage, Double> e : labeledFormulas.entrySet()) {
			formulas[i] = new LabeledFormula();
			formulas[i].label = e.getValue();
			formulas[i].lineage = e.getKey();
			i++;
		}

		TupleToLearn[] tuplesToLearn = new TupleToLearn[tuples.size()];
		for (i = 0; i < tuples.size(); i++) {
			tuplesToLearn[i] = new TupleToLearn("tuple" + i, (LineageLeaf)LineageArray.getInstance().getLineage(i));
		}
		return new ProblemInstance(pathToFile, formulas, tuplesToLearn);
	}
	
	private void loadLineageLeafs(TIntSet tuples) throws Exception {
		for (int i = 0; i < tuples.size(); i++) {
			if (!tuples.contains(i)) {
				throw new Exception("no tuple with id: " + i + ". they should be continous starting at 0");
			}
			int arrayId = LineageArray.getInstance().addLineage(LineageProvider.getInstance().getLineageLeaf(i, -1.0));
			if (arrayId != i) {
				throw new Exception("arrayId was differing");
			}
		}
	}
}
