package reasoner.io.weightlearning;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;

import org.antlr.v4.runtime.ANTLRFileStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.tree.ParseTreeWalker;

import reasoner.grounding.Tuple;
import reasoner.grounding.Interval;
import reasoner.grounding.IntervalProvider;
import reasoner.io.Time;
import reasoner.io.antlr.DeductionLexer;
import reasoner.io.antlr.DeductionParser;
import reasoner.io.antlr.ParseTreeVisitor;
import reasoner.io.antlr.DeductionParser.LabelsContext;
import reasoner.io.dataset.Dataset;
import reasoner.io.dataset.Relation;
import reasoner.query.parsed.Constant;
import util.io.FileScanner;

public class WeightLearningParser {
	
	private final Time time;
	private final IntervalProvider intervalProvider;
	
	public WeightLearningParser() throws Exception {
		time = Time.getInstance();
		time.setParsingOptions("-", "X", false, "time_begin", "time_end");
		intervalProvider = IntervalProvider.getInstance();
	}
	
	/**
   * Parses facts to be labels for the weight learning procedure.
   */
  public ParsedLabels parseLabels(String pathToFile, Dataset dataset) throws Exception {
  	System.out.println("Parsing " + pathToFile);
		DeductionLexer lexer = new DeductionLexer(new ANTLRFileStream(pathToFile));
    DeductionParser parser = new DeductionParser(new CommonTokenStream(lexer));
    LabelsContext labelsContext = parser.labels();
    ParseTreeWalker walker = new ParseTreeWalker();
    ParseTreeVisitor listener = new ParseTreeVisitor(dataset);
    listener.labels = new LinkedList<ParsedLabel>();
    walker.walk(listener, labelsContext);
    if (!listener.exceptions.isEmpty()) {
	  	throw listener.exceptions.get(0);
	  }
    
    ParsedLabels parsedLabels = new ParsedLabels();
    for (ParsedLabel label : listener.labels) {
    	String[] arguments = new String[label.arguments.size()];
    	int argId = 0;
    	for (Constant constant : label.arguments) {
    		arguments[argId] = constant.name;
    		argId += 1;
    	}
    	
    	String relationName = label.relation.name;
    	if (!parsedLabels.occuringRelations.containsKey(relationName)) {
    		parsedLabels.occuringRelations.put(relationName, label.relation);
    	}
    	
    	Tuple fact = new Tuple(label.relation, arguments);
    	String factString = fact.factString();
    	if (!parsedLabels.labels.containsKey(factString)) {
    		parsedLabels.labels.put(factString, fact);
    	}
    	fact = parsedLabels.labels.get(factString);
    	
    	int[] timePoints = time.string2int(label.temporalAnnotation.get(0).name);
    	int begin = timePoints[0];
    	timePoints = time.string2int(label.temporalAnnotation.get(1).name);
    	int end = timePoints[1];
    	Interval interval = intervalProvider.get(begin, end);
    	interval.confidence = label.confidenceLabel;
    	fact.intervals.add(interval);
    }
    return parsedLabels;
	}
	
  
  /**
   * Parses the priors for all facts in a relation.
   * 
   * Example line:
   * yago2_wasbornin
   */
  public Set<Relation> parseRelationsToLearn(String pathToFile, Dataset dataset) throws Exception {
  	System.out.println("Parsing " + pathToFile);
  	HashSet<Relation> relations = new HashSet<Relation>();
    FileScanner scanner = new FileScanner(pathToFile);
    while (scanner.hasNextLine()) {
    	String line = scanner.nextLine();
    	
    	String relationName = line.toLowerCase().trim();
    	Relation relation = dataset.relations.get(relationName);
    	if (relation == null) {
    		throw new Exception("Relation to learn was not in dataset: " + line);
    	}
    	if (!relation.isExtensional) {
    		throw new Exception("Only extensional relations can be labeled: " + line);
    	}
    	
    	relations.add(relation);
    }
    scanner.close();
    return relations;
  }
}
