package reasoner.lineage;

import reasoner.ReasonerConfig;
import util.provider.Providable;

import gnu.trove.iterator.TIntIntIterator;
import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TIntIntHashMap;
import gnu.trove.procedure.TIntIntProcedure;
import gnu.trove.set.hash.TIntHashSet;

public abstract class Lineage extends Providable {
	public Lineage[] subLineage;
	public TIntHashSet vars; // including children
	public TIntIntHashMap varId2count; // contains only pairs with value > 1
	public RetainMyVariables retainMyVariables;
	public double cachedConf;
	protected int maxShannon;
	
	protected Lineage(){
		vars = new TIntHashSet();
		retainMyVariables = new RetainMyVariables(vars);
		varId2count = new TIntIntHashMap();
		cachedConf = -1.0;
	}
	
	public void setNewChildren(Lineage... children) throws Exception {
	  if (ReasonerConfig.debugMode) {
	    if (children.length == 0) {
	    	throw new Exception("Invalid lineage with no children! " + this);
	    }
	    for (int i = 0; i < children.length; i++) {
	      if (children[i] == null) {
	        throw new Exception("Invalid lineage with child == null: " + this);
	      }
	    }
	  }
	  subLineage = children;
	  vars.clear();
	  for (Lineage l : children) {
	  	vars.addAll(l.vars);
    }
    varCount();
    computeMaxShannon();
    cachedConf = -1.0;
	}
	
	public void clear() {
	  subLineage = null;
	  vars.clear();
	  vars.compact();
	  varId2count.clear();
	  varId2count.compact();
	  cachedConf = -1.0;
	  maxShannon = 0;
	}
	
	public abstract String name();
	
	// computes <varId -> numberOfTimesUsed>
	private void varCount(){
		// init datastructure
	  varId2count.clear();
		TIntIterator it = this.vars.iterator();
		for (int i=0; i<this.vars.size(); i++){
			varId2count.put(it.next(), 0);
		}
		// count
		for (Lineage l : this.subLineage){
			it = l.vars.iterator();
			for (int i=0; i<l.vars.size(); i++){
				varId2count.increment(it.next());
			}
		}
		// remove elements with count==1
		TIntIntIterator itMap = varId2count.iterator();
		int orgSize = varId2count.size();
		for (int i=0; i<orgSize; i++){
			itMap.advance();
			if (itMap.value()==1) {
				itMap.remove();
			} else {
				// we keep it
			}
		}
	}
	
	private void computeMaxShannon() {
		for (int i = 0; i < subLineage.length; i++) {
			if (subLineage[i].maxShannon > maxShannon) {
				maxShannon = subLineage[i].maxShannon;
			}
		}
		maxShannon += varId2count.size();
	}
	
	/**
	 * inner class. Procedure returning true, if variable is contained in this lineage.
	 */
	private class RetainMyVariables implements TIntIntProcedure {
		private TIntHashSet varIds;  // careful, this is a pointer to @varIds of @Lineage. So, varCount() alters both here and in @Lineage.
		
		public RetainMyVariables(TIntHashSet varIds) {
			this.varIds = varIds;
		}
		
		public boolean execute(int key, int value) {
			return varIds.contains(key);
		}
	}
	
	public abstract void acceptVisitor(LineageVisitor visitor) throws Exception;
	
	public int getMaxShannon() {
		return maxShannon;
	}
	
	public String toString(){
		StringBuffer s = new StringBuffer();
		this.toString(s);
		return s.toString();
	}
	
	private void toString(StringBuffer s){
		s.append(this.name());
		if (subLineage != null
		    && this.subLineage.length > 0){
			s.append("[");
			for (Lineage l : this.subLineage){
				l.toString(s);
				s.append(" ");
			}
			s.append("]");
		}
	}
}
