package reasoner.lineage;

import reasoner.io.dataset.Dataset;
import reasoner.lineage.providers.LineageProvider;

/**
 * Keeps the lineage array corresponding
 * to the arrayIds stored in database.
 * 
 * Implemented as a singleton.
 */
public class LineageArray {
	private Lineage[] array;
	private int nextArrayId;
	private Dataset currentDataset;
	private LineageProvider lineageProvider;
	
	private static LineageArray instance;
	
	private LineageArray() {
		array = new Lineage[1024];
		nextArrayId = 0;
		lineageProvider = null;  // loading it eagerly creates an infinite loop!
	}

	public static LineageArray getInstance() {
		if (instance == null) {
			instance = new LineageArray();
		}
		return instance;
	}
	
	public Lineage getLineage(int arrayId) throws Exception {
		if (arrayId >= array.length) {
			throw new Exception("too large arrayId");
		}
		if (array[arrayId] == null) {
			if (arrayId < currentDataset.numExtensionalFacts) {
				if (lineageProvider == null) {
					lineageProvider = LineageProvider.getInstance();
				}
				array[arrayId] = lineageProvider.getLineageLeaf(arrayId, -1.0);
				currentDataset.loadConfidenceLazy(arrayId);
			} else {
				throw new Exception("lineage was null, but non-extensional");
			}
		}
		return array[arrayId];
	}
	
	public void setLeafConfidence(int arrayId, double confidence) throws Exception {
		if (arrayId >= currentDataset.numExtensionalFacts) {
			throw new Exception("cannot set confidence of non-leaf lineage");
		}
		if (arrayId >= array.length) {
			throw new Exception("too large arrayId");
		}
		if (array[arrayId] == null) {
			throw new Exception("no leaf loaded? " + arrayId);
		}
		((LineageLeaf)array[arrayId]).conf = confidence;
	}
	
	public boolean wasLoaded(int arrayId) throws Exception {
		if (arrayId >= array.length) {
			throw new Exception("too large arrayId");
		}
		if (arrayId >= currentDataset.numExtensionalFacts) {
			throw new Exception("Do not query non-existensional lineage to be loaded");
		}
		return array[arrayId] != null;
	}
	
	/**
	 * Adds lineage to the array.
	 * @return the arrayId where the lineage is stored.
	 */
	public int addLineage(Lineage l) {
		if (array.length == nextArrayId) {
			Lineage[] newArray = new Lineage[array.length * 2];
			System.arraycopy(array, 0, newArray, 0, array.length);
			array = newArray;
		}
		int id = nextArrayId;
		array[id] = l;
		nextArrayId++;
		return id;
	}
	
	/**
	 * Clear all lineage except lineage leafs.
	 * To be used after executing a query.
	 */
	public void clearQuery() throws Exception {
		for (int i = currentDataset.numExtensionalFacts; i < array.length; i++) {
			if (array[i] != null) {
				LineageProvider.getInstance().handBack(array[i], false);
				array[i] = null;
			}
		}
		nextArrayId = currentDataset.numExtensionalFacts;
	}
	
	public void finalizeGrounding() throws Exception {
		currentDataset.eagerConfidenceLoading();
	}
	
	/**
	 * Clears all lineage including lineage leaf.
	 * Execute if you want to load another dataset.
	 */
	public void clearAll() throws Exception {
		for (int i = 0; i < array.length; i++) {
			if (array[i] != null) {
				LineageProvider.getInstance().handBack(array[i], true);
				array[i] = null;
			}
		}
		nextArrayId = 0;
		currentDataset = null;
	}
	
	public void setDataset(Dataset dataset) {
		currentDataset = dataset;
		nextArrayId = currentDataset.numExtensionalFacts;
		while (array.length < nextArrayId) {
			array = new Lineage[array.length * 2];
		}
	}
	
	public int getNextArrayId() {
		return nextArrayId;
	}
}
