package reasoner.lineage.confidence;

import gnu.trove.iterator.TIntIntIterator;
import gnu.trove.map.hash.TIntIntHashMap;
import reasoner.ReasonerConfig;
import reasoner.lineage.Lineage;
import reasoner.lineage.LineageAnd;
import reasoner.lineage.LineageArray;
import reasoner.lineage.LineageDisjointOr;
import reasoner.lineage.LineageFalse;
import reasoner.lineage.LineageLeaf;
import reasoner.lineage.LineageNot;
import reasoner.lineage.LineageOr;
import reasoner.lineage.LineageTrue;
import reasoner.lineage.LineageVisitor;
import util.Doubles;
import util.datastructures.Stack;

/**
 * This class computes the confidence based on lineage.
 * It assumes a tuple independent database.
 * For each thread a separate instance is necessary.
 * 
 * Shannon Expansions are implemented here.
 */
public class TupleIndependentConfidence implements LineageVisitor, ConfidenceCalculator {
	
	private Stack<ShannonState> stack;
	private ShannonStateProvider shannonStateProvider;
	private LineageArray lineages;
	
	public TupleIndependentConfidence() throws Exception {
		stack = new Stack<ShannonState>(128);
		shannonStateProvider = new ShannonStateProvider(128);
		lineages = LineageArray.getInstance();
	}
	
	/**
	 * Calculate confidence of the lineage.
	 */
	public double confidence(Lineage lineage) throws Exception {
		if (lineage == null) {
			throw new Exception("Cannot compute confidence of null");
		}
		stack.push(shannonStateProvider.get()); // empty
		confCached(lineage);
		double result = stack.peek().conf;
		shannonStateProvider.handBack(stack.pop());
		shannonStateProvider.handBack(stack.pop());
		if (ReasonerConfig.debugMode) {
			if (!stack.isEmpty()) {
				throw new Exception("The stack was non-empty");
			}
			if (!shannonStateProvider.allObjectsReturned()) {
				throw new Exception("not all objects were returned");
			}
			if (result < 0.0 && !!Doubles.equals(result, 0.0)) {
				throw new Exception("too low confidence");
			}
			if (result > 1.0 && !!Doubles.equals(result, 1.0)) {
				throw new Exception("too high confidence");
			}
		}
		return result;
	}
	
	/**
	 * Calculate derivative of a variable @varId given the lineage @lineage.
	 */
	public double derivative(Lineage lineage, int varId) throws Exception {
		if (ReasonerConfig.debugMode) {
			if (!lineage.vars.contains(varId)) {
				throw new Exception("Lineage to calculate derivative on, does not contain the leaf");
			}
		}
		
		stack.push(shannonStateProvider.get()); // here we store the setting of @varId
		
		stack.peek().arrayId2bool.put(varId, 1);
		confCached(lineage);
		double positiveConf = stack.peek().conf;
		shannonStateProvider.handBack(stack.pop());
		
		stack.peek().arrayId2bool.put(varId, -1);
		confCached(lineage);
		double negativeConf = stack.peek().conf;
		shannonStateProvider.handBack(stack.pop());
		
		shannonStateProvider.handBack(stack.pop());
		
		double derivative = positiveConf - negativeConf;
		if (ReasonerConfig.debugMode) {
			if (!stack.isEmpty()) {
				throw new Exception("The stack was non-empty");
			}
			if (!shannonStateProvider.allObjectsReturned()) {
				throw new Exception("not all objects were returned");
			}
			if (derivative < -1.0 && !Doubles.equals(derivative, -1.0)) {
				throw new Exception("too low derivative");
			}
			if (derivative > 1.0 && !Doubles.equals(derivative, 1.0)) {
				throw new Exception("too high derivative");
			}
		}
		return derivative;
	}
	
	/**
	 * This is the first method being called to calculate confidences.
	 * Stack elements are pushed here.
	 */
	private void confCached(Lineage lineage) throws Exception {
		ShannonState state = shannonStateProvider.get();
		if (ReasonerConfig.debugMode) {
			if (!state.arrayId2bool.isEmpty() || state.conf != -1.0) {
				throw new Exception("New shannon state must be empty");
			}
		}
		state.arrayId2bool.putAll(stack.peek().arrayId2bool);
		state.arrayId2bool.retainEntries(lineage.retainMyVariables);
		stack.push(state);
		if (!state.arrayId2bool.isEmpty()) {
			shannonManage(lineage);
		} else {
			if (lineage.cachedConf > -1.0) {
				stack.peek().conf = lineage.cachedConf;
			} else {
				shannonManage(lineage);
				lineage.cachedConf = stack.peek().conf;
			}
		}
	}
	
	/**
	 * repeatedly invokes shannon expansions
	 */
	private void shannonManage(Lineage lineage) throws Exception {
		// the stack element on top is already the one we need
		TIntIntHashMap newShannons = shannonNew(lineage);
		if (newShannons.isEmpty()) {
			lineage.acceptVisitor(this);
		} else {
			double result = 0.0;
			stack.peek().arrayId2bool.putAll(newShannons);
			for (int i = 0; i < Math.pow(2, newShannons.size()); i++) {
				double confShannonState = shannonState(newShannons);
				if (!Doubles.equals(confShannonState, 0.0)) {
					lineage.acceptVisitor(this);
					confShannonState *= stack.peek().conf;
					result += confShannonState;
				}
				
				shannonNext(newShannons);
			}
			stack.peek().conf = result;
		}
	}
	
	/**
	 * returns a hashMap with all the new Shannon Expansions
	 * old shannon expansions coming from a parent won't be added again
	 */
	private TIntIntHashMap shannonNew(Lineage lineage) throws Exception {
		TIntIntHashMap newShannons = new TIntIntHashMap();
		TIntIntIterator it = lineage.varId2count.iterator();
		for (int i = 0; i < lineage.varId2count.size(); i++){
			it.advance();
			if (stack.peek().arrayId2bool.containsKey(it.key())) {
				continue;
			}
			newShannons.put(it.key(), 0);
		}
		return newShannons;
	}
	
	/**
	 * computes confidence of current state in Shannon Expansion mode 
	 */
	private double shannonState(TIntIntHashMap toConsider) throws Exception {
		double conf = 1.0;
		TIntIntIterator it = toConsider.iterator();
		for (int i = 0; i < toConsider.size(); i++) {
			it.advance();
			double confLeaf = ((LineageLeaf)lineages.getLineage(it.key())).conf;
			if (stack.peek().arrayId2bool.get(it.key()) == 1) {
				conf *= confLeaf;
			} else {
				conf *= 1.0 - confLeaf;
			}
		}
		return conf;
	}
	
	/**
	 * moves the map @state to the next shannon state, where only variables in @toConsider are changed
	 */
	private void shannonNext(TIntIntHashMap toConsider) throws Exception {
		TIntIntIterator it = toConsider.iterator();
		for (int i = 0; i < toConsider.size(); i++) {
			it.advance();
			int val = stack.peek().arrayId2bool.get(it.key());
			if (val == 0){
				stack.peek().arrayId2bool.put(it.key(), 1);
				break;
			} else {
				stack.peek().arrayId2bool.put(it.key(), 0);
			}
		}
	}
	
	/**
	 * All the visit methods implement the specific confidence computations.
	 * They pop the stack element of their children from the stack.
	 */
	public void visit(LineageAnd and) throws Exception {
		stack.peek().conf = 1.0;
		for (Lineage child : and.subLineage) {
			confCached(child);
			double confChild = stack.peek().conf;
			shannonStateProvider.handBack(stack.pop());
			if (Doubles.equals(confChild, 0.0)) {
		    stack.peek().conf = 0.0;
		    break;
		  } else {
		  	stack.peek().conf *= confChild;
		  }
		}
	}

	public void visit(LineageOr or) throws Exception {
		stack.peek().conf = 1.0;
		for (Lineage child : or.subLineage) {
			confCached(child);
			double confChild = 1.0 - stack.peek().conf;
			shannonStateProvider.handBack(stack.pop());
			if (Doubles.equals(confChild, 0.0)) {
				stack.peek().conf = 0.0;
		    break;
			} else {
				stack.peek().conf *= confChild;
			}
		}
		stack.peek().conf = 1.0 - stack.peek().conf;
	}

	public void visit(LineageDisjointOr disjointOr) throws Exception {
		if (stack.peek().arrayId2bool.contains(disjointOr.leaf.id)) {
			if (stack.peek().arrayId2bool.get(disjointOr.leaf.id) == 1) {
				confCached(disjointOr.subLineage[0]);
				double confChild = stack.peek().conf;
				shannonStateProvider.handBack(stack.pop());
				stack.peek().conf = confChild;
			} else {
				confCached(disjointOr.subLineage[1]);
				double confChild = stack.peek().conf;
				shannonStateProvider.handBack(stack.pop());
				stack.peek().conf = confChild;
			}
		} else {
			confCached(disjointOr.subLineage[0]);
			double confPosChild = stack.peek().conf;
			shannonStateProvider.handBack(stack.pop());
			confCached(disjointOr.subLineage[1]);
			double confNegChild = stack.peek().conf;
			shannonStateProvider.handBack(stack.pop());
			stack.peek().conf = disjointOr.leaf.conf * confPosChild
				+ (1.0 - disjointOr.leaf.conf) * confNegChild;
		}
	}

	public void visit(LineageNot not) throws Exception {
		confCached(not.subLineage[0]);
		double confChild = stack.peek().conf;
		shannonStateProvider.handBack(stack.pop());
		stack.peek().conf = 1.0 - confChild;
	}

	public void visit(LineageLeaf leaf) throws Exception {
		if (ReasonerConfig.debugMode) {
	    if (lineages.getLineage(leaf.id) != leaf) {
	      throw new Exception("wrong array id in lineage leaf! " + leaf);
	    }
	    if (stack.peek().arrayId2bool.size() > 1) {
	    	throw new Exception("Only one variable in Leaf, how should I be so many shannons? " + leaf);
	    }
	    if (stack.peek().arrayId2bool.size() == 1
	    		&& !stack.peek().arrayId2bool.containsKey(leaf.id)) {
	    	throw new Exception("Wrong array id in shannon map " + leaf);
	    }
	    if (leaf.conf < 0.0 || leaf.conf > 1.0) {
	    	throw new Exception("Invalid confidence in leaf: " + leaf);
	    }
	  }
		if (stack.peek().arrayId2bool.contains(leaf.id)) {
			if (stack.peek().arrayId2bool.get(leaf.id) == 1) {
				stack.peek().conf = 1.0;
			} else {
				stack.peek().conf = 0.0;
			}
		} else {
			stack.peek().conf = leaf.conf;
		}
	}

	public void visit(LineageTrue lineage) throws Exception {
		stack.peek().conf = 1.0;
	}

	public void visit(LineageFalse lineage) throws Exception {
		stack.peek().conf = 0.0;
	}

	public void clear() {
		shannonStateProvider.compact();
	}
}
