package reasoner.lineage.providers;

import java.util.LinkedList;

import gnu.trove.iterator.TIntIntIterator;
import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TIntIntHashMap;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;
import reasoner.ReasonerConfig;
import reasoner.lineage.Lineage;
import reasoner.lineage.LineageAnd;
import reasoner.lineage.LineageArray;
import reasoner.lineage.LineageOr;
import reasoner.lineage.shannonmaterializer.ShannonMaterializer;
import reasoner.lineage.shannonmaterializer.ShannonMaterializerProvider;
import util.datastructures.MappedUnionFind;
import util.provider.AbstractProvider;
import util.time.Timer;

/**
 * This class should be extended by lineageOr and lineageAnd providers.
 * It alters lineage to speed up confidence computations exponentially.
 * In practice, this has huge impact.
 * 
 * 1. Feature: It detects sets of children with disjoint shannon variables
 *   Example:
 *   OR [ (var1 /\ var2), (var1 /\ var3), (var4 /\ var5), (var4 /\ var6) ]
 *   would have two Shannon expansions. One Expansion on var1 and one on var2.
 *   Splitting the OR into:
 *   OR [ OR [ (var1 /\ var2), (var1 /\ var3) ], OR [ (var4 /\ var5), (var4 /\ var6) ] ]
 *   does not alter the semantics (since OR is associative).
 *   However the confidence calculations are sped up by a factor of 2 (more general,
 *   exponentially!).
 * 
 * 2. Feature: If feature 1 is not applicable, it will perform materializations
 *    of Shannon Expansions creating DisjointOr instances.
 *    It picks the Shannon variable common to most children.\
 * 
 * Feature 1 & 2 are called recursively. It breaks when all children have less
 * than @maxShannonExpansion Shannon variables.
 *
 * @param <T> the lineage type, e.g. LineageAnd
 */
public abstract class AbstractLineageProvider<T extends Lineage> extends AbstractProvider<T> {
	
	private LineageProvider lineageProvider;
	private LineageArray lineageArray;
	private ShannonMaterializerProvider shannonMaterializerProvider;
	private Timer timer;
  
  protected AbstractLineageProvider(
  		String name,
  		int initialSize,
  		LineageProvider lineageProvider,
  		ShannonMaterializerProvider shannonMaterializerProvider) throws Exception {
    super(name, initialSize);
    this.lineageProvider = lineageProvider;
    lineageArray = LineageArray.getInstance();
    this.shannonMaterializerProvider = shannonMaterializerProvider;
    timer = Timer.getInstance();
    timer.initTimer(AbstractLineageProvider.class.getName());
  }
  
  protected Lineage split(T input) throws Exception {
  	if (!ReasonerConfig.useLineagePreprocessing) {
  		return removeRedundantLineage(input);
  	}
  	
  	if (input.varId2count.isEmpty()) {
  		return removeRedundantLineage(input);
  	}
    
  	boolean measureTime = false;
  	if (!lineageProvider.measureTime) {
  		// the split function is called recursively.
  		// So @lineageProvider.measureTime is used to track, that only the most outer call is measuring time.
  		measureTime = true;
  		lineageProvider.measureTime = true;
  		timer.startTimer(AbstractLineageProvider.class.getName());
  	}
  	
    MappedUnionFind groupedShannonVariables = groupShannonVariables(input);
    if (groupedShannonVariables.numSets() == 1) {
      // no splitting is possible
    	if (measureTime) {
      	timer.stopTimer(AbstractLineageProvider.class.getName());
      	lineageProvider.measureTime = false;
      }
    	if (input.getMaxShannon() < ReasonerConfig.maxShannonPerNode) {
    		return input;
    	} else {
    		return materializeShannon(input);
    	}
    }
    
    TIntHashSet[] childrenIds = getChildrenGroups(groupedShannonVariables, input);
    T splitted = getSplitLineage(childrenIds, input);
    
    if (measureTime) {
    	timer.stopTimer(AbstractLineageProvider.class.getName());
    	lineageProvider.measureTime = false;
    }
    
    handBack(input);
    return splitted;
  }
  
  /**
   * Computes sets of Shannon expansion variables.
   * Each set represents a disjoint group of children (of the lineage @input)
   * which entirely cover these Shannon expansions.
   * 
   * If there is more than one set, we can split @input into several lineages
   * to achieve the speed-up.
   * 
   * @param input : a lineage object with at least two Shannon expansions
   * @return
   */
  private MappedUnionFind groupShannonVariables(T input) {
    // fill the union find data structure 
    MappedUnionFind uf = new MappedUnionFind(input.varId2count.keys());
    for (Lineage child : input.subLineage) {
      TIntIterator setIt = child.vars.iterator();
      int toMerge = -1;
      for (int i = 0; i < child.vars.size(); i++) {
        int varId = setIt.next();
        if (!input.varId2count.containsKey(varId)) {
          // the variable is no Shannon variable
          continue;
        }
        if (toMerge == -1) {
          // it is the first Shannon variable of this child
          toMerge = varId;
          continue;
        }
        // more than one Shannon variable -> union their sets
        uf.union(toMerge, varId);
      }
    }
    return uf;
  }
  
  /**
   * computes sets of children of @input such that all children having variables
   * in the same set of @sets are in the same set.
   * 
   * @return sets of ids. The last set corresponds to the children without any
   * variable in @sets
   */
  private TIntHashSet[] getChildrenGroups(MappedUnionFind sets, T input) {
    // one set for each Shannon group and one set for all children without Shannon variables
    TIntHashSet[] childrenGroups = new TIntHashSet[sets.numSets() + 1];
    for (int i = 0; i < childrenGroups.length; i++) {
    	childrenGroups[i] = new TIntHashSet();
    }
    // maps variable ids (of shannon variables) to the above array
    TIntIntHashMap shannonId2group = new TIntIntHashMap();
    
    for (int childId = 0; childId < input.subLineage.length; childId++) {
      TIntSet vars = input.subLineage[childId].vars;
      TIntIterator it = vars.iterator();
      boolean wasAdded = false;
      for (int j = 0; j < vars.size() && !wasAdded; j++) {
      	int varId = it.next();
      	if (!input.varId2count.containsKey(varId)) {
      		// the variable is no Shannon variable
      		continue;
      	}
      	int shannonId = sets.find(varId);
      	if (!shannonId2group.containsKey(shannonId)) {
      		shannonId2group.put(shannonId, shannonId2group.size());
      	}
      	childrenGroups[shannonId2group.get(shannonId)].add(childId);
      	wasAdded = true;  // -> the loop breaks
      }
      if (!wasAdded) {
      	// no shannon variable in the child -> add to last set
      	childrenGroups[childrenGroups.length - 1].add(childId);
      }
    }
    return childrenGroups;
  }
  
  /**
   * creates a splitted up lineage as in the example on top.
   * 
   * 
   * @param childrenIds : sets of ids of children of @input. 
   * 											Each set will be the children of a child of @return
   * @param input : the original lineage to be split up 
   */
  private T getSplitLineage(TIntHashSet[] childrenIds, T input) throws Exception {
    LinkedList<Lineage> childrenTop = new LinkedList<Lineage>();
    for (int i = 0; i < childrenIds.length; i++) {
    	if (childrenIds[i].isEmpty()) {
    		continue;
    	}
    	Lineage[] childrenLow = new Lineage[childrenIds[i].size()];
    	TIntIterator it = childrenIds[i].iterator();
    	for (int j = 0; j < childrenIds[i].size(); j++) {
    		childrenLow[j] = input.subLineage[it.next()];
    	}
    	T child = getCleanObject();
    	child.setNewChildren(childrenLow);
    	if (child.varId2count.size() < ReasonerConfig.maxShannonPerNode) {
    		childrenTop.add(child);
    	} else {
    		childrenTop.add(materializeShannon(child));
    	}
    }
    T newOne = getCleanObject();
    newOne.setNewChildren(childrenTop.toArray(new Lineage[childrenTop.size()]));
    return newOne;
  }
  
  /**
   * Materializes one shannon expansion of the input.
   * Chooses the shannon variable occuring in the maximum number of children.
   * 
   * @param lineage : lineage to be materialized
   * @return disjointOr expansion of the input
   */
  private Lineage materializeShannon(T lineage) throws Exception {
  	// determine var id to branch on
  	int maxVarId = -1;
  	if (!lineage.varId2count.isEmpty()) {
	  	int maxNumChildren = -1;
	  	int size = lineage.varId2count.size();
	  	TIntIntIterator it = lineage.varId2count.iterator();
	  	for (int i = 0; i < size; i++) {
	  		it.advance();
	  		if (maxNumChildren < it.value()) {
	  			maxNumChildren = it.value();
	  			maxVarId = it.key();
	  		}
	  	}
  	} else {
  		maxVarId = lineage.vars.iterator().next();
  	}
  	
  	lineageArray.addLineage(lineage);  // so it will cleaned up
  	
  	// due to recursion, we need to get a clean instance here
  	ShannonMaterializer materializer = shannonMaterializerProvider.getMaterializer();
  	Lineage result = materializer.materializeVariable(maxVarId, lineage);
  	shannonMaterializerProvider.handBack(materializer);
  	return result;
  }

  /**
   * Will drop redundant lineage such as LineageAnd / LineageOr instances
   * with exactly one child.
   */
  private Lineage removeRedundantLineage(Lineage lineage) throws Exception {
  	if (lineage.subLineage.length == 1 && (lineage instanceof LineageOr || lineage instanceof LineageAnd)) {
  		Lineage toReturn = lineage.subLineage[0];
  		lineage.subLineage = new Lineage[0];  // otherwise the recursion of handing back will affect the children 
  		lineageProvider.handBack(lineage, false);
  		return toReturn;
  	} else {
  		return lineage;
  	}
  }
}
