package reasoner.lineage.providers;

import reasoner.ReasonerConfig;
import reasoner.lineage.Lineage;
import reasoner.lineage.LineageAnd;
import reasoner.lineage.LineageDisjointOr;
import reasoner.lineage.LineageFalse;
import reasoner.lineage.LineageLeaf;
import reasoner.lineage.LineageNot;
import reasoner.lineage.LineageOr;
import reasoner.lineage.LineageTrue;
import reasoner.lineage.LineageVisitor;
import reasoner.lineage.shannonmaterializer.ShannonMaterializerProvider;
import util.time.Timer;

/**
 * hides providers for `and', `not', `or', `leaf'
 * keeps constant objectes for `true', 'false'
 */
public class LineageProvider implements LineageVisitor {
  private static LineageProvider instance;
  private LineageOrProvider orProvider;
  private LineageDisjointOrProvider disjointOrProvider;
  private LineageAndProvider andProvider;
  private LineageNotProvider notProvider;
  private LineageLeafProvider leafProvider;
  // LineageTrue / LineageFalse are unique, they exist only once.
  // That is fine, since they are stateless.
  private LineageTrue lineageTrue;
  private LineageFalse lineageFalse;
  
  private ShannonMaterializerProvider shannonMaterializerProvider;
  
  private boolean handBackLeafs;
  
  // Used to measure run-time of preprocessing lineage in @AbstractLineageProvider.
  // @AbstractLineageProvider has recursive calls, so we store the variable here.
  public boolean measureTime;
  
  private LineageProvider() throws Exception {
  	shannonMaterializerProvider = new ShannonMaterializerProvider(this);
    orProvider = new LineageOrProvider(this, shannonMaterializerProvider);
    disjointOrProvider = new LineageDisjointOrProvider();
    notProvider = new LineageNotProvider();
    andProvider = new LineageAndProvider(this, shannonMaterializerProvider);
    leafProvider = new LineageLeafProvider();
    lineageTrue = new LineageTrue();
    lineageFalse = new LineageFalse();
    measureTime = false;
  }
  
  public static LineageProvider getInstance() throws Exception {
    if (instance == null) {
      instance = new LineageProvider();
    }
    return instance;
  }
  
  public Lineage getLineageOr(Lineage... children) throws Exception {
    return orProvider.get(children);
  }
  
  public Lineage getLineageAnd(Lineage... children) throws Exception {
    return andProvider.get(children);
  }
  
  public LineageNot getLineageNot(Lineage child) throws Exception {
    return notProvider.get(child);
  }
  
  public LineageLeaf getLineageLeaf(int varId, double conf) throws Exception {
    return leafProvider.get(varId, conf);
  }
  
  public Lineage getLineageConstant(boolean isTrue) {
  	if (isTrue) {
  		return lineageTrue;
  	} else {
  		return lineageFalse;
  	}
  }
  
  public LineageDisjointOr getLineageDisjointOr(Lineage childPos, Lineage childNeg, LineageLeaf var) throws Exception {
  	return disjointOrProvider.get(childPos, childNeg, var);
  }
  
  public boolean allObjectsReturned() {
    boolean allReturned = true;
    allReturned = orProvider.allObjectsReturned() & allReturned;
    allReturned = disjointOrProvider.allObjectsReturned() & allReturned;
    allReturned = andProvider.allObjectsReturned() & allReturned;
    allReturned = notProvider.allObjectsReturned() & allReturned;
    allReturned = leafProvider.allObjectsReturned() & allReturned;
    allReturned = shannonMaterializerProvider.allObjectsReturned() & allReturned;
    return allReturned;
  }
  
  public void clear() throws Exception {
  	measureTime = false;
  	Timer.getInstance().reset(AbstractLineageProvider.class.getName());
  }
  
  /**
   * Hands back lineage objects to the provider.
   * Implemented as a visitor.
   * 
   * @param toReturn : the lineage formula to return
   * @param handBackLeafs : if set, instances of @LineageLeaf are not returned. Useful for keeping existential relation.
   * @throws Exception
   */
  public void handBack(Lineage toReturn, boolean handBackLeafs) throws Exception {
    this.handBackLeafs = handBackLeafs;
    toReturn.acceptVisitor(this);
  }

	public void visit(LineageAnd and) throws Exception {
		if (!and.isInArray) {
			for (Lineage l : and.subLineage) {
				l.acceptVisitor(this);
			}
			andProvider.handBack(and);
		}
	}

	public void visit(LineageOr or) throws Exception {
		if (!or.isInArray) {
			for (Lineage l : or.subLineage) {
				l.acceptVisitor(this);
			}
			orProvider.handBack(or);
		}
	}
	
	public void visit(LineageDisjointOr disjointOr) throws Exception {
		if (!disjointOr.isInArray) {
			for (Lineage l : disjointOr.subLineage) {
				l.acceptVisitor(this);
			}
			disjointOrProvider.handBack(disjointOr);
		}
	}

	public void visit(LineageNot not) throws Exception {
		if (!not.isInArray) {
			not.subLineage[0].acceptVisitor(this);
			notProvider.handBack(not);
		}
	}

	public void visit(LineageLeaf leaf) throws Exception {
		if (handBackLeafs && !leaf.isInArray) {
			leafProvider.handBack(leaf);
		}
	}

	public void visit(LineageTrue lineage) throws Exception {
		if (ReasonerConfig.debugMode) {
			if (lineage != lineageTrue) {
				throw new Exception("Another lineageTrue was used!");
			}
		}
	}

	public void visit(LineageFalse lineage) throws Exception {
		if (ReasonerConfig.debugMode) {
			if (lineage != lineageFalse) {
				throw new Exception("Another lineageFalse was used!");
			}
		}
	}
	//// End handing back visitor ////
}
