package reasoner.lineage.shannonmaterializer;

import java.util.LinkedList;

import util.datastructures.Stack;
import util.provider.Providable;

import reasoner.ReasonerConfig;
import reasoner.lineage.Lineage;
import reasoner.lineage.LineageAnd;
import reasoner.lineage.LineageArray;
import reasoner.lineage.LineageDisjointOr;
import reasoner.lineage.LineageFalse;
import reasoner.lineage.LineageLeaf;
import reasoner.lineage.LineageNot;
import reasoner.lineage.LineageOr;
import reasoner.lineage.LineageTrue;
import reasoner.lineage.LineageVisitor;
import reasoner.lineage.providers.LineageProvider;

/**
 * Materializes a Shannon expansion, by changing the lineage
 * formula: 
 * phi = (var /\ phi_[var -> true]) \/ (~var /\ phi_[var -> false])
 * 
 * Be carefull, peforming this operation on every variable will
 * result in exponential run-time.
 * 
 * Implements a push-down automaton as a visitor to the lineage
 * data structure.
 * 
 * Reuses Lineage which was not changed.
 * Lineage, which was unnecesarily created is added to LineageArray,
 * so it will be handedBack correctly.
 */
public class ShannonMaterializer extends Providable implements LineageVisitor {
	
	private LineageProvider lineageProvider;
	private StackElementProvider stackElementProvider;
	private LineageArray lineageArray;
	
	private int varId;
	private Stack<StackElement> stack;
	private int depth;
	private boolean valueToSet;
	
	public ShannonMaterializer() throws Exception {
		stackElementProvider = new StackElementProvider(256);
		lineageArray = LineageArray.getInstance();
		stack = new Stack<StackElement>(256); 
	}
	
	public void setLineageProvider(LineageProvider lineageProvider) {
		this.lineageProvider = lineageProvider;
	}
	
	public Lineage materializeVariable(int varId, Lineage lineage) throws Exception {
		if (ReasonerConfig.debugMode) {
			if (lineageProvider == null) {
				throw new Exception("Set lineage provider before materializing shannon expansions");
			}
		}
		if (!lineage.vars.contains(varId)) {
			return lineage;
		}
		
		this.varId = varId;
		
		depth = 0;
		valueToSet = true;
		lineage.acceptVisitor(this);
		StackElement pos = stack.pop();
		if (ReasonerConfig.debugMode) {
			if (!stack.isEmpty()) {
				throw new Exception("Stack must be empty");
			}
			if (depth != 0) {
				throw new Exception("depth must 0");
			}
		}
		
		depth = 0;
		valueToSet = false;
		lineage.acceptVisitor(this);
		StackElement neg = stack.pop();
		if (ReasonerConfig.debugMode) {
			if (!stack.isEmpty()) {
				throw new Exception("Stack must be empty");
			}
			if (depth != 0) {
				throw new Exception("depth must 0");
			}
		}
		
		depth = -1;
		this.varId = -1;
		
		LineageLeaf leaf = (LineageLeaf)lineageArray.getLineage(varId);
		Lineage toReturn = lineageProvider.getLineageDisjointOr(pos.lineage, neg.lineage, leaf);
		
		stackElementProvider.handBack(pos);
		stackElementProvider.handBack(neg);

		if (ReasonerConfig.debugMode) {
			if (!stackElementProvider.allObjectsReturned()) {
				throw new Exception("Some stack elements were not returned");
			}
		}
		return toReturn;
	}

	@Override
	public void visit(LineageAnd and) throws Exception {
		handleAndOr(and, false);
	}

	@Override
	public void visit(LineageOr or) throws Exception {
		handleAndOr(or, true);
	}
	
	/**
	 * traverses the lineage for LineageAnd and LineageOr
	 * 
	 * @param lineage : the lineage to be updated
	 * @param fastEvalValue : true for LineageOr, false for LineageAnd
	 */
	private void handleAndOr(Lineage lineage, boolean fastEvalValue) throws Exception {
		depth += 1;
		StackElement myResult = null;
		for (int i = 0; i < lineage.subLineage.length; i++) {
			if (lineage.subLineage[i].vars.contains(varId)) {
				lineage.subLineage[i].acceptVisitor(this);
				if (ReasonerConfig.debugMode) {
					if (stack.peek().depth != depth) {
						throw new Exception("illegal stack depth");
					}
				}
				if (stack.peek().lineage instanceof LineageTrue
						|| stack.peek().lineage instanceof LineageFalse) {
					boolean isTrue = stack.peek().lineage instanceof LineageTrue;
					if (fastEvalValue == isTrue) {
						// stop processing! The result is @fastEvalValue
						while (!stack.isEmpty() && stack.peek().depth == depth) {
							// remove children already processed
							checkPeekLineage();
							stackElementProvider.handBack(stack.pop());
						}
						myResult = stackElementProvider.get(lineageProvider.getLineageConstant(fastEvalValue), depth - 1, true);
						break;
					} else {
						// we can drop the child
						stackElementProvider.handBack(stack.pop());
					}
				}
			} else {
				pushToStack(stackElementProvider.get(lineage.subLineage[i], depth, false));
			}
		}
		depth -= 1;
		
		LinkedList<Lineage> newChildren = new LinkedList<Lineage>();
		if (myResult == null) {	
			while (!stack.isEmpty() && stack.peek().depth == depth + 1) {
				StackElement child = stack.pop();
				newChildren.add(child.lineage);
				stackElementProvider.handBack(child);
			}
			if (newChildren.isEmpty()) {
				myResult = stackElementProvider.get(lineageProvider.getLineageConstant(false), depth, false);
			} else if (newChildren.size() == 1) {
				myResult = stackElementProvider.get(newChildren.getFirst(), depth, true);
			} else {
				Lineage newLineage = null;
				if (fastEvalValue) {
					newLineage = lineageProvider.getLineageOr(newChildren.toArray(new Lineage[newChildren.size()]));
				} else {
					newLineage = lineageProvider.getLineageAnd(newChildren.toArray(new Lineage[newChildren.size()]));
				}
				myResult = stackElementProvider.get(newLineage, depth, true);
			}
		}
		pushToStack(myResult);
	}

	@Override
	public void visit(LineageNot not) throws Exception {
		depth += 1;
		not.subLineage[0].acceptVisitor(this);
		depth -= 1;
		
		checkPeekLineage();
		StackElement stackElement = stack.pop();
		
		if (ReasonerConfig.debugMode) {
			if (stackElement.depth != depth + 1) {
				throw new Exception("Invalid depth in stack element");
			}
			if (stackElement.depth != depth + 1) {
				throw new Exception("Invalid depth in stack element");
			}
			if (!stack.isEmpty() && stack.peek().depth == depth + 1) {
				throw new Exception("More than one stack element for a not?");
			}
		}
		
		StackElement newStackElement = null;
		if (stackElement.lineage instanceof LineageTrue || stackElement.lineage instanceof LineageFalse) {
			boolean val = stackElement.lineage instanceof LineageFalse;
			newStackElement = stackElementProvider.get(lineageProvider.getLineageConstant(val), depth, false);
		} else {
			LineageNot lineage = lineageProvider.getLineageNot(stackElement.lineage);
			newStackElement = stackElementProvider.get(lineage, depth, true);
		}
		stackElementProvider.handBack(stackElement);
		pushToStack(newStackElement);
	}

	@Override
	public void visit(LineageLeaf leaf) throws Exception {
		if (ReasonerConfig.debugMode) {
			if (leaf.id != varId) {
				throw new Exception("Other variables should be reused by and / or");
			}
		}
		
		Lineage lineage = lineageProvider.getLineageConstant(valueToSet);
		StackElement stackElement = stackElementProvider.get(lineage, depth, false);
		pushToStack(stackElement);
	}

	public void visit(LineageTrue lineage) throws Exception {
		pushToStack(stackElementProvider.get(lineage, depth, false));
	}

	public void visit(LineageFalse lineage) throws Exception {
		pushToStack(stackElementProvider.get(lineage, depth, false));
	}

	public void visit(LineageDisjointOr disjointOr) throws Exception {
		if (varId == disjointOr.leaf.id) {
			// the disjoint or vanishes -> return the appropriate child
			if (valueToSet) {
				pushToStack(stackElementProvider.get(disjointOr.subLineage[0], depth, false));
			} else {
				pushToStack(stackElementProvider.get(disjointOr.subLineage[1], depth, false));
			}
		} else {
			// handle children
			depth += 1;
			for (int i = 0; i < 2; i++) {
				if (disjointOr.subLineage[i].vars.contains(varId)) {
					disjointOr.subLineage[i].acceptVisitor(this);
				} else {
					pushToStack(stackElementProvider.get(disjointOr.subLineage[i], depth, false));
				}
				if (stack.peek().depth != depth) {
					throw new Exception("invalid depth!");
				}
			}
			depth -= 1;
			
			// push new one on stack
			StackElement neg = stack.pop();
			StackElement pos = stack.pop();
			if (neg.lineage instanceof LineageFalse && pos.lineage instanceof LineageFalse) {
				// it evaluates to false!
				pushToStack(stackElementProvider.get(lineageProvider.getLineageConstant(false), depth, false));
			} else	if (neg.lineage instanceof LineageTrue && pos.lineage instanceof LineageTrue) {
				// it evaluates to true: (~var /\ true) \/ (var /\ true) = ~var \/ var = true
				pushToStack(stackElementProvider.get(lineageProvider.getLineageConstant(false), depth, true));
			} else	if (neg.lineage instanceof LineageFalse && pos.lineage instanceof LineageTrue) {
				pushToStack(stackElementProvider.get(disjointOr.leaf, depth, false));
			} else	if (neg.lineage instanceof LineageTrue && pos.lineage instanceof LineageFalse) {
				pushToStack(stackElementProvider.get(lineageProvider.getLineageNot(disjointOr.leaf), depth, true));
			} else if (pos.lineage instanceof LineageFalse) {
				pushToStack(stackElementProvider.get(lineageProvider.getLineageAnd(lineageProvider.getLineageNot(disjointOr.leaf), neg.lineage), depth, true));
				pos.lineage = null;
				neg.lineage = null;
			} else if (neg.lineage instanceof LineageFalse) {
				pushToStack(stackElementProvider.get(lineageProvider.getLineageAnd(disjointOr.leaf, pos.lineage), depth, true));
				pos.lineage = null;
				neg.lineage = null;
			} else {
				pushToStack(stackElementProvider.get(lineageProvider.getLineageDisjointOr(pos.lineage, neg.lineage, disjointOr.leaf), depth, true));
				pos.lineage = null;
				neg.lineage = null;
			}
			// handing back
			if (pos.lineage != null && pos.newLineage) {
				lineageArray.addLineage(pos.lineage);
			}
			stackElementProvider.handBack(pos);
			if (neg.lineage != null && neg.newLineage) {
				lineageArray.addLineage(neg.lineage);
			}
			stackElementProvider.handBack(neg);
		}
	}
	
	/**
	 * saves newly created lineage in the lineage array -> it will be cleaned up
	 */
	private void checkPeekLineage() throws Exception {
		if (stack.peek().newLineage) {
			lineageArray.addLineage(stack.peek().lineage);  // for garbage collection
		}
	}
	
	private void pushToStack(StackElement element) throws Exception {
		if (ReasonerConfig.debugMode) {
			if (element.lineage.vars.contains(varId)) {
				throw new Exception("Stack may not contain varId!");
			}
		}
		stack.push(element);
	}
	
	public void clear() throws Exception {
		if (ReasonerConfig.debugMode) {
			if (depth != -1 || varId != -1 || !stack.isEmpty()) {
				throw new Exception("ShannonMaterializer was not cleared");
			}
		}
	}
}
