package reasoner.query;

import java.util.Collection;
import java.util.LinkedList;

import reasoner.ReasonerConfig;
import reasoner.grounding.Tuple;
import reasoner.grounding.Grounder;
import reasoner.grounding.IntervalLogic;
import reasoner.grounding.IntervalProvider;
import reasoner.io.DatabaseManager;
import reasoner.io.DatasetNameParser;
import reasoner.io.QueryParser;
import reasoner.lineage.Lineage;
import reasoner.lineage.LineageArray;
import reasoner.lineage.confidence.ConfidenceCalculator;
import reasoner.lineage.confidence.TupleIndependentConfidence;
import reasoner.lineage.providers.AbstractLineageProvider;
import reasoner.query.compiled.CompiledDeductionRule;
import reasoner.query.compiled.Query;
import reasoner.query.compiled.QueryStep;
import util.io.Files;
import util.time.Timer;

/**
 * Manages executing a query.
 * 
 * Fills constants into a query.
 * check via hasNextQuery()
 * and then prepareNextQuery().
 * 
 * provides methods to:
 *  + ground a query
 *  + ground constraints
 *  + calculate the results' confidences
 */
public class QueryExecuter {
	public Grounder grounder;
	public Query query;
	
	private int constantId;  // id in query.constantsToSet
	private TupleIndependentConfidence confidenceCalculator;
	private Timer timer;
	
	public QueryExecuter() throws Exception {
		confidenceCalculator = new TupleIndependentConfidence();
		timer = Timer.getInstance();
	}
	
	public void initializeSettings() throws Exception {
		// clean up db
		DatabaseManager manager = DatabaseManager.getInstance();
		manager.clearDBall();
		
		// load dataset
		String path = QuerySettings.pathToFolder;
		String datasetName = DatasetNameParser.getName(path + "/dataset.txt");
		QuerySettings.dataset = manager.getDataset(datasetName);
		
		// parse query
    QueryParser queryParser = new QueryParser(QuerySettings.dataset);
    query = queryParser.parseQueryPlan(path + "/deductionRules.txt"); 
    
    // parse constants
    if (Files.fileExists(path + "/constants.txt")) {
    	queryParser.parseConstants(path + "/constants.txt", query);
    } else {
    	query.constantsToSet.add("EMPTY");
    }
    
    // parse output relations
    queryParser.parseOutputRelations(path + "/queryRelations.txt", query);
    
    // parse constraints
    if (QuerySettings.useConstraints) {
    	if (!Files.fileExists(path + "/constraints.txt")) {
    		throw new Exception("Constraint strategy was set, but constraints.txt does not exist.");
    	}
    	queryParser.parseConstraints(path + "/constraints.txt", query);
    	query.useConstraints = QuerySettings.useConstraints;
    }
    
		// create necessary objects
		grounder = new Grounder();
		
		constantId = 0;
	}
	
	/**
	 * Injects the next constant into the query plan.
	 * Returns false, if no such query exists.
	 */
	public boolean hasNextQuery() throws Exception {
		if (constantId >= query.constantsToSet.size()) {
			return false;
		}
		String nextConstant = query.constantsToSet.get(constantId);
		if (!nextConstant.equals("EMPTY")) {
			int numConstantsSet = 0;
			for (QueryStep step : query.steps) {
				for (CompiledDeductionRule rule : step.rules) {
					if (rule.requiresConstant) {
						numConstantsSet++;
						rule.stmt.setString(1, nextConstant);
					}
				}
			}
			if (numConstantsSet == 0) {
				throw new Exception("Tried to set constant: " + nextConstant + ", but query plan did not allow setting constants");
			}
			if (ReasonerConfig.debugMode) {
				System.out.println("Constant " + nextConstant + " was set " + numConstantsSet + " times.");
			}
		}
		constantId++;
		return true;
	}
	
	public Results groundQuery() throws Exception {
		if (query.dirty) {
			throw new Exception("Clean query before executing again");
		}
		
		query.dirty = true;
		Results results = new Results();
		
	  // rule grounding
    timer.startTimer(Grounder.class.getName());
    grounder.queryScheduler(query);
    results.ruleGroundingTime = timer.elapsedAndRemoveTimer(Grounder.class.getName());
    System.out.println("Grounding time: " + results.ruleGroundingTime + " ms");
    
    //  ground constraints
    results.constraintLineage = null;
    if (query.useConstraints) {
    	timer.startTimer(Grounder.class.getName());
    	results.constraintLineage = grounder.groundConstraints(query.constraints);
    	results.constraintGroundingTime = timer.elapsedAndRemoveTimer(Grounder.class.getName());
    	System.out.println("Constraints grounding time: " + results.constraintGroundingTime + " ms");
    	results.numConstraintClauses = grounder.numberConstraintClauses;
    	if (results.constraintLineage != null) {
    		System.out.println("Max #Shannon Expansion Variables: " + results.constraintLineage.getMaxShannon());
    	}
    }
    
    results.intervalCompTime = timer.getElapsedTime(IntervalLogic.class.getName());
    System.out.println("Interval Computation time: " + results.intervalCompTime + " ms");
    results.preprocessingTime = timer.getElapsedTime(AbstractLineageProvider.class.getName());
    System.out.println("Preprocessing time: " + results.preprocessingTime + " ms");
    
    // gather result tuples
    results.resultTuples = grounder.output(query, results.constraintLineage);
	  return results;
	}
	
	public void computeConfidences(Results results) throws Exception {
		double constraintConfidence = 1.0;
		if (results.constraintLineage != null) {
			timer.startTimer(ConfidenceCalculator.class.getName());
			constraintConfidence = confidenceCalculator.confidence(results.constraintLineage);
			results.confidenceComputationTime = timer.elapsedAndRemoveTimer(ConfidenceCalculator.class.getName());
			System.out.println("Constraints confidence time: " + results.confidenceComputationTime + " ms");
		}
		LineageArray lineageArray = LineageArray.getInstance();
		timer.startTimer(TupleIndependentConfidence.class.getName());
		int i = 0;
		for (Tuple tuple : results.resultTuples){
			if (ReasonerConfig.debugMode){
				if (i % 1000 == 0) {
					System.out.println("Tuple confidence calculation " + i);
				}
			}
      tuple.calcConfExact(lineageArray, confidenceCalculator, constraintConfidence);
      i++;
    }
		results.confidenceComputationTime += timer.elapsedAndRemoveTimer(TupleIndependentConfidence.class.getName());
		System.out.println("Total confidence time: " + results.confidenceComputationTime + " ms");
	}
	
	public void clearQuery(Results results) throws Exception {
		grounder.clear();
		returnIntervals(results.resultTuples);
    query.dataset.clear();
    DatabaseManager.getInstance().clearDBintensional();
    confidenceCalculator.clear();
    query.dirty = false;
    System.gc();
	}
	
	public void clean() throws Exception {
		ReasonerConfig.clearAll();
	}
	
	private void returnIntervals(Collection<Tuple> results) throws Exception {
	  for (Tuple fact : results){
      IntervalProvider.getInstance().handBack(fact.intervals);
    }
	}
	
	public String getCurrentConstant() {
		return query.constantsToSet.get(constantId - 1);
	}
	
	public class Results {
		public long ruleGroundingTime;
		public long constraintGroundingTime;
		public long preprocessingTime;
		public long intervalCompTime;
		public long confidenceComputationTime;
		public int numConstraintClauses;
		public LinkedList<Tuple> resultTuples;
		public Lineage constraintLineage;
	}
}
