package reasoner.query.compiled;

import java.sql.PreparedStatement;
import java.sql.ResultSet;

import reasoner.grounding.Interval;
import reasoner.grounding.IntervalProvider;
import reasoner.query.parsed.RelationalLiteral;

public class CompiledDeductionRule {
	/*
	 * SELECT headVar1, headVar2, ... 
	 *        tX.time_begin, tX.time_end, ...
	 *        t0.array_id, t1.array_id, ... , tn.array_id, ... // positive array_ids first, then negative ones
	 * FROM relationName AS t0 JOIN relationName2 AS t1 ... LEFT OUTER JOIN negatedRelation1 AS tn ...
	 * WHERE tX.arg='Constant' ...
	 *       AND tX.time_end < tY.time_begin ...
	 */ 
	public PreparedStatement stmt;
	// @positiveLiteralStmt needs a constant to be set 
	public final boolean requiresConstant;
	// one entry per argument
	public final HeadArgument[] headArguments;
	public final TemporalHeadArgument[] temporalHeadArguments;
	// array_id's necessary for a rule body as ids in rset
	public final int[] positiveArrayIdsInRset;
	// array_ids in negative literal. Result might be null.
	// One arrayId entry per negated literal in the body.
	public final int[] negativeArrayIdsInRset;
	
	public CompiledDeductionRule(RelationalLiteral headLiteral, int numPosLiterals, int numNegativeLiterals, boolean requiresConstant) {
		headArguments = new HeadArgument[headLiteral.relation.numNonTempArgs];
		temporalHeadArguments = new TemporalHeadArgument[2];
		positiveArrayIdsInRset = new int[numPosLiterals];
		negativeArrayIdsInRset = new int[numNegativeLiterals];
		this.requiresConstant = requiresConstant;
	}
	
	/**
	 * Creates a string representing the head's variables' bindings
	 * at the current result set scan position.
	 * The used seperator is '\t'.
	 */
	public String getNonTemporalArguments(ResultSet rset) throws Exception {
		StringBuffer buffer = new StringBuffer();
		for (int i = 0; i < headArguments.length; i++){
			String argToAppend = headArguments[i].getArgument(rset);
			buffer.append(argToAppend);
			buffer.append("\t");
		}
		return buffer.toString();
	}
	
	/**
	 * Determine the interval limits as defined in the rule
	 * as of the current scan position of the result set.
	 * 
	 * @return the deduced interval, or null if there is none (end <= begin)
	 */
	public Interval getDeducedInterval(ResultSet rset, IntervalProvider intervalProvider) throws Exception {
		int begin = temporalHeadArguments[0].getTimePoint(rset);
		int end = temporalHeadArguments[1].getTimePoint(rset);
		
    if (begin < end) {
    	return intervalProvider.get(begin, end, -1);
    } else {
    	return null;
    }
	}
}
