package reasoner.query.parsed;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import reasoner.io.dataset.Dataset;
import reasoner.query.parsed.JoinOrdering.Join;

public class ParsedConjunction {
	public final List<RelationalLiteral> positiveLiterals;
	public final List<RelationalLiteral> negativeLiterals;
	public final List<ArithmeticLiteral> arithmeticLiterals;
	
	public ParsedConjunction() {
		positiveLiterals = new LinkedList<RelationalLiteral>();
		negativeLiterals = new LinkedList<RelationalLiteral>();
		arithmeticLiterals = new LinkedList<ArithmeticLiteral>();
	}
	
	public Map<Variable, String> compileConjunction(HashMap<RelationalLiteral, String> literal2tableName, StringBuffer from, StringBuffer where) throws Exception {
		HashMap<Variable, String> seenVar2name = new HashMap<Variable, String>();
		
		// from, where clauses for joining
		JoinOrdering joinOrdering = getJoinOrdering();
		int tableId = 0;
		for (Join join : joinOrdering.order) {
			if (tableId > 0) {
				if (join.variables.isEmpty()) {
					from.append(", ");
				} else {
					from.append("JOIN ");
				}
			}
			
			// table name
			from.append(join.literal.relation.name);
			from.append(" AS ");
			String tableName = "t" + tableId;
			from.append(tableName);
			from.append(" ");
			literal2tableName.put(join.literal, tableName);
			
			// join condition
			Map<Variable, String> varNames = join.literal.getVariableNames(tableName);
			if (!join.variables.isEmpty()) {
				if (tableId > 0) {
					from.append("ON ");
					int varNum = 0;
					for (Variable var : join.variables) {
						if (varNum > 0) {
							from.append("AND ");
						}
						from.append(seenVar2name.get(var));
						from.append("=");
						from.append(varNames.get(var));
						from.append(" ");
						varNum += 1;
					}
				}
			}
			
			// update seen variables
			seenVar2name.putAll(varNames);  // old entries can be overwritten
			
			// where clause
			updateVariableConditions(join.literal, where, tableName);
			
			tableId += 1;
		}
		
		// from, where clause: negated
		for (RelationalLiteral l : negativeLiterals) {
			String tableName = "t" + tableId;
			literal2tableName.put(l, tableName);
			
			// from clause
			from.append("LEFT OUTER JOIN ");
			from.append(l.relation.name);
			from.append(" AS ");
			from.append(tableName);
			
			// join condition
			from.append(" ON ");
			int varId = 0;
			Map<Variable, String> varNames = l.getVariableNames(tableName);
			for (Entry<Variable, String> entry : varNames.entrySet()) {
				if (varId > 0) {
					from.append("AND ");
				}
				from.append(seenVar2name.get(entry.getKey()));
				from.append("=");
				from.append(entry.getValue());
				from.append(" ");
				varId += 1;
			}
			
			// where clause
			updateVariableConditions(l, from, tableName);
			
			tableId += 1;
		}
		
		// where clause: arithmetic
		for (ArithmeticLiteral l : arithmeticLiterals) {
			if (where.length() > 0) {
				where.append("AND ");
			}
			if (l.negated) {
				where.append("NOT (");
			}
			if (l.arguments[0] instanceof Variable) {
				where.append(seenVar2name.get(l.arguments[0]));
			} else {
				where.append(escapeConstant((Constant)l.arguments[0]));
			}
			where.append(l.operator);
			if (l.arguments[1] instanceof Variable) {
				where.append(seenVar2name.get(l.arguments[1]));
			} else {
				where.append(escapeConstant((Constant)l.arguments[1]));
			}
			if (l.negated) {
				where.append(") ");
			} else {
				where.append(" ");
			}
		}
		
		return seenVar2name;
	}
	
	private void updateVariableConditions(RelationalLiteral literal, StringBuffer sql, String tableName) throws Exception {
		String condition = literal.getConstantConditions(tableName);
		if (condition.length() > 0) {
			if (sql.length() > 0) {
				sql.append("AND ");
			}
			sql.append(condition);
			sql.append(" ");
		}
		
		condition = literal.getRepeatedVariables(tableName);
		if (condition.length() > 0) {
			if (sql.length() > 0) {
				sql.append("AND ");
			}
			sql.append(condition);
		}
	}
	
	private String escapeConstant(Constant constant) {
		if (constant.toSet) {
			return "?";
		} else if (constant.isTemporal) {
			return constant.name;
		} else {
			return "\"" + constant.name + "\"";
		}
	}
	
	/**
	 * Computes a join ordering for the positive literals.
	 */
	private JoinOrdering getJoinOrdering() throws Exception {
		JoinOrdering ordering = new JoinOrdering();
		List<RelationalLiteral> remainingLiterals = new LinkedList<RelationalLiteral>();
		remainingLiterals.addAll(positiveLiterals);
		HashSet<Variable> boundVariables = new HashSet<Variable>();
		// first one is arbitrary: first element in list
		RelationalLiteral firstLiteral = remainingLiterals.get(0);
		ordering.add(firstLiteral, new HashSet<Variable>());
		boundVariables.addAll(firstLiteral.getVariables());
		remainingLiterals.remove(firstLiteral);
		
		while (!remainingLiterals.isEmpty()) {
			RelationalLiteral nextLiteral = getNextJoinLiteral(boundVariables, remainingLiterals);
			remainingLiterals.remove(nextLiteral);
			
			Set<Variable> variables = nextLiteral.getVariables();
			variables.retainAll(boundVariables);
			ordering.add(nextLiteral, variables);
			
			variables = nextLiteral.getVariables();
			boundVariables.addAll(variables);
		}
		return ordering;
	}
	
	/**
	 * Determines the literal with maximum number of variables to be joined.
	 */
	private RelationalLiteral getNextJoinLiteral(HashSet<Variable> boundVariables, List<RelationalLiteral> remainingLiterals) throws Exception {
		if (remainingLiterals.isEmpty()) {
			throw new Exception("Cannot find a literal, if remaining literals are empty");
		}
		int bestOverlap = -1;
		RelationalLiteral bestLiteral = null;
		for (RelationalLiteral l : remainingLiterals) {
			Set<Variable> variables = l.getVariables();
			variables.retainAll(boundVariables);
			if (variables.size() > bestOverlap) {
				bestOverlap = variables.size();
				bestLiteral = l;
			}
		}
		if (bestOverlap == 0) {
			System.err.println("No variables to join on. Did you use a cartesian product? This might be slow ... " + this);
		}
		return bestLiteral;
	}
	
	/**
	 * Invokes the consitency checks.
	 */
	public void consistencyCheck() throws Exception {
		for (RelationalLiteral l : positiveLiterals) {
			l.consistencyCheck();
		}
		for (RelationalLiteral l : negativeLiterals) {
			l.consistencyCheck();
		}
		for (ArithmeticLiteral l : arithmeticLiterals) {
			l.consistencyCheck();
		}
	}
	
	/**
	 * Links the relation names to objects
	 * stored in the dataset class.
	 * 
	 * @return true on success
	 */
	public boolean inferRelations(Dataset dataset) throws Exception {
		// positive literals in body
		// collect all positive variables
		HashMap<String, Variable> variables = new HashMap<String, Variable>();
		for (RelationalLiteral l : positiveLiterals) {
			if (l.relation == null) {
				if (dataset.relations.containsKey(l.relationName)) {
					l.relation = dataset.relations.get(l.relationName);
					l.relationName = null;
				} else {
					if (l.relationName.startsWith(dataset.name)) {
						throw new Exception("Misspelling of extensional relation? " + l.relationName);
					}
					return false;  // infering does not terminate
				}
			}
			if (l.relation.isTemporal) {
				makeArgumentsTemporal(l);
			}
			for (AbstractArgument arg : l.arguments) {
				if (arg instanceof Variable) {
					variables.put(arg.name, (Variable)arg);
				}
			}
		}
		
		// negative literals in body
		for (RelationalLiteral l : negativeLiterals) {
			if (l.relation == null) {
				if (!dataset.relations.containsKey(l.relationName)) {
					return false;  // infering does not terminate
				}
				l.relation = dataset.relations.get(l.relationName);
				l.relationName = null;
			}
		}
		
		// arithmetic literals
		for (ArithmeticLiteral l : arithmeticLiterals) {
			if (l.arguments[0].isTemporal
					|| l.arguments[1].isTemporal
					|| l.operator != ArithmeticLiteral.Operator.EQUALS) {
				l.arguments[0].isTemporal = true;
				l.arguments[1].isTemporal = true;
			}
		}
		
		return true;
	}
	
	/**
	 * Sets the last two arguments to be temporal.
	 */
	private void makeArgumentsTemporal(RelationalLiteral l) throws Exception {
		int numArgs = l.arguments.size();
		if (numArgs < 3) {
			throw new Exception("Too few arguments in the literal: " + l);
		}
		l.arguments.get(numArgs - 1).isTemporal = true;
		l.arguments.get(numArgs - 2).isTemporal = true;
	}
	
	@Override
	public String toString() {
		StringBuffer buffer = new StringBuffer();
		for (RelationalLiteral l : positiveLiterals) {
			if (buffer.length() > 0) {
				buffer.append(" & ");
			}
			buffer.append(l);
		}
		for (RelationalLiteral l : negativeLiterals) {
			if (buffer.length() > 0) {
				buffer.append(" & ");
			}
			buffer.append("~");
			buffer.append(l);
		}
		for (ArithmeticLiteral l : arithmeticLiterals) {
			if (buffer.length() > 0) {
				buffer.append(" & ");
			}
			buffer.append(l);
		}
		return buffer.toString();
	}
}
