package reasoner.query.parsed;

import java.sql.Connection;
import java.util.HashMap;
import java.util.Map;

import reasoner.io.DatabaseManager;
import reasoner.io.Time;
import reasoner.io.dataset.Dataset;
import reasoner.io.dataset.Relation;
import reasoner.query.compiled.CompiledDeductionRule;
import reasoner.query.compiled.HeadArgument;
import reasoner.query.compiled.TemporalHeadArgument;

public class ParsedDeductionRule {
	public RelationalLiteral head;
	public ParsedConjunction body;
	
	/**
	 * Compiles the parsed deduction rule to SQL.
	 * Stores additional information in @return regarding the result set.
	 */
	public CompiledDeductionRule compileRule() throws Exception {
		CompiledDeductionRule compiledRule = initCompiledDeductionRule();
		
		StringBuffer from = new StringBuffer();
		StringBuffer where = new StringBuffer();
		HashMap<RelationalLiteral, String> literal2tableName = new HashMap<RelationalLiteral, String>();
		Map<Variable, String> seenVar2name = body.compileConjunction(literal2tableName, from, where);
		
		// select clause: nonTemporal arguments
		StringBuffer select = new StringBuffer();
		int argId = 0;
		int rsetId = 1;
		for (AbstractArgument arg : head.arguments) {
			if (arg.isTemporal) {
				break;
			}
			if (arg instanceof Variable) {
				compiledRule.headArguments[argId] = new HeadArgument(rsetId);
				select.append(seenVar2name.get(arg));
				select.append(", ");
				rsetId += 1;
			} else {
				// constant
				compiledRule.headArguments[argId] = new HeadArgument(arg.name);
			}
			argId += 1;
		}
		
		// select clause: temporal arguments
		Time time = Time.getInstance();
		if (head.relation.isTemporal) {
			int tempArgId = head.arguments.size() - 2;
			for (int i = 0; i < 2; i++) {
				AbstractArgument arg = head.arguments.get(tempArgId + i);
				if (arg instanceof Variable) {
					compiledRule.temporalHeadArguments[i] = new TemporalHeadArgument(-1, rsetId);
					select.append(seenVar2name.get(arg));
					select.append(", ");
					rsetId += 1;
				} else {
					// it is a constant
					String tempConstant = arg.name;
					int[] timeInterval = time.string2int(tempConstant);
					if (tempConstant.equals("time_end")) {
						compiledRule.temporalHeadArguments[i] = new TemporalHeadArgument(timeInterval[1], -1);
					} else {
						compiledRule.temporalHeadArguments[i] = new TemporalHeadArgument(timeInterval[0], -1);
					}
				}
			}
		} else {
			// no temporal argument
			compiledRule.temporalHeadArguments[0] = new TemporalHeadArgument(time.minTime, -1);
			compiledRule.temporalHeadArguments[1] = new TemporalHeadArgument(time.maxTime, -1);
		}
		
		// select clause: arrayIds
		int literalId = 0;
		for (RelationalLiteral l : body.positiveLiterals) {
			if (!l.relation.isUncertain) {
				continue;
			}
			if (literalId > 0) {
				select.append(", ");
			}
			select.append(literal2tableName.get(l));
			select.append(".array_id");
			compiledRule.positiveArrayIdsInRset[literalId] = rsetId;
			rsetId += 1;
			literalId += 1;
		}
		
		literalId = 0;
		for (RelationalLiteral l : body.negativeLiterals) {
			if (!l.relation.isUncertain) {
				continue;
			}
			select.append(", ");
			select.append(literal2tableName.get(l));
			select.append(".array_id");
			compiledRule.negativeArrayIdsInRset[literalId] = rsetId;
			literalId += 1;
			rsetId += 1;
		}
		
		// compile SQL stmt
		StringBuffer sql = new StringBuffer();
		sql.append("SELECT ");
		sql.append(select);
		sql.append(" FROM ");
		sql.append(from);
		if (where.length() > 0) {
			sql.append("WHERE ");
			sql.append(where);
		}
		Connection con = DatabaseManager.getInstance().getConnection();
		compiledRule.stmt = con.prepareStatement(sql.toString());
		
		return compiledRule;
	}
	
	private CompiledDeductionRule initCompiledDeductionRule() throws Exception {
		int requiresConstant = 0;
		int numPositive = 0;
		for (RelationalLiteral l : body.positiveLiterals) {
			if (l.relation.isUncertain) {
				numPositive += 1;
			}
			for (AbstractArgument arg : l.arguments) {
				if (arg instanceof Constant
						&& ((Constant)arg).toSet) {
					requiresConstant += 1;
				}
			}
		}
		if (numPositive == 0) {
			throw new Exception("There must be at least one uncertain positive literal in the deduction rule body");
		}
		
		int numNegative = 0;
		for (RelationalLiteral l : body.negativeLiterals) {
			if (l.relation.isUncertain) {
				numNegative += 1;
			}
			for (AbstractArgument arg : l.arguments) {
				if (arg instanceof Constant
						&& ((Constant)arg).toSet) {
					requiresConstant += 1;
				}
			}
		}
		
		for (ArithmeticLiteral l : body.arithmeticLiterals) {
			for (int i = 0; i < 2; i++) {
				if (l.arguments[i] instanceof Constant
						&& ((Constant)l.arguments[i]).toSet) {
					requiresConstant += 1;
				} 
			}
		}
		
		if (requiresConstant > 1) {
			throw new Exception("You can set at most one constant in a deduction rule.");
		}
		
		return new CompiledDeductionRule(head, numPositive, numNegative, requiresConstant == 1);
	}
	
	/**
	 * Infers the relations from the relation names of the literals.
	 * Performs a consistency check on each literal.
	 * 
	 * @return true, if all relations are processed.
	 * @throws Exception : a consistency check failed.
	 */
	public boolean updateRelations(Dataset dataset) throws Exception {
		if (head.relation == null) {
			inferRelations(dataset);
		}
		if (head.relation != null) {
			consistencyCheck();
			return true;
		} else {
			return false;
		}
	}
	
	/**
	 * Invokes the consitency checks.
	 */
	private void consistencyCheck() throws Exception {
		head.consistencyCheck();
		body.consistencyCheck();
	}
	
	/**
	 * Infers the relations from the rule body.
	 * Sets arguments to be temporal, if the relation
	 * is temporal.
	 * 
	 * If head.relation == null after this method terminates,
	 * then the inference did not succeed. 
	 */
	private void inferRelations(Dataset dataset) throws Exception {
		boolean success = body.inferRelations(dataset);
		if (!success) {
			return;
		}
		
		// head literal
		if (dataset.relations.containsKey(head.relationName)) {
			head.relation = dataset.relations.get(head.relationName);
		} else {
			// we need to infer it!
			int numArgs = head.arguments.size();
			boolean isTemporal = head.arguments.get(numArgs - 1).isTemporal && head.arguments.get(numArgs - 2).isTemporal;
			if (isTemporal) {
				numArgs -= 2;
			}
			Connection con = DatabaseManager.getInstance().getConnection();
			Relation relation = dataset.createRelation(head.relationName, numArgs, isTemporal, false, true, con);
			head.relation = relation;
		}
		head.relationName = null;
	}
	
	@Override
	public String toString() {
		StringBuffer buffer = new StringBuffer();
		if (head != null) {
			buffer.append(head);
		}
		buffer.append(" <- ");
		buffer.append(body);
		return buffer.toString();
	}
}
