package reasoner.query.parsed;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import reasoner.io.dataset.Dataset;
import reasoner.io.dataset.Relation;
import reasoner.query.compiled.CompiledDeductionRule;
import reasoner.query.compiled.Query;
import reasoner.query.compiled.QueryStep;

public class ParsedQuery {
	public List<ParsedDeductionRule> rules;
	
	public ParsedQuery() {
		rules = new LinkedList<ParsedDeductionRule>();
	}
	
	public Query compileQuery(Dataset dataset) throws Exception {
		inferRelations(dataset);
		Map<Relation, List<ParsedDeductionRule>> groupedRules = groupRulesByHeadRelation();
		Query query = determineQueryPlan(groupedRules, dataset);
		return query;
	}
	
	/**
	 * Determines the relations in `dataset.relations'
	 * from the parsed deduction rules. More specifically
	 * from the head literals of the deduction rules.
	 */
	private void inferRelations(Dataset dataset) throws Exception {
		List<ParsedDeductionRule> toProcess = new LinkedList<ParsedDeductionRule>();
		toProcess.addAll(rules);
		
		int numIterations = 0;
		while (!toProcess.isEmpty()) {
			List<ParsedDeductionRule> remainingRules = new LinkedList<ParsedDeductionRule>();
			for (ParsedDeductionRule rule : toProcess) {
				if (!rule.updateRelations(dataset)) {
					remainingRules.add(rule);
				}
			}
			toProcess = remainingRules;
			numIterations += 1;
			if (numIterations == 1000) {
				throw new Exception("Could not resolve the relations of some body literals.");
			}
		}
	}
	
	/**
	 * Gathers all parsed deduction rules with the
	 * same relation in the head literal.
	 */
	private Map<Relation, List<ParsedDeductionRule>> groupRulesByHeadRelation() {
		HashMap<Relation, List<ParsedDeductionRule>> map = new HashMap<Relation, List<ParsedDeductionRule>>();
		for (ParsedDeductionRule rule : rules) {
			Relation headRelation = rule.head.relation;
			if (!map.containsKey(headRelation)) {
				map.put(headRelation, new LinkedList<ParsedDeductionRule>());
			}
			map.get(headRelation).add(rule);
		}
		return map;
	}
	
	/**
	 * Determines the order of executing
	 * the deduction rules and compiles
	 * the rules to a query.
	 */
	private Query determineQueryPlan(Map<Relation, List<ParsedDeductionRule>> groupedRules, Dataset dataset) throws Exception {
		List<QueryStep> querySteps = new LinkedList<QueryStep>();
		
		Set<Relation> definedRelations = new HashSet<Relation>();
		// gather all extensional relations
		for (Relation relation : dataset.relations.values()) {
			if (relation.isExtensional) {
				definedRelations.add(relation);
			}
		}
		
		int numIterations = 0;
		while (!groupedRules.isEmpty()) {
			Map<Relation, List<ParsedDeductionRule>> newGroupedRules = new HashMap<Relation, List<ParsedDeductionRule>>();
			for (Entry<Relation, List<ParsedDeductionRule>> step : groupedRules.entrySet()) {
				boolean allDefined = true;
				// check all relation of all deduction rules
				for (ParsedDeductionRule rule : step.getValue()) {
					for (RelationalLiteral l : rule.body.positiveLiterals) {
						if (!definedRelations.contains(l.relation)) {
							allDefined = false;
							break;
						}
					}
					if (!allDefined) {
						break;
					}
					for (RelationalLiteral l : rule.body.negativeLiterals) {
						if (!definedRelations.contains(l.relation)) {
							allDefined = false;
							break;
						}
					}
					if (!allDefined) {
						break;
					}
				}
				
				if (allDefined) {
					List<CompiledDeductionRule> compiledRules = new LinkedList<CompiledDeductionRule>();
					for (ParsedDeductionRule rule : step.getValue()) {
						compiledRules.add(rule.compileRule());
					}
					QueryStep queryStep = new QueryStep(step.getKey(), compiledRules);
					querySteps.add(queryStep);
					definedRelations.add(step.getKey());
				} else {
					newGroupedRules.put(step.getKey(), step.getValue());
				}
			}
			groupedRules = newGroupedRules;
			numIterations += 1;
			if (numIterations > 1000) {
				throw new Exception("Infinite loop while determining order of deduction rules?");
			}
		}
		
		if (querySteps.isEmpty()) {
			throw new Exception("No query steps? Empty query?");
		}
		Query query = new Query(dataset, querySteps);
		return query;
	}
}
