package reasoner.query.parsed;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import reasoner.io.Time;
import reasoner.io.dataset.Relation;

public class RelationalLiteral {
	
	public Relation relation;
	public String relationName;  // used for intensional relations at first
	public List<AbstractArgument> arguments;
	
	public RelationalLiteral() {
		arguments = new LinkedList<AbstractArgument>();
	}
	
	public void consistencyCheck() throws Exception {
		if (relation == null) {
			throw new Exception("The relation was null");
		}
		if (relationName != null) {
			throw new Exception("The relation name was not cleared");
		}
		if (relation.isTemporal) {
			if (arguments.size() < 3) {
				throw new Exception("Too few arguments in literal: " + this);
			}
			for (int i = 1; i < 3; i++) {
				AbstractArgument arg = arguments.get(arguments.size() - i);
				if (!arg.isTemporal) {
					throw new Exception("Non-Temporal argument: " + arg + " but the relation " + relation.name + " is temporal");
				}
			}
			if (arguments.size() != relation.numNonTempArgs + 2) {
				throw new Exception("Wrong number of arguments in relation: " + this);
			}
			for (int i = 0; i < arguments.size() - 2; i++) {
				if (arguments.get(i).isTemporal) {
					throw new Exception("Only the last two arguments should be temporal: " + this);
				}
			}
		} else {
			if (arguments.size() != relation.numNonTempArgs) {
				throw new Exception("Wrong number of arguments in relation: " + this);
			}
			for (int i = 0; i < arguments.size(); i++) {
				if (arguments.get(i).isTemporal) {
					throw new Exception("No arguments should be temporal: " + this);
				}
			}
		}
	}
	
	public Set<Variable> getVariables() {
		Set<Variable> variables = new HashSet<Variable>();
		for (AbstractArgument arg : arguments) {
			if (arg instanceof Variable) {
				variables.add((Variable)arg);
			}
		}
		return variables;
	}
	
	/**
	 * Gathers all variables and their name as refered to
	 * by `tableName' in a SQL statement.
	 */
	public Map<Variable, String> getVariableNames(String tableName) {
		HashMap<Variable, String> map = new HashMap<Variable, String>();
		int i = 0;
		for (AbstractArgument arg : arguments) {
			if (arg instanceof Variable) {
				if (i < relation.numNonTempArgs) {
					map.put((Variable)arg, tableName + ".arg" + i);
				} else {
					if (i == relation.numNonTempArgs) {
						map.put((Variable)arg, tableName + ".time_begin");
					} else {
						map.put((Variable)arg, tableName + ".time_end");
					}
				}
			}
			i += 1;
		}
		return map;
	}
	
	/**
	 * Determines a SQL WHERE clause condition that
	 * ensures the constants occuring in the arguments to be set.
	 */
	public String getConstantConditions(String tableName) throws Exception {
		Time time = Time.getInstance();
		StringBuffer buffer = new StringBuffer();
		int i = 0;
		for (AbstractArgument arg : arguments) {
			if (arg instanceof Constant) {
				Constant constant = (Constant)arg;
				if (buffer.length() > 0) {
					buffer.append("AND ");
				}
				
				buffer.append(tableName);
				if (i < relation.numNonTempArgs) {
					buffer.append(".arg");
					buffer.append(i);
					buffer.append("=");
					if (constant.toSet) {
						buffer.append("? ");
					} else {
						buffer.append("\'");
						buffer.append(constant.name);
						buffer.append("\' ");
					}
				} else {
					String date = constant.name;
					int[] timePoints = time.string2int(date);
					if (i == relation.numNonTempArgs) {
						buffer.append(".time_begin=");
						if (constant.toSet) {
							buffer.append("?");
						} else {
							buffer.append(timePoints[0]);
						}
					} else {
						buffer.append(".time_end=");
						if (constant.toSet) {
							buffer.append("?");
						} else {
							if (date.equals("time_end")) {
								buffer.append(timePoints[1]);
							} else {
								buffer.append(timePoints[0]);
							}
						}
					}
					buffer.append(" ");
				}
			}
			i += 1;
		}
		return buffer.toString();
	}
	
	/**
	 * If a variable occurs repeatedly in the literal,
	 * this method determines a condition for a SQL WHERE clause
	 * which assures that all these arguments to take the same value.
	 */
	public String getRepeatedVariables(String tableName) throws Exception {
		Map<Variable, List<Integer>> map = new HashMap<Variable, List<Integer>>();
		int argId = 0;
		for (AbstractArgument arg : arguments) {
			if (arg instanceof Variable) {
				if (!map.containsKey(arg)) {
					map.put((Variable)arg, new LinkedList<Integer>());
				}
				map.get(arg).add(argId);
			}
			argId += 1;
		}
		
		StringBuffer condition = new StringBuffer();
		for (Entry<Variable, List<Integer>> entry : map.entrySet()) {
			if (entry.getValue().size() <= 1) {
				continue;  // the variable is used at most once
			}
			if (entry.getKey().isTemporal) {
				throw new Exception("Don't use the same variable for begin and end of the time interval: " + this);
			}
			List<Integer> argIds = entry.getValue();
			// create chain of equalities
			for (int i = 0; i < argIds.size() - 1; i++) {
				if (condition.length() > 0) {
					condition.append("AND ");
				}
				condition.append(tableName);
				condition.append(".arg");
				condition.append(argIds.get(i));
				condition.append("=");
				condition.append(tableName);
				condition.append(".arg");
				condition.append(argIds.get(i + 1));
				condition.append(" ");
			}
		}
		return condition.toString();
	}
	
	@Override
	public String toString() {
		StringBuffer buffer = new StringBuffer();
		if (relation != null) {
			buffer.append(relation.name);
		} else if (relationName != null){
			buffer.append(relationName);
		}
		buffer.append("(");
		int argNum = 0;
		for (AbstractArgument arg : arguments) {
			if (argNum > 0) {
				buffer.append(", ");
			}
			buffer.append(arg);
			argNum++;
		}
		buffer.append(")");
		return buffer.toString();
	}
}
