package reasoner.weightlearning;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import reasoner.grounding.Grounder;
import reasoner.grounding.Interval;
import reasoner.grounding.IntervalProvider;
import reasoner.grounding.Tuple;
import reasoner.io.dataset.Dataset;
import reasoner.io.dataset.Relation;
import reasoner.io.weightlearning.ParsedLabels;
import reasoner.lineage.LineageArray;
import reasoner.lineage.LineageLeaf;
import reasoner.query.compiled.Query;
import reasoner.weightlearning.problem.LabeledFormula;
import reasoner.weightlearning.problem.ProblemInstance;
import reasoner.weightlearning.problem.TupleToLearn;
import util.time.Timer;

public class WeightLearningGrounder {
	
	private final Grounder grounder;
	private final LineageArray lineageArray;

	public WeightLearningGrounder() throws Exception {
		grounder = new Grounder();
		lineageArray = LineageArray.getInstance();
	}	
	
	/**
	 * invokes the grounding process.
	 * Ensure that constants were propagated.
	 * 
	 * @return run time in ms
	 */
	public long createLineage(Query query) throws Exception {
		Timer timer = Timer.getInstance();
		timer.startTimer(WeightLearningGrounder.class.getName());
		grounder.queryScheduler(query);
		return timer.elapsedAndRemoveTimer(WeightLearningGrounder.class.getName());
	}
	
	public ProblemInstance getProblemInstance(String name, Set<Relation> relationsToLearn, ParsedLabels labels) throws Exception {
		TupleToLearn[] tuples = getTuplesToLearn(relationsToLearn);
		LabeledFormula[] formulas = getLabeledFormulas(labels);
		return new ProblemInstance(name, formulas, tuples);
	}
	
	/**
	 * Creates one tupleToLearn per fact in the relations in @relationToPrior
	 */
	private TupleToLearn[] getTuplesToLearn(Set<Relation> relationsToLearn) throws Exception {
		List<TupleToLearn> tuplesToLearn = new LinkedList<TupleToLearn>();
		for (Relation relation : relationsToLearn) {
			List<Tuple> facts = relation.getFacts();
			for (Tuple fact : facts) {
				for (Interval interval : fact.intervals) {
					LineageLeaf leaf = (LineageLeaf)lineageArray.getLineage(interval.lineageArrayIds.toArray()[0]);
					TupleToLearn tupleToLearn = new TupleToLearn(fact.factString(), leaf);
					tuplesToLearn.add(tupleToLearn);
				}
				IntervalProvider.getInstance().handBack(fact.intervals);
			}
		}
		return tuplesToLearn.toArray(new TupleToLearn[tuplesToLearn.size()]);
	}
	
	private LabeledFormula[] getLabeledFormulas(ParsedLabels labels) throws Exception {
		LinkedList<LabeledFormula> labeledFormulas = new LinkedList<LabeledFormula>();
		for (Relation labeledRelation : labels.occuringRelations.values()) {
			LinkedList<Tuple> facts = labeledRelation.getFacts();
			for (Tuple fact : facts) {
				if (!labels.labels.containsKey(fact.factString())) {
					IntervalProvider.getInstance().handBack(fact.intervals);
					continue;  // no label applies
				} else {
					// label all grounded intervals that are contained in a labeled interval
					List<Interval> labeledIntervals = labels.labels.get(fact.factString()).intervals; 
					for (Interval labeledInterval : labeledIntervals) {
						for (Interval groundedInterval : fact.intervals) {
							if (groundedInterval.isContainedIn(labeledInterval)) {
								LabeledFormula formula = new LabeledFormula();
								formula.label = labeledInterval.confidence;
								formula.lineage = lineageArray.getLineage(groundedInterval.lineageArrayIds.toArray()[0]);
								labeledFormulas.add(formula);
							}
						}
					}
				}
				IntervalProvider.getInstance().handBack(fact.intervals);
			}
		}
		
		for (Map.Entry<String, Tuple> entry : labels.labels.entrySet()) {
			IntervalProvider.getInstance().handBack(entry.getValue().intervals);
		}
		return labeledFormulas.toArray(new LabeledFormula[labeledFormulas.size()]);
	}
}