package reasoner.weightlearning.problem;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import gnu.trove.iterator.TIntDoubleIterator;
import gnu.trove.iterator.TIntIterator;
import gnu.trove.list.linked.TIntLinkedList;
import gnu.trove.map.TIntDoubleMap;
import gnu.trove.map.hash.TIntDoubleHashMap;
import gnu.trove.map.hash.TIntIntHashMap;
import gnu.trove.set.hash.TIntHashSet;
import reasoner.lineage.LineageArray;
import reasoner.lineage.LineageCacheClearer;
import reasoner.lineage.confidence.ConfidenceCalculator;
import reasoner.lineage.providers.LineageProvider;
import reasoner.weightlearning.WeightLearningConfig;
import util.datastructures.UnionFind;
import util.io.FileWriter;

public class ProblemInstance {

	public final String fileName;
	public final TupleToLearn[] tuples;
	public final LabeledFormula[] formulas;
	public final TIntHashSet labeledTupleIds;
	public final List<ComputationTask> independentTasks;
	
	public ProblemInstance(String fileName, LabeledFormula[] formulas, TupleToLearn[] tuples) throws Exception {
		this.formulas = formulas;
		this.tuples = tuples;
		this.fileName = fileName;
		labeledTupleIds = new TIntHashSet();
		TIntIntHashMap arrayId2tupleId = collectLabeledTuples(tuples, formulas);
		createTupleToFormulasIndex(arrayId2tupleId);
		if (WeightLearningConfig.multithreading) {
			independentTasks = tasksFormulas(computeIndependentTasks(tuples, formulas, arrayId2tupleId));
		} else {
			independentTasks = null;
		}
		if (labeledTupleIds.isEmpty()) {
			throw new Exception("No tuples labeled!");
		}
	}
	
	/**
	 * public for testing
	 */
	public TIntIntHashMap collectLabeledTuples(TupleToLearn[] tuples, LabeledFormula[] formulas) {
		TIntHashSet allArrayIds = new TIntHashSet();
		for (int i = 0; i < tuples.length; i++) {
			allArrayIds.add(tuples[i].leaf.id);
		}

		TIntHashSet arrayIdsInFormulas = new TIntHashSet();
		for (LabeledFormula formula : formulas) {
			arrayIdsInFormulas.addAll(formula.lineage.vars);
		}
		arrayIdsInFormulas.retainAll(allArrayIds);
		TIntIntHashMap arrayId2tupleId = new TIntIntHashMap();
		for (int i = 0; i < tuples.length; i++) {
			arrayId2tupleId.put(tuples[i].leaf.id, i);
			if (arrayIdsInFormulas.contains(tuples[i].leaf.id)) {
				labeledTupleIds.add(i);
			}
		}
		return arrayId2tupleId;
	}
	
	private void createTupleToFormulasIndex(TIntIntHashMap arrayId2tupleId) throws Exception {
		for (int i = 0; i < formulas.length; i++) {
			TIntIterator it = formulas[i].lineage.vars.iterator();
			for (int j = 0; j < formulas[i].lineage.vars.size(); j++) {
				int arrayId = it.next();
				if (arrayId2tupleId.containsKey(arrayId)) {
					tuples[arrayId2tupleId.get(arrayId)].labeledFormulas.add(i);
				}
			}
		}
	}
	
	/**
	 * computes a list of indepedent computation tasks.
	 * independence is with respect to tuples to learn.
	 * However, lineage leafs whose confidence is fixed
	 * might be shared!
	 * 
	 * public for testing
	 */
	public List<ComputationTask> computeIndependentTasks(
			TupleToLearn[] tuples,
			LabeledFormula[] formulas,
			TIntIntHashMap arrayId2tupleId) throws Exception {
		UnionFind uf = new UnionFind(arrayId2tupleId.size());
		
		// union all sets, which occur together in a formula
		for (int i = 0; i < formulas.length; i++) {
			TIntLinkedList tupleIds = new TIntLinkedList();
			TIntIterator it = formulas[i].lineage.vars.iterator();
			for (int j = 0; j < formulas[i].lineage.vars.size(); j++) {
				int arrayId = it.next();
				if (arrayId2tupleId.containsKey(arrayId)) {
					// variables not to be learned are skipped
					int tupleId = arrayId2tupleId.get(arrayId);
					tupleIds.add(tupleId);
				}
			}
			
			int[] tupleIdArray = tupleIds.toArray();
			for (int j = 0; j < tupleIdArray.length; j++) {
				for (int k = j + 1; k < tupleIdArray.length; k++) {
					uf.union(tupleIdArray[j], tupleIdArray[k]);
				}
			}
		}
		
		HashMap<Integer, ComputationTask> setId2computation = new HashMap<Integer, ComputationTask>();
		for (int i = 0; i < tuples.length; i++) {
			if (labeledTupleIds.contains(i)) {
				int setId = uf.find(i);
				if (!setId2computation.containsKey(setId)) {
					ComputationTask computation = new ComputationTask();
					setId2computation.put(setId, computation);
				}
				setId2computation.get(setId).tupleIds.add(i);
			}
		}
		LinkedList<ComputationTask> computationList = new LinkedList<ComputationTask>();
		computationList.addAll(setId2computation.values());
		return computationList;
	}
	
	/**
	 * compute formula ids for each task
	 */
	public List<ComputationTask> tasksFormulas(List<ComputationTask> list) {
		for (ComputationTask task : list) {
			int numTuples = task.tupleIds.size();
			TIntIterator it = task.tupleIds.iterator();
			for (int i = 0; i < numTuples; i++) {
				task.formulaIds.addAll(tuples[it.next()].labeledFormulas);
			}
		}
		return list;
	}
	
	public double addGradient(double originalWeight, double learningRate, double gradient) {
		return originalWeight + learningRate * gradient;
	}
	
	/**
	 * Calculates the confidences of all formulas
	 * @param tupleConfidences
	 * @param cacheClearer for concurrency issues it is an argument
	 * @return confidences of the formulas
	 * @throws Exception
	 */
	public double[] calculateAllConfidences(double[] tupleConfidences, LineageCacheClearer cacheClearer, ConfidenceCalculator confidenceCalculator) throws Exception {
		if (tupleConfidences.length != tuples.length) {
			throw new Exception("All tuples must have a confidence!");
		}
		// set confidence
		for (int i = 0; i < tupleConfidences.length; i++) {
			tuples[i].leaf.conf = tupleConfidences[i];
		}
		// clear lineage cache
		for (int i = 0; i < formulas.length; i++) {
			formulas[i].lineage.acceptVisitor(cacheClearer);
		}
		// calculate the formulas confidences
		double[] formulaConfidences = new double[formulas.length];
		for (int i = 0; i < formulas.length; i++) {
			formulaConfidences[i] = confidenceCalculator.confidence(formulas[i].lineage);
		}
		return formulaConfidences;
	}
	
	/** 
	 * Computes confidences of the formulas of the updated tuple
	 * 
	 * @param formulaConfidences : this array will be updated
	 * @return formulaId -> oldConfidence
	 */
	public TIntDoubleMap updateConfidences(
			int updatedTupleId,
			double updatedConfidence,
			double[] formulaConfidences,
			LineageCacheClearer cacheClearer,
			ConfidenceCalculator confidenceCalculator) throws Exception {
		// set confidence
		tuples[updatedTupleId].leaf.conf = updatedConfidence;
		
		// clear lineage cache and recalculate the confidence
		TIntDoubleHashMap formulaIdToOldConfidence = new TIntDoubleHashMap();
		TIntIterator it = tuples[updatedTupleId].labeledFormulas.iterator();
		int numFormulas = tuples[updatedTupleId].labeledFormulas.size();
		for (int i = 0; i < numFormulas; i++) {
			int formulaId = it.next();
			formulaIdToOldConfidence.put(formulaId, formulaConfidences[formulaId]);
			cacheClearer.clearCache(formulas[formulaId].lineage, tuples[updatedTupleId].leaf.id);
			formulaConfidences[formulaId] = confidenceCalculator.confidence(formulas[formulaId].lineage);
		}
		return formulaIdToOldConfidence;
	}
	
	public void resetConfidence(int tupleId, double oldConfidence, TIntDoubleMap formulaIdToOldConfidence, double[] formulaConfidences, LineageCacheClearer cacheClearer) throws Exception {
		// set confidence
		tuples[tupleId].leaf.conf = oldConfidence;
		
		TIntDoubleIterator it = formulaIdToOldConfidence.iterator();
		for (int i = 0; i < formulaIdToOldConfidence.size(); i++) {
			it.advance();
			formulaConfidences[it.key()] = it.value();
			cacheClearer.clearCache(formulas[it.key()].lineage, tuples[tupleId].leaf.id);
		}
	}
	
	public void writeToFile(String path, double[] tupleConfidences) throws Exception {
		if (tupleConfidences.length != tuples.length) {
			throw new Exception("each tuple must have a confidence");
		}
		FileWriter writer = new FileWriter(path);
		for (int i = 0; i < tuples.length; i++) {
			writer.append(tuples[i].leaf.id + "");
			writer.append("\t");
			writer.append(tupleConfidences[i] + "");
			writer.append("\t");
			writer.append(tuples[i].name);
			writer.append("\n");
		}
		writer.close();
	}
	
	public void clear() throws Exception {
		LineageCacheClearer cacheClearer = new LineageCacheClearer();
		for (int i = 0; i < formulas.length; i++) {
			formulas[i].lineage.acceptVisitor(cacheClearer);
		}
	}
	
	public void delete() throws Exception {
		for (int i = 0; i < formulas.length; i++) {
			LineageProvider.getInstance().handBack(formulas[i].lineage, false);
		}
		LineageArray.getInstance().clearQuery();
	}
}
