package reasoner.weightlearning.strategies;

import gnu.trove.TIntCollection;
import gnu.trove.list.linked.TIntLinkedList;
import reasoner.io.weightlearning.WeightLearningEpochesWriter;
import reasoner.lineage.LineageCacheClearer;
import reasoner.lineage.confidence.ConfidenceCalculator;
import reasoner.lineage.confidence.TupleIndependentConfidence;
import reasoner.weightlearning.WeightLearningConfig;
import reasoner.weightlearning.problem.ProblemInstance;
import reasoner.weightlearning.strategies.learningrate.GlobalLearningRate;
import reasoner.weightlearning.strategies.learningrate.LearningRate;
import util.rand.Logit;
import util.time.Timer;

public class GradientDescent implements WeightLearningOptimizer {
	
	private final ProblemInstance problem;
	private final Timer timer;
	private final LearningRate learningRate;
	private final int[] tupleIds;
	private final TIntCollection formulaIds;
	private final ConfidenceCalculator confidenceCalculator;
	private final LineageCacheClearer cacheClearer;
	private final WeightLearningEpochesWriter epochesWriter;
	
	private double initialError;
	private int iteration;
	private TupleConfidenceState currentState;
	private int numNoImprovement;
	
	public GradientDescent(ProblemInstance problem, LearningRate learningRate, WeightLearningEpochesWriter epochesWriter) throws Exception {
		this.epochesWriter = epochesWriter;
		this.problem = problem;
		timer = Timer.getInstance();
		if (learningRate instanceof GlobalLearningRate) {
			this.learningRate = learningRate;
		} else {
			throw new Exception("Gradient Descent requires the global learning rate");
		}
		tupleIds = problem.labeledTupleIds.toArray();
		formulaIds = new TIntLinkedList();
		for (int i = 0; i < problem.formulas.length; i++) {
			formulaIds.add(i);
		}
		confidenceCalculator = new TupleIndependentConfidence();
		cacheClearer = new LineageCacheClearer();
	}

	@Override
	public double getInitialError() throws Exception {
		return initialError;
	}

	@Override
	public double getNumIterations() throws Exception {
		return iteration;
	}

	@Override
	public long getPreprocessingTime() throws Exception {
		return timer.getElapsedTime(GradientDescent.class.getName() + "-prepro");
	}

	@Override
	public long getRunTime() throws Exception {
		return timer.getElapsedTime(GradientDescent.class.getName());
	}

	@Override
	public double run() throws Exception {
		timer.reset(GradientDescent.class.getName() + "-prepro");
		timer.startTimer(GradientDescent.class.getName() + "-prepro");
		WeightLearningConfig.objective.init(problem);
		learningRate.init();
		currentState = new TupleConfidenceState(problem);
		currentState.initAsFirstState();
		iteration = 0;
		timer.stopTimer(GradientDescent.class.getName() + "-prepro");
		
		timer.reset(GradientDescent.class.getName());
		timer.startTimer(GradientDescent.class.getName());
		double oldError = 1.0;
		double error = currentState.calculateError(tupleIds, formulaIds);
		initialError = error;
		if (WeightLearningConfig.resultPath != null) {
			epochesWriter.appendResults(timer.getElapsedTime(GradientDescent.class.getName()), error);
		}
		while (moreIterations(oldError, error)) {
			if (WeightLearningConfig.printing) {
				System.out.println(Timer.getInstance().getElapsedTime(GradientDescent.class.getName()) + "ms Next Epoche. Error: " + error);
			}
			
			TupleConfidenceState nextState = nextState();
			
			oldError = error;
			error = nextState.calculateError(tupleIds, formulaIds);
			
			boolean improved = oldError > error;
			
			for (int i = 0; i < tupleIds.length; i++) {
				learningRate.registerResult(0, improved);
			}
			learningRate.newEpoche();
			
			if (improved) {
				double relativeError = (oldError - error) / oldError;
				if (relativeError > WeightLearningConfig.relativeErrorBound) {
					numNoImprovement = 0;
				} else {
					numNoImprovement += 1;
				}
				currentState = nextState;
			} else {
				numNoImprovement += 1;
				error = oldError;
			}
			
			iteration += 1;
			if (WeightLearningConfig.resultPath != null) {
				epochesWriter.appendResults(timer.getElapsedTime(GradientDescent.class.getName()), error);
			}
		}
		timer.stopTimer(GradientDescent.class.getName());
		return error;
	}

	private TupleConfidenceState nextState() throws Exception {
		
		double[] gradient = new double[currentState.tupleConfidences.length];
		for (int i = 0; i < tupleIds.length; i++) {
			gradient[tupleIds[i]] = WeightLearningConfig.objective.calculateTupleGradient(tupleIds[i], problem, currentState, confidenceCalculator);
		}
		
		TupleConfidenceState nextState = new TupleConfidenceState(problem);
		for (int i = 0; i < tupleIds.length; i++) {
			nextState.tupleWeights[tupleIds[i]] = problem.addGradient(currentState.tupleWeights[tupleIds[i]], learningRate.getLearningRate(tupleIds[i]), gradient[tupleIds[i]]);
			nextState.tupleConfidences[tupleIds[i]] = Logit.probability(nextState.tupleWeights[tupleIds[i]]);
		}
		nextState.formulaConfidences = problem.calculateAllConfidences(nextState.tupleConfidences, cacheClearer, confidenceCalculator);
		return nextState;
	}
	
	private boolean moreIterations(double oldError, double newError) {
		if (iteration >= WeightLearningConfig.maxIterations
				|| newError < WeightLearningConfig.absoluteErrorBound) {
			return false;
		}
		if (numNoImprovement >= 5) {
			return false;
		}
		return true;
	}
}
