package reasoner.weightlearning.strategies;

import reasoner.io.weightlearning.WeightLearningEpochesWriter;
import reasoner.weightlearning.WeightLearningConfig;
import reasoner.weightlearning.problem.ProblemInstance;
import reasoner.weightlearning.strategies.learningrate.LearningRate;
import util.time.Timer;

public class StochasticGradientDescent implements WeightLearningOptimizer {
	
	private final ProblemInstance problem;
	private final Timer timer;
	private final LearningRate learningRate;
	private final StochasticGradientDescentUpdater updater;
	
	public StochasticGradientDescent(ProblemInstance problem, WeightLearningEpochesWriter epochesWriter, LearningRate learningRate) throws Exception {
		this.problem = problem;
		timer = Timer.getInstance();
		this.learningRate = learningRate;
		if (WeightLearningConfig.multithreading
				&& problem.independentTasks.size() > 1) {
			updater = new StochasticGradientDescentMultiThreadedUpdater();
		} else {
			StochasticGradientDescentSingleThreadedUpdater updater = new StochasticGradientDescentSingleThreadedUpdater();
			updater.setEpochesWriter(epochesWriter);
			this.updater = updater; 
		}
	}

	@Override
	public double run() throws Exception {
		timer.reset(StochasticGradientDescent.class.getName() + "-prepro");
		timer.startTimer(StochasticGradientDescent.class.getName() + "-prepro");
		WeightLearningConfig.objective.init(problem);
		learningRate.init();
		TupleConfidenceState state = new TupleConfidenceState(problem);
		state.initAsFirstState();
		updater.init(problem, learningRate, state);
		timer.stopTimer(StochasticGradientDescent.class.getName() + "-prepro");
		
		timer.reset(StochasticGradientDescent.class.getName());
		timer.startTimer(StochasticGradientDescent.class.getName());
		double error = updater.run();
		timer.stopTimer(StochasticGradientDescent.class.getName());
		if (WeightLearningConfig.resultPath != null) {
			state.writeToFile(WeightLearningConfig.resultPath + "/tupleConfidences.txt");
			learningRate.writeToFile(WeightLearningConfig.resultPath + "/learningRate.txt");
		}
		return error;
	}

	@Override
	public long getRunTime() throws Exception {
		return timer.getElapsedTime(StochasticGradientDescent.class.getName());
	}

	@Override
	public long getPreprocessingTime() throws Exception {
		return timer.getElapsedTime(StochasticGradientDescent.class.getName()+"-prepro");
	}

	@Override
	public double getNumIterations() throws Exception {
		return updater.getIterations();
	}

	@Override
	public double getInitialError() throws Exception {
		return updater.getInitialError();
	}
}
