package reasoner.weightlearning.strategies;

import gnu.trove.list.linked.TIntLinkedList;
import gnu.trove.set.hash.TIntHashSet;

import java.util.Collections;
import java.util.List;
import java.util.Random;

import reasoner.lineage.LineageCacheClearer;
import reasoner.lineage.confidence.ConfidenceCalculator;
import reasoner.lineage.confidence.TupleIndependentConfidence;
import reasoner.weightlearning.WeightLearningConfig;
import reasoner.weightlearning.problem.ComputationTask;
import reasoner.weightlearning.problem.ProblemInstance;
import reasoner.weightlearning.strategies.learningrate.LearningRate;
import util.Integers;
import util.time.Timer;

public class StochasticGradientDescentMultiThreadedUpdater implements StochasticGradientDescentUpdater {
	
	private Random javaRandom;
	private List<ComputationTask> tasks;
	private Executer[] threads;

	public void init(ProblemInstance problem, LearningRate learningRate, TupleConfidenceState state) throws Exception {
		if (!learningRate.multithreadable()) {
			throw new Exception("Learning rate " + learningRate + " is not multithreadable");
		}
		tasks = problem.independentTasks;
		javaRandom = new Random(System.currentTimeMillis());
		createThreads(tasks.size(), learningRate, state);
		distributeTuplesToThreads();
	}
	
	@Override
	public double run() throws Exception {
		// start all threads
		for (int i = 0; i < threads.length; i++) {
			threads[i].start();
		}
		// wait for all threads to finish
		for (int i = 0; i < threads.length; i++) {
			try {
				threads[i].join();
			} catch (InterruptedException e) {
				System.err.print("Join interrupted\n");
			}
		}
		if (WeightLearningConfig.printing) {
			System.out.println("All threads are finished");
		}
		double error = 0.0;
		for (int i = 0; i < threads.length; i++) {
			error += threads[i].error;
		}
		return error / threads.length;
	}
	
	@Override
	public double getIterations() {
		double iterations = 0.0;
		for (int i = 0; i < threads.length; i++) {
			iterations += threads[i].iteration;
		}
		return iterations / threads.length;
	}
	
	/**
	 * creates empty lists, one for each thread
	 */
	private void createThreads(int numTasks, LearningRate learningRate, TupleConfidenceState state) throws Exception {
		int numCores = Runtime.getRuntime().availableProcessors();
		if (numCores > 1) {
			numCores -= 1;
		}
		if (numCores > numTasks) {
			numCores = numTasks;
		}
		threads = new Executer[numCores];
		for (int i = 0; i < numCores; i++) {
			threads[i] = new Executer(i, learningRate, state);
		}
		if (WeightLearningConfig.printing) {
			System.out.println("There are " + numTasks + " independent tasks");
			System.out.println("Updating will use " + numCores + " threads");
		}
	}
	
	/**
	 * computes linked lists of tuple ids for each thread
	 */
	private void distributeTuplesToThreads() throws Exception {
		if (tasks.isEmpty() || threads.length == 0) {
			return;
		}
		
		if (WeightLearningConfig.printing) {
			System.out.println("Distributing tasks to theads");
		}
		TIntLinkedList[] tuplesForThreads = new TIntLinkedList[threads.length];
		for (int i = 0; i < threads.length; i++) {
			tuplesForThreads[i] = new TIntLinkedList();
		}
		Collections.shuffle(tasks, javaRandom);
		
		for (ComputationTask task : tasks) {
			// linear search for smallest list
			// this won't scale for many threads
			int minList = -1;
			int minListSize = Integer.MAX_VALUE;
			for (int i = 0; i < threads.length; i++) {
				if (tuplesForThreads[i].size() < minListSize) {
					minListSize = tuplesForThreads[i].size();
					minList = i;
				}
			}
			
			// append task to smallest list
			// a heuristic for balancing the thread's load
			tuplesForThreads[minList].addAll(task.tupleIds);
			threads[minList].formulaIds.addAll(task.formulaIds);
		}
		for (int i = 0; i < threads.length; i++) {
			threads[i].tupleIds = tuplesForThreads[i].toArray();
		}
		if (WeightLearningConfig.printing) {
			for (int i = 0; i < threads.length; i++) {
				System.out.println("Thread " + i + " updates " + threads[i].tupleIds.length + " tuples");
			}
		}
	}
	
	private class Executer extends Thread {
		
		public int iteration;
		public double error;
		public double initialError;
		public int[] tupleIds;
		public final TIntHashSet formulaIds;
		
		private final int threadId;
		private final LineageCacheClearer cacheClearer;
		private final TupleConfidenceState state;
		private final LearningRate learningRate;
		private final ConfidenceCalculator confidenceCalculator;
		
		public Executer(int id, LearningRate learningRate, TupleConfidenceState state) throws Exception {
			threadId = id;
			formulaIds = new TIntHashSet();
			cacheClearer = new LineageCacheClearer();
			confidenceCalculator = new TupleIndependentConfidence();
			this.state = state;
			this.learningRate = learningRate;
		}
		
		public void run() {
			try {
				iteration = 0;
				double oldError = 1.0;
				error = error();
				initialError = error;
				while (WeightLearningConfig.objective.moreIterationsNecessary(oldError, error, iteration, threads.length)) {
					if (WeightLearningConfig.printing) {
						System.out.println("Thread " + threadId + " " + Timer.getInstance().getElapsedTime(StochasticGradientDescent.class.getName()) + "ms Next Epoche. Error: " + error);
					}
					Integers.shuffle(tupleIds);
					for (int i = 0; i < tupleIds.length; i++) {
						boolean improved = state.updateTuple(tupleIds[i], learningRate.getLearningRate(tupleIds[i]), cacheClearer, confidenceCalculator);	
						learningRate.registerResult(tupleIds[i], improved);
					}
					oldError = error;
					error = error();
					iteration += 1;
				}
				if (WeightLearningConfig.printing) {
					System.out.println("Thread " + threadId + " done " + Timer.getInstance().getElapsedTime(StochasticGradientDescent.class.getName()) + "ms Error: " + error);
				}
			} catch (Exception e) {
				System.err.println(e.getMessage());
				throw new RuntimeException(e.getMessage());
			}
		}
		
		private double error() throws Exception {
			return state.calculateError(tupleIds, formulaIds);
		}
	}

	public double getInitialError() {
		double initialError = 0.0;
		for (int i = 0; i < threads.length; i++) {
			initialError += threads[i].initialError;
		}
		return initialError;
	}
}
