package reasoner.weightlearning.strategies;

import gnu.trove.list.linked.TIntLinkedList;
import reasoner.io.weightlearning.WeightLearningEpochesWriter;
import reasoner.lineage.LineageCacheClearer;
import reasoner.lineage.confidence.ConfidenceCalculator;
import reasoner.lineage.confidence.TupleIndependentConfidence;
import reasoner.weightlearning.WeightLearningConfig;
import reasoner.weightlearning.problem.ProblemInstance;
import reasoner.weightlearning.strategies.learningrate.LearningRate;
import util.Integers;
import util.time.Timer;

public class StochasticGradientDescentSingleThreadedUpdater implements StochasticGradientDescentUpdater {
	
	private int[] tupleIds;
	private TIntLinkedList formulaIds;
	private int iteration;
	private double initialError;
	private LearningRate learningRate;
	private LineageCacheClearer cacheClearer;
	private TupleConfidenceState state;
	private ConfidenceCalculator confidenceCalculator;
	private WeightLearningEpochesWriter epochesWriter;

	@Override
	public void init(ProblemInstance problem, LearningRate learningRate, TupleConfidenceState state) throws Exception {
		tupleIds = problem.labeledTupleIds.toArray();
		formulaIds = new TIntLinkedList();
		for (int i = 0; i < problem.formulas.length; i++) {
			formulaIds.add(i);
		}
		confidenceCalculator = new TupleIndependentConfidence();
		this.learningRate = learningRate;
		cacheClearer = new LineageCacheClearer();
		this.state = state;
	}
	
	public void setEpochesWriter(WeightLearningEpochesWriter epochesWriter) {
		this.epochesWriter = epochesWriter;
	}

	@Override
	public double run() throws Exception {
		iteration = 0;
		double oldError = 1.0;
		double newError = error();
		initialError = newError;
		
		if (WeightLearningConfig.resultPath != null) {
			epochesWriter.appendResults(iteration, newError);
		}
		while (WeightLearningConfig.objective.moreIterationsNecessary(oldError, newError, iteration, 1)) {
			if (WeightLearningConfig.printing) {
				System.out.println(Timer.getInstance().getElapsedTime(StochasticGradientDescent.class.getName()) + "ms Next Epoche. Error: " + newError);
			}
			Integers.shuffle(tupleIds);
			for (int i = 0; i < tupleIds.length; i++) {
				boolean improved = state.updateTuple(tupleIds[i], learningRate.getLearningRate(tupleIds[i]), cacheClearer, confidenceCalculator);	
				learningRate.registerResult(tupleIds[i], improved);
			}
			oldError = newError;
			newError = error();
			learningRate.newEpoche();
			iteration += 1;
			if (WeightLearningConfig.resultPath != null) {
				epochesWriter.appendResults(Timer.getInstance().getElapsedTime(StochasticGradientDescent.class.getName()), newError);
			}
		}
		return newError;
	}

	private double error() throws Exception {
		return state.calculateError(tupleIds, formulaIds);
	}

	public double getIterations() {
		return iteration;
	}

	public double getInitialError() {
		return initialError;
	}
}
