package reasoner.weightlearning.strategies;

import gnu.trove.TIntCollection;
import gnu.trove.map.TIntDoubleMap;
import reasoner.lineage.LineageCacheClearer;
import reasoner.lineage.confidence.ConfidenceCalculator;
import reasoner.lineage.confidence.TupleIndependentConfidence;
import reasoner.weightlearning.WeightLearningConfig;
import reasoner.weightlearning.problem.ProblemInstance;
import util.rand.Logit;
import util.rand.Uniform;

public class TupleConfidenceState {
	
	public ProblemInstance problem;
	public double[] tupleWeights;
	public double[] tupleConfidences;
	public double[] tupleError;
	
	public double[] formulaConfidences;
	
	public TupleConfidenceState(ProblemInstance problem) throws Exception {
		this.problem = problem;
		tupleConfidences = new double[problem.tuples.length];
		tupleWeights = new double[problem.tuples.length];
		tupleError = new double[problem.tuples.length];
	}
	
	public void initAsFirstState() throws Exception {
		for (int i = 0; i < tupleConfidences.length; i++) {
			if (problem.labeledTupleIds.contains(i)) {
				tupleWeights[i] = Uniform.getInstance().getDouble(-2.0, 2.0);
			} else {
				tupleWeights[i] = 0.0;
			}
			tupleConfidences[i] = Logit.probability(tupleWeights[i]);
		}
		formulaConfidences = problem.calculateAllConfidences(tupleConfidences, new LineageCacheClearer(), new TupleIndependentConfidence());
		for (int i = 0; i < tupleConfidences.length; i++) {
			tupleError[i] = WeightLearningConfig.objective.calculateError(problem, this, problem.tuples[i].labeledFormulas, i);
		}
	}
	
	/**
	 * tries to update the weight of the tuple.
	 * If the updated weight yields an improvement the update is kept.
	 * Otherwise the old state is restored.
	 * 
	 * @return true if the tuple resulted in an improvement
	 * @throws Exception
	 */
	public boolean updateTuple(int tupleId, double learningRate, LineageCacheClearer cacheClearer, ConfidenceCalculator confidenceCalculator) throws Exception {
		// save old state
		double oldConfidence = tupleConfidences[tupleId];
		
		// create new state
		double gradient = WeightLearningConfig.objective.calculateTupleGradient(tupleId, problem, this, confidenceCalculator);
		double newTupleWeight = problem.addGradient(tupleWeights[tupleId], learningRate, gradient);
		double newTupleConf = Logit.probability(newTupleWeight);
		TIntDoubleMap formulaIdToOldConfidence = problem.updateConfidences(tupleId, newTupleConf, formulaConfidences, cacheClearer, confidenceCalculator);
		
		double newError = WeightLearningConfig.objective.calculateError(problem, this, problem.tuples[tupleId].labeledFormulas, tupleId);
		if (newError < tupleError[tupleId]) {
			tupleError[tupleId] = newError;
			tupleWeights[tupleId] = newTupleWeight;
			tupleConfidences[tupleId] = newTupleConf;
			return true;
		} else {
			problem.resetConfidence(tupleId, oldConfidence, formulaIdToOldConfidence, formulaConfidences, cacheClearer);
			return false;
		}
	}
	
	public double calculateError(int[] tupleIds, TIntCollection formulaIds) throws Exception {
		return WeightLearningConfig.objective.calculateError(problem, this, formulaIds, tupleIds);
	}
	
	public void writeToFile(String pathToFile) throws Exception {
		problem.writeToFile(pathToFile, tupleConfidences);
	}
}
