package tests.reasoner.grounding;

import static org.junit.Assert.*;

import java.util.LinkedList;
import java.util.List;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import reasoner.grounding.Tuple;
import reasoner.grounding.Grounder;
import reasoner.grounding.Interval;
import reasoner.grounding.IntervalProvider;
import reasoner.io.DatabaseManager;
import reasoner.io.QueryParser;
import reasoner.io.Time;
import reasoner.io.dataset.Dataset;
import reasoner.io.dataset.Relation;
import reasoner.lineage.Lineage;
import reasoner.lineage.LineageArray;
import reasoner.lineage.providers.LineageProvider;
import reasoner.query.compiled.Query;
import reasoner.query.compiled.QueryStep;

public class GrounderDbTest {
  private LineageProvider lineageProvider;
  private LineageArray lineageArray;
  private String pathToCode;
  private Dataset dataset;
  private Grounder grounder;
  private Relation[] extensionalRelations;
  private QueryParser parser;
  
  public GrounderDbTest() throws Exception {
    lineageProvider = LineageProvider.getInstance();
    lineageArray = LineageArray.getInstance();
    grounder = new Grounder();
    pathToCode = System.getProperty("user.dir");
  }
  
  @Before
  public void setUp() throws Exception {
  	lineageArray.clearAll();
  	
    // create dataset
  	DatabaseManager manager = DatabaseManager.getInstance();
  	try {
  		manager.deleteDataset("testDataset");
  	} catch (Exception e) {
  		// the dataset did not exist.
  	}
    manager.clearDBall();
    int datasetId = manager.createDataset("testDataset");
    manager.newRelation(datasetId, "occ_test", 2, true, true, true).createTable();
    manager.newRelation(datasetId, "occ_test2", 2, true, true, true).createTable();
    manager.newRelation(datasetId, "occ_test3", 2, true, true, true).createTable();
    manager.newRelation(datasetId, "occ_certain", 2, true, true, false).createTable();
    manager.newRelation(datasetId, "occ_atemporal", 2, false, true, true).createTable();
    manager.finalizeDataset(datasetId, 1);
    
    // load dataset
    dataset = manager.getDataset("testDataset");
    extensionalRelations = new Relation[5];
    extensionalRelations[0] = dataset.relations.get("occ_test");
    extensionalRelations[1] = dataset.relations.get("occ_test2");
    extensionalRelations[2] = dataset.relations.get("occ_test3");
    extensionalRelations[3] = dataset.relations.get("occ_certain");
    extensionalRelations[4] = dataset.relations.get("occ_atemporal");
    
    parser = new QueryParser(dataset);
  }
  
  @After
  public void cleanUp() throws Exception {
    DatabaseManager.getInstance().deleteDataset(dataset.name);
    lineageArray.clearAll();
  }
  
  @Test
  public void testInsertAndRead() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.5, 10, 11);
    
    Query query = new Query(dataset, new LinkedList<QueryStep>());
    query.output.add(extensionalRelations[0]);

    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 11, 1);
    checkIntervalLimits(expected, tuple.intervals);
  }

  @Test
  public void testNot() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.4, 10, 30);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.5, 30, 100);
    
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.2, 10, 30);
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.6, 50, 200);
    
    Query query = parseQueryPlan("/input/testing/grounding/not.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 2);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 30, 1);
    addInterval(expected, 30, 100, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageAnd(lineageProvider.getLineageLeaf(0, 0.0),
                                               lineageProvider.getLineageNot(
                                                 lineageProvider.getLineageLeaf(0, 0.0))));
    lineages.add(lineageProvider.getLineageAnd(lineageProvider.getLineageLeaf(0, 0.0)));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testNotAtemporal() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.4, 10, 30);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.5, 30, 100);
    
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.2, 20, 50);
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.6, 50, 200);
    
    Query query = parseQueryPlan("/input/testing/grounding/notAtemporal.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 2);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 30, 1);
    addInterval(expected, 30, 100, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageAnd(lineageProvider.getLineageLeaf(0, 0.0),
                                               lineageProvider.getLineageNot(
                                                 lineageProvider.getLineageOr(
                                                   lineageProvider.getLineageLeaf(0, 0.0),
                                                   lineageProvider.getLineageLeaf(0, 0.0)))));
    lineages.add(lineageProvider.getLineageAnd(lineageProvider.getLineageLeaf(0, 0.0),
        																			 lineageProvider.getLineageNot(
        																			   lineageProvider.getLineageOr(
        																			     lineageProvider.getLineageLeaf(0, 0.0),
        																			     lineageProvider.getLineageLeaf(0, 0.0)))));

    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
   
  @Test
  public void testNotConstant() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.4, 10, 30);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.5, 30, 100);
    
    addTuple(extensionalRelations[1], "testArg1", "testArg3", 0.2, 10, 30);
    addTuple(extensionalRelations[1], "testArg1", "testArg3", 0.6, 50, 200);
    
    Query query = parseQueryPlan("/input/testing/grounding/notConstant.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 2);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 30, 1);
    addInterval(expected, 30, 100, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageAnd(lineageProvider.getLineageLeaf(0, 0.0),
                                               lineageProvider.getLineageNot(
                                                 lineageProvider.getLineageLeaf(0, 0.0))));
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testNotTemporalConstant() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.4, 10, 30);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.5, 30, 100);
    
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.2, 10, Time.getInstance().maxTime);
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.6, 50, 200);
    
    Query query = parseQueryPlan("/input/testing/grounding/notTemporalConstant.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 2);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 30, 1);
    addInterval(expected, 30, 100, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageAnd(lineageProvider.getLineageLeaf(0, 0.0),
                                               lineageProvider.getLineageNot(
                                                 lineageProvider.getLineageLeaf(0, 0.0))));
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }

 
  @Test
  public void testNotEmpty() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.4, 10, 30);
    
    // occ_test2 is empty
    
    Query query = parseQueryPlan("/input/testing/grounding/not.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 30, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testNotAndJoin() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 30);
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.2, 20, 40);
    addTuple(extensionalRelations[2], "testArg1", "testArg2", 0.3, 10, 30);
    
    Query query = parseQueryPlan("/input/testing/grounding/notAndJoin.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 40, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageAnd(lineageProvider.getLineageLeaf(0, 0.0),
                                               lineageProvider.getLineageLeaf(0, 0.0),
                                               lineageProvider.getLineageNot(
                                                 lineageProvider.getLineageLeaf(0, 0.0))));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testNotTwice() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 30);
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.2, 10, 30);
    addTuple(extensionalRelations[2], "testArg1", "testArg2", 0.3, 10, 30);
    
    Query query = parseQueryPlan("/input/testing/grounding/notTwice.txt");     
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 30, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageAnd(lineageProvider.getLineageLeaf(0, 0.0),
                                               lineageProvider.getLineageNot(
                                                   lineageProvider.getLineageOr(
                                                       lineageProvider.getLineageLeaf(0, 0.0),
                                                       lineageProvider.getLineageLeaf(0, 0.0)))));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testOneJoin() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 30);
    addTuple(extensionalRelations[1], "testArg2", "testArg3", 0.2, 20, 40);
    addTuple(extensionalRelations[1], "testArg2", "testArg4", 0.3, 20, 40);
    
    Query query = parseQueryPlan("/input/testing/grounding/oneJoin.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 2);
    for (int i = 0; i < 2; i++) {
      Tuple tuple = output.get(i);
      assertTrue(tuple.intervals.size() == 1);
      
      LinkedList<Interval> expected = new LinkedList<Interval>();
      addInterval(expected, 10, 20, 1);
      checkIntervalLimits(expected, tuple.intervals);
      
      LinkedList<Lineage> lineages = new LinkedList<Lineage>();
      lineages.add(lineageProvider.getLineageAnd(lineageProvider.getLineageLeaf(0, 0.0),
                                                 lineageProvider.getLineageLeaf(0, 0.0)));
      checkLineageTypes(lineages, tuple.intervals, lineageArray);
    }
  }
  
  @Test
  public void testOneJoinTemporal() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 30);
    addTuple(extensionalRelations[1], "testArg3", "testArg4", 0.2, 30, 40);
    addTuple(extensionalRelations[1], "testArg3", "testArg5", 0.3, 20, 50);
    
    Query query = parseQueryPlan("/input/testing/grounding/oneJoinTemporal.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 40, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageAnd(lineageProvider.getLineageLeaf(0, 0.0),
                                               lineageProvider.getLineageLeaf(0, 0.0)));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testTwoJoins() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 30);
    addTuple(extensionalRelations[1], "testArg1", "testArg3", 0.2, 20, 40);
    addTuple(extensionalRelations[2], "testArg4", "testArg3", 0.3, 30, 50);
    addTuple(extensionalRelations[2], "testArg5", "testArg3", 0.4, 40, 50);
    
    Query query = parseQueryPlan("/input/testing/grounding/twoJoins.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 40, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageOr(lineageProvider.getLineageAnd(
                                                  lineageProvider.getLineageLeaf(0, 0.0),
                                                  lineageProvider.getLineageLeaf(0, 0.0),
                                                  lineageProvider.getLineageLeaf(0, 0.0)),
                                              lineageProvider.getLineageAnd(
                                                  lineageProvider.getLineageLeaf(0, 0.0),
                                                  lineageProvider.getLineageLeaf(0, 0.0),
                                                  lineageProvider.getLineageLeaf(0, 0.0))));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testEmptyJoinCondition() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 30);
    addTuple(extensionalRelations[0], "testArg1", "testArg3", 0.2, 10, 30);
    addTuple(extensionalRelations[1], "testArg3", "testArg4", 0.3, 20, 40);
    addTuple(extensionalRelations[1], "testArg3", "testArg5", 0.4, 30, 50);
    
    Query query = parseQueryPlan("/input/testing/grounding/emptyJoinCondition.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 2);
    
    LinkedList<Interval> expectedInterval = new LinkedList<Interval>();
    addInterval(expectedInterval, 10, 30, 1);
    
    LinkedList<Lineage> expectedLineage = new LinkedList<Lineage>();
    expectedLineage.add(lineageProvider.getLineageOr(lineageProvider.getLineageAnd(
                                                  lineageProvider.getLineageLeaf(0, 0.0),
                                                  lineageProvider.getLineageLeaf(0, 0.0)),
                                              lineageProvider.getLineageAnd(
                                                  lineageProvider.getLineageLeaf(0, 0.0),
                                                  lineageProvider.getLineageLeaf(0, 0.0))));
    
    for (Tuple tuple : output) {
	    assertTrue(tuple.intervals.size() == 1);
	    checkIntervalLimits(expectedInterval, tuple.intervals);
	    checkLineageTypes(expectedLineage, tuple.intervals, lineageArray);
    }
  }
  
  @Test
  public void testTwoRules() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 30);
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.2, 20, 40);
    
    Query query = parseQueryPlan("/input/testing/grounding/twoRules.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 3);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 20, 1);
    addInterval(expected, 20, 30, 1);
    addInterval(expected, 30, 40, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    lineages.add(lineageProvider.getLineageOr(lineageProvider.getLineageLeaf(0, 0.0),
                                              lineageProvider.getLineageLeaf(0, 0.0)));
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testIntervalBegin() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 30);
    
    Query query = parseQueryPlan("/input/testing/grounding/intervalBegin.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, Time.getInstance().minTime, 30, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }

 
  @Test
  public void testIntervalBoth() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 30);
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.2, 40, 50);
    
    Query query = parseQueryPlan("/input/testing/grounding/intervalBoth.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 50, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageAnd(
                   lineageProvider.getLineageLeaf(0, 0.0),
                   lineageProvider.getLineageLeaf(0, 0.0)));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
   
  @Test
  public void testIntervalEnd() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 30);
    
    Query query = parseQueryPlan("/input/testing/grounding/intervalEnd.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, Time.getInstance().maxTime, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testNotEquals() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg1", 0.1, 10, 30);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.2, 20, 40);
    
    Query query = parseQueryPlan("/input/testing/grounding/notEquals.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 20, 40, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testBefore() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 20);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.2, 10, 50); // not in
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.1, 20, 60);
    
    Query query = parseQueryPlan("/input/testing/grounding/before.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 20, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageAnd(
                    lineageProvider.getLineageLeaf(0, 0.0),
                    lineageProvider.getLineageLeaf(0, 0.0)));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testOverlaps() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 20); // not in
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.2, 10, 50);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.3, 40, 100);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.4, 30, 40);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.5, 20, 60);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.5, 60, 100); // not in

    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.6, 20, 60);
    
    Query query = parseQueryPlan("/input/testing/grounding/overlaps.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 6);
  }

  @Test
  public void testDisjoint() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 20); // in
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.2, 10, 50);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.3, 40, 100);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.4, 30, 40);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.5, 20, 60);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.6, 60, 100); // in

    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.7, 20, 60);
    
    Query query = parseQueryPlan("/input/testing/grounding/disjoint.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 2);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageAnd(
                    lineageProvider.getLineageLeaf(0, 0.0),
                    lineageProvider.getLineageLeaf(0, 0.0)));
    lineages.add(lineageProvider.getLineageAnd(
        lineageProvider.getLineageLeaf(0, 0.0),
        lineageProvider.getLineageLeaf(0, 0.0)));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
    
  @Test
  public void testDualInterval() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 30);
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.7, 20, 60);
    
    Query query = parseQueryPlan("/input/testing/grounding/dualInterval.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    Interval interval = tuple.intervals.get(0);
    assertTrue(interval.begin == 20);
    assertTrue(interval.end == 30);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageAnd(
                    lineageProvider.getLineageLeaf(0, 0.0),
                    lineageProvider.getLineageLeaf(0, 0.0)));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
 
  @Test
  public void testConstantInHead() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 20, 60);
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.7, 20, 60);
    
    Query query = parseQueryPlan("/input/testing/grounding/constantInHead.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageOr(
                    lineageProvider.getLineageLeaf(0, 0.0),
                    lineageProvider.getLineageLeaf(0, 0.0)));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
 
  @Test
  public void testConstantInBody() throws Exception {
  	addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.5, 10, 11);
  	addTuple(extensionalRelations[0], "testArg1", "testArg3", 0.5, 10, 11);
  	
  	Query query = parseQueryPlan("/input/testing/grounding/constantInBody.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }

  @Test
  public void testTemporalConstantInBody() throws Exception {
  	Time.getInstance().setParsingOptions("-", "X", false, "time_begin", "time_end");
  	addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.5, 10, Time.getInstance().string2int("2014-01-01")[0]);
  	addTuple(extensionalRelations[0], "testArg1", "testArg3", 0.5, 10, 11);
  	
  	Query query = parseQueryPlan("/input/testing/grounding/temporalConstantInBody.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    assertTrue(tuple.intervals.get(0).end == Time.getInstance().maxTime);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testNegationOnBody() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 20, 60);
    addTuple(extensionalRelations[1], "testArg1", "testArg3", 0.2, 20, 60);
    addTuple(extensionalRelations[2], "testArg1", "testArg3", 0.3, 20, 60);
    
    Query query = parseQueryPlan("/input/testing/grounding/negationOnBody.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageAnd(
                    lineageProvider.getLineageLeaf(0, 0.0),
                    lineageProvider.getLineageLeaf(0, 0.0),
                    lineageProvider.getLineageNot(lineageProvider.getLineageLeaf(0, 0.0))));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
 
  @Test
  public void testEmptyIntervalDeduced() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 30, 60);
    addTuple(extensionalRelations[1], "testArg1", "testArg3", 0.2, 10, 20);
    
    Query query = parseQueryPlan("/input/testing/grounding/emptyInterval.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 0);
  }
  
  @Test
  public void testCommentAndWhiteSpace() throws Exception {
  	addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 30);
    addTuple(extensionalRelations[1], "testArg2", "testArg3", 0.2, 20, 40);
    addTuple(extensionalRelations[1], "testArg2", "testArg4", 0.3, 20, 40);
    
    Query query = parseQueryPlan("/input/testing/grounding/commentAndWhiteSpace.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 2);
    for (int i = 0; i < 2; i++) {
      Tuple tuple = output.get(i);
      assertTrue(tuple.intervals.size() == 1);
      
      LinkedList<Interval> expected = new LinkedList<Interval>();
      addInterval(expected, 10, 20, 1);
      checkIntervalLimits(expected, tuple.intervals);
      
      LinkedList<Lineage> lineages = new LinkedList<Lineage>();
      lineages.add(lineageProvider.getLineageAnd(lineageProvider.getLineageLeaf(0, 0.0),
                                                 lineageProvider.getLineageLeaf(0, 0.0)));
      checkLineageTypes(lineages, tuple.intervals, lineageArray);
    }
  }
  
  @Test
  public void testSameVariable() throws Exception {
  	addTuple(extensionalRelations[0], "testArg1", "testArg1", 0.1, 10, 30);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.2, 20, 40);
    
    Query query = parseQueryPlan("/input/testing/grounding/sameVariable.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
    
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
    
    assertTrue(output.size() == 1);
    Tuple tuple = output.get(0);
    assertTrue(tuple.intervals.size() == 1);
    
    LinkedList<Interval> expected = new LinkedList<Interval>();
    addInterval(expected, 10, 30, 1);
    checkIntervalLimits(expected, tuple.intervals);
    
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testSameTemporalVariable() throws Exception {
    try {
    	parseQueryPlan("/input/testing/grounding/sameTemporalVariable.txt");
    	assertTrue(false);
    } catch (Exception e) {
    	// good!
    }
  }
  

  @Test
  public void constraintJoin() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 20);
    addTuple(extensionalRelations[0], "testArg1", "testArg3", 0.2, 10, 20);
    addTuple(extensionalRelations[1], "testArg1", "testArg3", 0.3, 10, 50);
    
    Lineage groundedConstraints = groundConstraints("/input/testing/constraints/join.txt");
    
	  Lineage correctLineage = lineageProvider.getLineageAnd(
																	lineageProvider.getLineageAnd(lineageArray.getLineage(0),
																																lineageArray.getLineage(1)),
	    														lineageProvider.getLineageAnd(lineageArray.getLineage(0),
	    																													lineageArray.getLineage(2)));
    assertTrue(checkLineage(correctLineage, groundedConstraints));
  }
  
  @Test
  public void constraintJoinNegated() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 20);
    addTuple(extensionalRelations[0], "testArg1", "testArg3", 0.2, 10, 20);
    addTuple(extensionalRelations[1], "testArg1", "testArg3", 0.3, 10, 50);
    
    Lineage groundedConstraints = groundConstraints("/input/testing/constraints/joinNegated.txt");
    
	  Lineage correctLineage = lineageProvider.getLineageAnd(
	  															lineageProvider.getLineageNot(
	  																	lineageProvider.getLineageAnd(
	  																			lineageArray.getLineage(0),
	  																			lineageArray.getLineage(1))),
	  															lineageProvider.getLineageNot(
	  																	lineageProvider.getLineageAnd(
	  																			lineageArray.getLineage(0),
	    																		lineageArray.getLineage(2))));
    assertTrue(checkLineage(correctLineage, groundedConstraints));
  }

  
  @Test
  public void constraintJoins() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 20);
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.2, 10, 20);
    addTuple(extensionalRelations[2], "testArg1", "testArg2", 0.3, 10, 50);
    
    Lineage groundedConstraints = groundConstraints("/input/testing/constraints/joins.txt");
    
	  Lineage correctLineage = lineageProvider.getLineageAnd(
	  		lineageArray.getLineage(0),
				lineageArray.getLineage(1),
				lineageArray.getLineage(2));
    assertTrue(checkLineage(correctLineage, groundedConstraints));
  }
  
  @Test
  public void constraintNotEquals() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 20);
    addTuple(extensionalRelations[0], "testArg1", "testArg3", 0.2, 10, 20); // will be omitted
    addTuple(extensionalRelations[1], "testArg1", "testArg3", 0.3, 10, 50);
    
    Lineage groundedConstraints = groundConstraints("/input/testing/constraints/notEquals.txt");
    
    Lineage correctLineage = lineageProvider.getLineageAnd(
    		lineageArray.getLineage(0),													
    		lineageArray.getLineage(2));
    assertTrue(checkLineage(correctLineage, groundedConstraints));
  }

  @Test
  public void constraintBefore() throws Exception {
  	addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.0, 10, 20);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 20);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.2, 10, 50);  // will be omited
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.3, 30, 50);
    
    Lineage groundedConstraints = groundConstraints("/input/testing/constraints/before.txt");
    
    Lineage correctLineage =
    	lineageProvider.getLineageAnd(
    	lineageProvider.getLineageAnd(
    			lineageArray.getLineage(1),													
    			lineageArray.getLineage(4)),
    	lineageProvider.getLineageAnd(
    			lineageArray.getLineage(2),													
    			lineageArray.getLineage(4)));
    assertTrue(checkLineage(correctLineage, groundedConstraints));
  }
  
  @Test
  public void constraintDisjoint() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 20);
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.2, 50, 60);  
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.3, 10, 40);  // will be omited
    addTuple(extensionalRelations[1], "testArg1", "testArg2", 0.4, 30, 50);
    
    Lineage groundedConstraints = groundConstraints("/input/testing/constraints/disjoint.txt");
    
    Lineage correctLineage =
    	lineageProvider.getLineageAnd(
    			lineageProvider.getLineageAnd(
    					lineageArray.getLineage(1),
    					lineageArray.getLineage(4)),
    			lineageProvider.getLineageAnd(
        			lineageArray.getLineage(2),
        			lineageArray.getLineage(4)));
    assertTrue(checkLineage(correctLineage, groundedConstraints));
  }
  
  @Test
  public void constraintsEmpty() throws Exception {
  	Lineage groundedConstraints = groundConstraints("/input/testing/constraints/empty.txt");
    assertNull(groundedConstraints);
  }
  
  @Test
  public void outputWithConstraints() throws Exception {
    addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 10, 20);
    addTuple(extensionalRelations[0], "testArg1", "testArg3", 0.2, 50, 60);
    Lineage constraintLineage = lineageProvider.getLineageLeaf(3, 0.5);
    
    Query query = new Query(dataset, new LinkedList<QueryStep>());
    query.output.add(extensionalRelations[0]);
    
    LinkedList<Tuple> facts = grounder.output(query, constraintLineage);
    
    assertTrue(facts.size() == 2);
    Lineage correctLineage = lineageProvider.getLineageAnd(lineageArray.getLineage(0), lineageArray.getLineage(1));
    for (Tuple fact : facts) {
    	assertTrue(checkLineage(lineageArray.getLineage(fact.intervals.get(0).lineageArrayIds.toArray()[0]), correctLineage));
    }
  }

  @Test
  public void oneCertain() throws Exception {
  	addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 20, 60);
    addTuple(extensionalRelations[0], "testArg2", "testArg2", 0.2, 20, 60);
    addTuple(extensionalRelations[3], "testArg2", "testArg3", 1.0, 20, 60);
     
    Query query = parseQueryPlan("/input/testing/grounding/oneCertain.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
     
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
     
    assertTrue(output.size() == 2);
    for (Tuple tuple : output) {
	    assertTrue(tuple.intervals.size() == 1);
	    
	    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
	    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
	    checkLineageTypes(lineages, tuple.intervals, lineageArray);
    }
  }
  
  @Test
  public void twoCertain() throws Exception {
  	addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 20, 60);
    addTuple(extensionalRelations[3], "testArg1", "testArg2", 1.0, 20, 60);
    addTuple(extensionalRelations[3], "testArg2", "testArg3", 1.0, 20, 60);
    
    Query query = parseQueryPlan("/input/testing/grounding/twoCertain.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
     
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
     
    assertTrue(output.size() == 1);
    Tuple tuple = output.getFirst();
    assertTrue(tuple.intervals.size() == 1);
     
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void only() throws Exception {
  	try {
     parseQueryPlan("/input/testing/grounding/onlyCertain.txt");
     assertTrue(false);
  	} catch (Exception e) {
  		// good!
  	}
  }
  
  @Test
  public void certainInNegation() throws Exception {
  	addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 20, 60);
    addTuple(extensionalRelations[3], "testArg1", "testArg2", 1.0, 20, 60);
    
    Query query = parseQueryPlan("/input/testing/grounding/certainInNegation.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
     
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
     
    assertTrue(output.size() == 1);
    Tuple tuple = output.getFirst();
    assertTrue(tuple.intervals.size() == 1);
     
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testAtemporal() throws Exception {
  	addTuple(extensionalRelations[4], "testArg1", "testArg2", 0.1);
    
    Query query = parseQueryPlan("/input/testing/grounding/atemporal.txt");
    Relation relation = dataset.relations.get("intensional_test");
    assertFalse(relation.isTemporal);
    query.output.add(relation);
     
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
     
    assertTrue(output.size() == 1);
    Tuple tuple = output.getFirst();
    assertTrue(tuple.intervals.size() == 1);
     
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testAtemporalTwoLevels() throws Exception {
  	addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 20, 60);
  	addTuple(extensionalRelations[4], "testArg1", "testArg2", 0.1);
    
    Query query = parseQueryPlan("/input/testing/grounding/atemporalTwoLevels.txt"); 
    query.output.add(dataset.relations.get("intensional_test"));
     
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
     
    assertTrue(output.size() == 1);
    Tuple tuple = output.getFirst();
    assertTrue(tuple.intervals.size() == 1);
     
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(
    		lineageProvider.getLineageAnd(
    				lineageProvider.getLineageLeaf(0, 0.0),
    				lineageProvider.getLineageLeaf(0, 0.0)));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testTemporalFromAtemporal() throws Exception {
  	addTuple(extensionalRelations[4], "testArg1", "testArg2", 0.1);
    
    Query query = parseQueryPlan("/input/testing/grounding/temporalFromAtemporal.txt"); 
    Relation relation = dataset.relations.get("intensional_test");
    assertTrue(relation.isTemporal);
    query.output.add(relation);
     
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
     
    assertTrue(output.size() == 1);
    Tuple tuple = output.getFirst();
    assertTrue(tuple.intervals.size() == 1);
     
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(lineageProvider.getLineageLeaf(0, 0.0));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  
  @Test
  public void testTemporalInferenceTwoLevels() throws Exception {
  	addTuple(extensionalRelations[0], "testArg1", "testArg2", 0.1, 20, 60);
  	addTuple(extensionalRelations[1], "testArg1", "testArg3", 0.1, 60, 80);
    
    Query query = parseQueryPlan("/input/testing/grounding/temporalInferenceTwoLevels.txt"); 
    Relation relation = dataset.relations.get("intensional_test");
    assertTrue(relation.isTemporal);
    query.output.add(relation);
     
    grounder.queryScheduler(query);
    LinkedList<Tuple> output = grounder.output(query);
     
    assertTrue(output.size() == 1);
    Tuple tuple = output.getFirst();
    assertTrue(tuple.intervals.size() == 1);
    Interval interval = tuple.intervals.get(0);
    assertTrue(interval.begin == 20);
    assertTrue(interval.end == 80);
     
    LinkedList<Lineage> lineages = new LinkedList<Lineage>();
    lineages.add(
    		lineageProvider.getLineageAnd(
    				lineageProvider.getLineageLeaf(0, 0.0),
    				lineageProvider.getLineageLeaf(0, 0.0)));
    checkLineageTypes(lineages, tuple.intervals, lineageArray);
  }
  

  private Lineage groundConstraints(String constraintsFile) throws Exception {
    Query query = new Query(dataset, new LinkedList<QueryStep>());
    parser.parseConstraints(pathToCode + constraintsFile, query);
    return grounder.groundConstraints(query.constraints);
  }
  
  private void checkLineageTypes(List<Lineage> expected, List<Interval> intervals, LineageArray lineageArray) throws Exception {
    for (int i = 0; i < expected.size(); i++) {
      assertTrue(checkLineage(expected.get(i), lineageArray.getLineage(intervals.get(i).lineageArrayIds.toArray()[0])));
    }
  }
  
  /**
   * compare two lineages whether they have exactly the same structure
   */
  private boolean checkLineage(Lineage a, Lineage b) {
    if (a.getClass().equals(b.getClass())) {
      if (a.subLineage != null && b.subLineage != null) {
        if (a.subLineage.length != b.subLineage.length) {
          return false;
        }
        for (int i = 0; i < a.subLineage.length; i++) {
          boolean correct = checkLineage(a.subLineage[i], b.subLineage[i]);
          if (!correct) {
            return false;
          }
        }
      } else {
        if (a.subLineage != null || b.subLineage != null) {
          return false;
        }
      }
      return true;
    } else {
      return false;
    }
  }
  
  private void addInterval(List<Interval> list, int begin, int end, int... arrayIds) throws Exception {
    Interval newInterval = IntervalProvider.getInstance().get(begin, end, arrayIds);
    list.add(newInterval);
  }
  
  /**
   * compare the time limits of the intervals
   * 
   * assumes that both lists follow the same order
   */
  private void checkIntervalLimits(List<Interval> expected, List<Interval> actual) {
    assertTrue(actual.size() == expected.size());
    for (int listPos = 0; listPos < actual.size(); listPos++) {
      assertEquals(expected.get(listPos).begin, actual.get(listPos).begin);
      assertEquals(expected.get(listPos).end, actual.get(listPos).end);
    }
  }
  
  private void addTuple(Relation relation, String arg1, String arg2, double conf) throws Exception {
  	addTuple(relation, arg1, arg2, conf, Time.getInstance().minTime, Time.getInstance().maxTime);
  }
  
  private void addTuple(Relation relation, String arg1, String arg2, double conf, int timeBegin, int timeEnd) throws Exception {
  	Interval interval = null;
  	if (relation.isUncertain) {
  		int lineageId = lineageArray.getNextArrayId();
  		lineageArray.addLineage(LineageProvider.getInstance().getLineageLeaf(lineageId, conf));
  		interval = IntervalProvider.getInstance().get(timeBegin, timeEnd, lineageId);
  	} else {
  		interval = IntervalProvider.getInstance().get(timeBegin, timeEnd);
  	}
    String[] arguments = {arg1, arg2};
    relation.addFactToBatch(arguments, conf, interval);
    relation.executeBatch();
  }
  
  private Query parseQueryPlan(String path) throws Exception {
  	return parser.parseQueryPlan(pathToCode + path);
  }
}
