package tests.reasoner.lineage;

import static org.junit.Assert.*;

import org.junit.After;
import org.junit.Test;

import reasoner.io.DatabaseManager;
import reasoner.lineage.LineageArray;
import reasoner.lineage.LineageDisjointOr;
import reasoner.lineage.LineageLeaf;
import reasoner.lineage.confidence.TupleIndependentConfidence;
import reasoner.lineage.providers.LineageProvider;
import util.Doubles;

public class LineageTest {
  private LineageProvider provider;
  private LineageArray lineageArray;
  private TupleIndependentConfidence confidenceCalculator;
  
  public LineageTest() throws Exception {
  	provider = LineageProvider.getInstance();
  	lineageArray = LineageArray.getInstance();
  	confidenceCalculator = new TupleIndependentConfidence();
  	lineageArray.clearAll();
  	DatabaseManager.getInstance().loadDummyDataset();
  }
  
  @After
  public void cleanUp() throws Exception {
  	lineageArray.clearAll();
  }
  
  @Test
  public void not() throws Exception {
    lineageArray.addLineage(provider.getLineageLeaf(0, 0.8));
    lineageArray.addLineage(provider.getLineageNot(lineageArray.getLineage(0)));
    double conf = confidenceCalculator.confidence(lineageArray.getLineage(1));
    assertTrue(Doubles.equals(conf, 0.2));
  }
  
  @Test
  public void independentAnd() throws Exception {
    lineageArray.addLineage(provider.getLineageLeaf(0, 0.8));
    lineageArray.addLineage(provider.getLineageLeaf(1, 0.5));
    lineageArray.addLineage(provider.getLineageAnd(lineageArray.getLineage(0), lineageArray.getLineage(1)));
    double conf = confidenceCalculator.confidence(lineageArray.getLineage(2));
    assertTrue(Doubles.equals(conf, 0.4));
  }
  
  @Test
  public void independentOr() throws Exception {
    lineageArray.addLineage(provider.getLineageLeaf(0, 0.8));
    lineageArray.addLineage(provider.getLineageLeaf(1, 0.4));
    lineageArray.addLineage(provider.getLineageOr(lineageArray.getLineage(0), lineageArray.getLineage(1)));
    double conf = confidenceCalculator.confidence(lineageArray.getLineage(2));
    assertTrue(Doubles.equals(conf, 0.88));
  }
  
  @Test
  public void shannon() throws Exception {
    lineageArray.addLineage(provider.getLineageLeaf(0, 0.8));
    lineageArray.addLineage(provider.getLineageLeaf(1, 0.4));
    lineageArray.addLineage(provider.getLineageLeaf(2, 0.3));
    lineageArray.addLineage(provider.getLineageAnd(lineageArray.getLineage(0), lineageArray.getLineage(1)));
    lineageArray.addLineage(provider.getLineageAnd(lineageArray.getLineage(1), lineageArray.getLineage(2)));
    lineageArray.addLineage(provider.getLineageOr(lineageArray.getLineage(3), lineageArray.getLineage(4)));
    double conf = confidenceCalculator.confidence(lineageArray.getLineage(5));
    assertTrue(Doubles.equals(conf, 0.344));
  }
  
  @Test 
  public void orConfOne() throws Exception {
    lineageArray.addLineage(provider.getLineageLeaf(0, 1.0));
    lineageArray.addLineage(provider.getLineageLeaf(1, 0.4));
    lineageArray.addLineage(provider.getLineageOr(lineageArray.getLineage(0), lineageArray.getLineage(1)));
    double conf = confidenceCalculator.confidence(lineageArray.getLineage(2));
    assertTrue(Doubles.equals(conf, 1.0));
  }
  
  @Test
  public void andConfZero() throws Exception {
    lineageArray.addLineage(provider.getLineageLeaf(0, 0.0));
    lineageArray.addLineage(provider.getLineageLeaf(1, 0.5));
    lineageArray.addLineage(provider.getLineageAnd(lineageArray.getLineage(0), lineageArray.getLineage(1)));
    double conf = confidenceCalculator.confidence(lineageArray.getLineage(2));
    assertTrue(Doubles.equals(conf, 0.0));
  }
  
  @Test
  public void shannonNested() throws Exception {
    lineageArray.addLineage(provider.getLineageLeaf(0, 0.8));
    lineageArray.addLineage(provider.getLineageLeaf(1, 0.4));
    lineageArray.addLineage(provider.getLineageLeaf(2, 0.3));
    lineageArray.addLineage(provider.getLineageAnd(lineageArray.getLineage(0), lineageArray.getLineage(1)));
    lineageArray.addLineage(provider.getLineageAnd(lineageArray.getLineage(1), lineageArray.getLineage(2)));
    lineageArray.addLineage(provider.getLineageOr(lineageArray.getLineage(3), lineageArray.getLineage(4)));
    lineageArray.addLineage(provider.getLineageLeaf(6, 0.1));
    lineageArray.addLineage(provider.getLineageOr(lineageArray.getLineage(2), lineageArray.getLineage(6)));
    lineageArray.addLineage(provider.getLineageAnd(lineageArray.getLineage(5), lineageArray.getLineage(7)));
    double conf = confidenceCalculator.confidence(lineageArray.getLineage(8));
    assertTrue(Doubles.equals(conf, 0.1424));
  }
  
  @Test
  public void testLazyConfComp() throws Exception {
  	lineageArray.addLineage(provider.getLineageLeaf(0, 0.8));
    lineageArray.addLineage(provider.getLineageLeaf(1, 0.4));
    lineageArray.addLineage(provider.getLineageLeaf(2, 0.3));
    lineageArray.addLineage(provider.getLineageAnd(lineageArray.getLineage(0), lineageArray.getLineage(1)));
    lineageArray.addLineage(provider.getLineageOr(lineageArray.getLineage(2), lineageArray.getLineage(3)));
    lineageArray.addLineage(provider.getLineageAnd(lineageArray.getLineage(2), lineageArray.getLineage(3)));
    double conf1 = confidenceCalculator.confidence(lineageArray.getLineage(4));
    double conf2 = confidenceCalculator.confidence(lineageArray.getLineage(5));
    assertTrue(Doubles.equals(conf1, 0.524));
    assertTrue(Doubles.equals(conf2, 0.096));
  }
  
  @Test
  public void testDisjointOr() throws Exception {
  	lineageArray.addLineage(provider.getLineageLeaf(0, 0.8));
    lineageArray.addLineage(provider.getLineageLeaf(1, 0.4));
    lineageArray.addLineage(provider.getLineageLeaf(2, 0.3));
    lineageArray.addLineage(provider.getLineageDisjointOr(lineageArray.getLineage(1), lineageArray.getLineage(2), (LineageLeaf)lineageArray.getLineage(0)));
    LineageDisjointOr disjointOr = (LineageDisjointOr)lineageArray.getLineage(3); 
    double conf = confidenceCalculator.confidence(disjointOr);
    assertTrue(Doubles.equals(conf, 0.8 * 0.4 + 0.2 * 0.3));
    // disjointOr contains all variables
    assertTrue(disjointOr.vars.contains(0));
    assertTrue(disjointOr.vars.contains(1));
    assertTrue(disjointOr.vars.contains(2));
    assertTrue(disjointOr.vars.size() == 3);
    // disjointOr children
    assertTrue(disjointOr.subLineage[0].vars.contains(1));
    assertTrue(disjointOr.subLineage[0].vars.size() == 1);
    assertTrue(disjointOr.subLineage[1].vars.contains(2));
    assertTrue(disjointOr.subLineage[1].vars.size() == 1);
  }
  
  @Test
  public void testDisjointOrShannon() throws Exception {
  	lineageArray.addLineage(provider.getLineageLeaf(0, 0.8));
    lineageArray.addLineage(provider.getLineageLeaf(1, 0.4));
    lineageArray.addLineage(provider.getLineageDisjointOr(lineageArray.getLineage(1), lineageArray.getLineage(1), (LineageLeaf)lineageArray.getLineage(0)));
    LineageDisjointOr disjointOr = (LineageDisjointOr)lineageArray.getLineage(2); 
    double conf = confidenceCalculator.confidence(disjointOr);
    assertTrue(Doubles.equals(conf, 0.8 * 0.4 + 0.2 * 0.4));
    assertTrue(disjointOr.varId2count.isEmpty());
  }
  
  @Test
  public void testTrue() throws Exception {
  	assertTrue(lineageArray.addLineage(provider.getLineageConstant(true)) == 0);
    double conf = confidenceCalculator.confidence(lineageArray.getLineage(0));
    assertTrue(Doubles.equals(conf, 1.0));
  }
  
  @Test
  public void testFalse() throws Exception {
  	assertTrue(lineageArray.addLineage(provider.getLineageConstant(false)) == 0);
    double conf = confidenceCalculator.confidence(lineageArray.getLineage(0));
    assertTrue(Doubles.equals(conf, 0.0));
  }
  
  @Test
  public void testDerivativeLeaf() throws Exception {
  	lineageArray.addLineage(provider.getLineageLeaf(0, 0.8));
  	double derivative = confidenceCalculator.derivative(lineageArray.getLineage(0), 0);
  	assertTrue(Doubles.equals(derivative, 1.0));
  }
  
  @Test
  public void testDerivativeNot() throws Exception {
  	lineageArray.addLineage(provider.getLineageLeaf(0, 0.8));
    lineageArray.addLineage(provider.getLineageNot(lineageArray.getLineage(0)));
  	double derivative = confidenceCalculator.derivative(lineageArray.getLineage(1), 0);
  	assertTrue(Doubles.equals(derivative, -1.0));
  }
  
  @Test
  public void testDerivativeAnd() throws Exception {
  	lineageArray.addLineage(provider.getLineageLeaf(0, 0.8));
    lineageArray.addLineage(provider.getLineageLeaf(1, 0.4));
    lineageArray.addLineage(provider.getLineageAnd(lineageArray.getLineage(0), lineageArray.getLineage(1)));
  	double derivative = confidenceCalculator.derivative(lineageArray.getLineage(2), 0);
  	assertTrue(Doubles.equals(derivative, 0.4));
  }
  
  @Test
  public void testDerivativeOr() throws Exception {
  	lineageArray.addLineage(provider.getLineageLeaf(0, 0.8));
    lineageArray.addLineage(provider.getLineageLeaf(1, 0.4));
    lineageArray.addLineage(provider.getLineageOr(lineageArray.getLineage(0), lineageArray.getLineage(1)));
  	double derivative = confidenceCalculator.derivative(lineageArray.getLineage(2), 0);
  	assertTrue(Doubles.equals(derivative, 0.6));
  }
  
  @Test
  public void testDerivativeShannon() throws Exception {
  	lineageArray.addLineage(provider.getLineageLeaf(0, 0.8));
    lineageArray.addLineage(provider.getLineageLeaf(1, 0.4));
    lineageArray.addLineage(provider.getLineageLeaf(2, 0.3));
    lineageArray.addLineage(provider.getLineageOr(lineageArray.getLineage(0), lineageArray.getLineage(1)));
    lineageArray.addLineage(provider.getLineageOr(lineageArray.getLineage(1), lineageArray.getLineage(2)));
    lineageArray.addLineage(provider.getLineageAnd(lineageArray.getLineage(3), lineageArray.getLineage(4)));
  	double derivative = confidenceCalculator.derivative(lineageArray.getLineage(5), 1);
  	assertTrue(Doubles.equals(derivative, 0.76));
  }
}
