package tests.reasoner.lineage.shannonmaterializer;

import static org.junit.Assert.assertTrue;

import org.junit.Test;

import reasoner.io.DatabaseManager;
import reasoner.lineage.Lineage;
import reasoner.lineage.LineageArray;
import reasoner.lineage.LineageLeaf;
import reasoner.lineage.providers.LineageProvider;
import reasoner.lineage.shannonmaterializer.ShannonMaterializer;

public class ShannonMaterializerTest {
	private LineageArray lineageArray;
	private LineageProvider provider;  // careful, tests might fail, if provider invokes shannon materializer!
	private ShannonMaterializer materializer;
	
	public ShannonMaterializerTest() throws Exception {
		DatabaseManager.getInstance().loadDummyDataset();
		lineageArray = LineageArray.getInstance();
		provider = LineageProvider.getInstance();
		materializer = new ShannonMaterializer();
		materializer.setLineageProvider(LineageProvider.getInstance());
	}
	
	@Test
	public void testAnd() throws Exception {
		lineageArray.addLineage(provider.getLineageLeaf(0, 0.1));
		lineageArray.addLineage(provider.getLineageLeaf(1, 0.2));
		lineageArray.addLineage(provider.getLineageAnd(lineageArray.getLineage(0), lineageArray.getLineage(1)));
		
		Lineage result = materializer.materializeVariable(0, lineageArray.getLineage(2));
		
		Lineage expected = provider.getLineageDisjointOr(
				lineageArray.getLineage(1),  // var = true
				provider.getLineageConstant(false),  // var = false
				(LineageLeaf)lineageArray.getLineage(0));  // variable
		
		compareLineages(expected, result);
		
		provider.handBack(result, true);
		provider.handBack(expected, true);
		lineageArray.clearAll();
		assertTrue(provider.allObjectsReturned());
	}
	
	@Test
	public void testOr() throws Exception {
		lineageArray.addLineage(provider.getLineageLeaf(0, 0.1));
		lineageArray.addLineage(provider.getLineageLeaf(1, 0.2));
		lineageArray.addLineage(provider.getLineageOr(lineageArray.getLineage(0), lineageArray.getLineage(1)));
		
		Lineage result = materializer.materializeVariable(0, lineageArray.getLineage(2));
		
		Lineage expected = provider.getLineageDisjointOr(
				provider.getLineageConstant(true),  // var = true
				lineageArray.getLineage(1),  // var = false
				(LineageLeaf)lineageArray.getLineage(0));  // variable
		
		compareLineages(expected, result);
		
		provider.handBack(result, true);
		provider.handBack(expected, true);
		lineageArray.clearAll();
		assertTrue(provider.allObjectsReturned());
	}
	
	@Test
	public void testDisjointOrSame() throws Exception {
		lineageArray.addLineage(provider.getLineageLeaf(0, 0.1));
		lineageArray.addLineage(provider.getLineageLeaf(1, 0.2));
		lineageArray.addLineage(provider.getLineageLeaf(2, 0.3));
		lineageArray.addLineage(provider.getLineageDisjointOr(lineageArray.getLineage(0), lineageArray.getLineage(1), (LineageLeaf)lineageArray.getLineage(2)));
		
		Lineage result = materializer.materializeVariable(2, lineageArray.getLineage(3));
		
		Lineage expected = provider.getLineageDisjointOr(
				lineageArray.getLineage(0),  // var = true
				lineageArray.getLineage(1),  // var = false
				(LineageLeaf)lineageArray.getLineage(2));  // variable
		
		compareLineages(expected, result);
		
		provider.handBack(result, true);
		provider.handBack(expected, true);
		lineageArray.clearAll();
		assertTrue(provider.allObjectsReturned());
	}
	
	@Test
	public void testDisjointOrOther() throws Exception {
		lineageArray.addLineage(provider.getLineageLeaf(0, 0.1));
		lineageArray.addLineage(provider.getLineageLeaf(1, 0.2));
		lineageArray.addLineage(provider.getLineageLeaf(2, 0.3));
		lineageArray.addLineage(provider.getLineageDisjointOr(lineageArray.getLineage(0), lineageArray.getLineage(1), (LineageLeaf)lineageArray.getLineage(2)));
		
		Lineage result = materializer.materializeVariable(0, lineageArray.getLineage(3));
		
		// DISJOINT_OR[DISJOINT_OR[True 0.2 ) AND[NOT[0.3 ) 0.2 ) )
		Lineage expected = provider.getLineageDisjointOr(
				provider.getLineageDisjointOr(provider.getLineageConstant(true), lineageArray.getLineage(1), (LineageLeaf)lineageArray.getLineage(2)),  // var = true
				provider.getLineageAnd(provider.getLineageNot(lineageArray.getLineage(2)), lineageArray.getLineage(1)),  // var = false
				(LineageLeaf)lineageArray.getLineage(0));  // variable
		
		compareLineages(expected, result);
		
		provider.handBack(result, true);
		provider.handBack(expected, true);
		lineageArray.clearAll();
		assertTrue(provider.allObjectsReturned());
	}
	
	@Test
	public void testDisjointOtherReversed() throws Exception {
		lineageArray.addLineage(provider.getLineageLeaf(0, 0.1));
		lineageArray.addLineage(provider.getLineageLeaf(1, 0.2));
		lineageArray.addLineage(provider.getLineageLeaf(2, 0.3));
		lineageArray.addLineage(provider.getLineageDisjointOr(lineageArray.getLineage(0), lineageArray.getLineage(1), (LineageLeaf)lineageArray.getLineage(2)));
		
		Lineage result = materializer.materializeVariable(1, lineageArray.getLineage(3));
		
		// DISJOINT_OR[DISJOINT_OR[0.1 True ) AND[0.3 0.1 ) )
		Lineage expected = provider.getLineageDisjointOr(
				provider.getLineageDisjointOr(lineageArray.getLineage(0), provider.getLineageConstant(true), (LineageLeaf)lineageArray.getLineage(2)),  // var = true
				provider.getLineageAnd(lineageArray.getLineage(2), lineageArray.getLineage(0)),  // var = false
				(LineageLeaf)lineageArray.getLineage(1));  // variable
		
		compareLineages(expected, result);
		
		provider.handBack(result, true);
		provider.handBack(expected, true);
		lineageArray.clearAll();
		assertTrue(provider.allObjectsReturned());
	}
	
	@Test
	public void testLeafDifferent() throws Exception {
		lineageArray.addLineage(provider.getLineageLeaf(0, 0.1));
		lineageArray.addLineage(provider.getLineageLeaf(0, 0.2));
		
		Lineage result = materializer.materializeVariable(1, lineageArray.getLineage(0));
		
		Lineage expected = lineageArray.getLineage(0);

		compareLineages(expected, result);
		
		provider.handBack(result, true);
		provider.handBack(expected, true);
		lineageArray.clearAll();
		assertTrue(provider.allObjectsReturned());
	}
	
	@Test
	public void testLeafSame() throws Exception {
		lineageArray.addLineage(provider.getLineageLeaf(0, 0.1));
		
		Lineage result = materializer.materializeVariable(0, lineageArray.getLineage(0));
		
		Lineage expected = provider.getLineageDisjointOr(
				provider.getLineageConstant(true),  // var = true
				provider.getLineageConstant(false),  // var = false
				(LineageLeaf)lineageArray.getLineage(0));  // variable

		compareLineages(expected, result);
		
		provider.handBack(result, true);
		provider.handBack(expected, true);
		lineageArray.clearAll();
		assertTrue(provider.allObjectsReturned());
	}
	
	@Test
	public void testNot() throws Exception {
		lineageArray.addLineage(provider.getLineageLeaf(0, 0.1));
		lineageArray.addLineage(provider.getLineageNot(lineageArray.getLineage(0)));
		
		Lineage result = materializer.materializeVariable(0, lineageArray.getLineage(1));
		
		Lineage expected = provider.getLineageDisjointOr(
				provider.getLineageConstant(false),  // var = true
				provider.getLineageConstant(true),  // var = false
				(LineageLeaf)lineageArray.getLineage(0));  // variable
		
		compareLineages(expected, result);
		
		provider.handBack(result, true);
		provider.handBack(expected, true);
		lineageArray.clearAll();
		assertTrue(provider.allObjectsReturned());
	}
	
	@Test
	public void testFalse() throws Exception {
		lineageArray.addLineage(provider.getLineageConstant(false));
		lineageArray.addLineage(provider.getLineageLeaf(0, 0.2));
		
		Lineage result = materializer.materializeVariable(1, lineageArray.getLineage(0));
		
		Lineage expected = provider.getLineageConstant(false);

		compareLineages(expected, result);
		
		provider.handBack(result, true);
		provider.handBack(expected, true);
		lineageArray.clearAll();
		assertTrue(provider.allObjectsReturned());
	}
	
	@Test
	public void testTrue() throws Exception {
		lineageArray.addLineage(provider.getLineageConstant(true));
		lineageArray.addLineage(provider.getLineageLeaf(0, 0.2));
		
		Lineage result = materializer.materializeVariable(1, lineageArray.getLineage(0));
		
		Lineage expected = provider.getLineageConstant(true);

		compareLineages(expected, result);
		
		provider.handBack(result, true);
		provider.handBack(expected, true);
		lineageArray.clearAll();
		assertTrue(provider.allObjectsReturned());
	}
	
	@Test
	public void testNested() throws Exception {
		lineageArray.addLineage(provider.getLineageLeaf(0, 0.1));
		lineageArray.addLineage(provider.getLineageLeaf(1, 0.2));
		lineageArray.addLineage(provider.getLineageLeaf(2, 0.3));
		lineageArray.addLineage(provider.getLineageAnd(lineageArray.getLineage(0), lineageArray.getLineage(1)));
		lineageArray.addLineage(provider.getLineageOr(lineageArray.getLineage(2), lineageArray.getLineage(3)));
		lineageArray.addLineage(provider.getLineageNot(lineageArray.getLineage(4)));
		
		Lineage result = materializer.materializeVariable(0, lineageArray.getLineage(5));
		
		// DISJOINT_OR[NOT[OR[0.2 0.3 ) ) NOT[0.3 ) )
		Lineage expected = provider.getLineageDisjointOr(
				provider.getLineageNot(provider.getLineageOr(lineageArray.getLineage(1), lineageArray.getLineage(2))),  // var = true
				provider.getLineageNot(lineageArray.getLineage(2)),  // var = false
				(LineageLeaf)lineageArray.getLineage(0));  // variable
		
		compareLineages(expected, result);
		
		provider.handBack(result, true);
		provider.handBack(expected, true);
		lineageArray.clearAll();
		assertTrue(provider.allObjectsReturned());
	}
	/**
	 * compares two lineages for structural identity
	 */
	private void compareLineages(Lineage expected, Lineage given) {
		if (!expected.getClass().equals(given.getClass())) {
			assertTrue(false);
		}
		if (expected.subLineage == null || given.subLineage == null) {
			assertTrue(expected.subLineage == given.subLineage);
			return;
		}
		if (expected.subLineage.length != given.subLineage.length) {
			assertTrue(false);
		}
		for (int i = 0; i < expected.subLineage.length; i++) {
			compareLineages(expected.subLineage[i], given.subLineage[i]);
		}
	}
}
