package tests.reasoner.query;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import reasoner.grounding.Tuple;
import reasoner.grounding.Grounder;
import reasoner.grounding.Interval;
import reasoner.grounding.IntervalProvider;
import reasoner.io.DatabaseManager;
import reasoner.io.QueryParser;
import reasoner.io.dataset.Dataset;
import reasoner.io.dataset.Relation;
import reasoner.lineage.LineageArray;
import reasoner.query.QueryExecuter;
import reasoner.query.QueryExecuter.Results;
import reasoner.query.compiled.Query;
import util.Doubles;

public class QueryExecuterTest {
	
	@Before
	public void setUp() throws Exception {
    DatabaseManager manager = DatabaseManager.getInstance();
    try {
    	manager.deleteDataset("testDataset");
    } catch (Exception e) {
    	// the dataset did not exist
    }
    manager.clearDBall();
		
		//	 create dataset
    int datasetId = manager.createDataset("testDataset");
    Relation extensionalRelation = manager.newRelation(datasetId, "occ_test", 2, true, true, true);
    extensionalRelation.createTable();
    
    // create relation and insert two facts
    String[] arguments = {"TestArg1", "TestArg2"};
    extensionalRelation.addFactToBatch(arguments, 0.1, IntervalProvider.getInstance().get(-10, 0, 0));
    arguments[0] = "TestArg1";
    arguments[1] = "TestArg3";
    extensionalRelation.addFactToBatch(arguments, 0.2, IntervalProvider.getInstance().get(-10, 0, 1));
    extensionalRelation.executeBatch();
    
    // finalize dataset
    manager.finalizeDataset(datasetId, 2);
	}
	
	@After
	public void cleanUp() throws Exception {
		DatabaseManager.getInstance().deleteDataset("testDataset");
	}

  @Test
  public void runQueryMultipleTimes() throws Exception {
    String pathToCode = System.getProperty("user.dir");
    
    DatabaseManager manager = DatabaseManager.getInstance();
    
    // load dataset
    Dataset dataset = manager.getDataset("testDataset");
    
    // parse query
    QueryParser queryParser = new QueryParser(dataset);
    Query query = queryParser.parseQueryPlan(pathToCode + "/input/testing/query/simpleRule.txt");
    query.output.add(dataset.relations.get("intensional_test"));
    
    QueryExecuter queryExecuter = new QueryExecuter();
    queryExecuter.grounder = new Grounder();  // we initialize manually
    queryExecuter.query = query;
    for (int i = 0; i < 10; i++) {
      Results results = queryExecuter.groundQuery();
      queryExecuter.computeConfidences(results);
      assertTrue(LineageArray.getInstance().getNextArrayId() == 3);
      assertTrue(results.resultTuples.size() == 1);
      Tuple tuple = results.resultTuples.get(0);
      assertTrue(tuple.intervals.size() == 1);
      Interval interval = tuple.intervals.get(0);
      assertTrue(interval.begin == -10);
      assertTrue(interval.end == 0);
      assertFalse(interval.isInArray);
      assertTrue(Doubles.equals(interval.confidence, 0.28));
      queryExecuter.clearQuery(results);
    }
    queryExecuter.clean();
  }

  @Test
  public void runQueryWithConstants() throws Exception {
  	String pathToCode = System.getProperty("user.dir");
    
    DatabaseManager manager = DatabaseManager.getInstance();
    
    // load dataset
    Dataset dataset = manager.getDataset("testDataset");
    
    // parse query
    QueryParser queryParser = new QueryParser(dataset);
    Query query = queryParser.parseQueryPlan(pathToCode + "/input/testing/query/withConstant.txt");
    query.output.add(dataset.relations.get("intensional_test"));
    query.constantsToSet.add("TestArg1");
    query.constantsToSet.add("TestArg2");
    
    QueryExecuter queryExecuter = new QueryExecuter();
    queryExecuter.grounder = new Grounder();  // we initialize manually
    queryExecuter.query = query;
    
    assertTrue(queryExecuter.hasNextQuery());
    Results results = queryExecuter.groundQuery();
    assertTrue(results.resultTuples.isEmpty());
    queryExecuter.clearQuery(results);
    
    assertTrue(queryExecuter.hasNextQuery());
    results = queryExecuter.groundQuery();
    assertTrue(results.resultTuples.size() == 1);
    queryExecuter.clearQuery(results);
    
    assertTrue(!queryExecuter.hasNextQuery());
    
    queryExecuter.clean();
  }
}
