package tests.reasoner.weightlearning;

import static org.junit.Assert.*;

import java.util.Set;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import reasoner.grounding.Interval;
import reasoner.grounding.IntervalProvider;
import reasoner.io.DatabaseManager;
import reasoner.io.QueryParser;
import reasoner.io.Time;
import reasoner.io.dataset.Dataset;
import reasoner.io.dataset.Relation;
import reasoner.io.weightlearning.ParsedLabels;
import reasoner.io.weightlearning.WeightLearningParser;
import reasoner.lineage.LineageArray;
import reasoner.lineage.providers.LineageProvider;
import reasoner.query.compiled.Query;
import reasoner.weightlearning.WeightLearningConfig;
import reasoner.weightlearning.WeightLearningExecuter;
import reasoner.weightlearning.WeightLearningGrounder;
import reasoner.weightlearning.problem.ProblemInstance;

public class WeightLearningTest {
	
	private Dataset dataset;
	private LineageArray lineageArray;
	private String pathToCode;
	private Time time;
	
	public WeightLearningTest() {
		lineageArray = LineageArray.getInstance();
		time = Time.getInstance();
		pathToCode = System.getProperty("user.dir") + "/input/testing/learning/";
	}
	
	@Before
  public void setUp() throws Exception {
		// create dataset
	  DatabaseManager manager = DatabaseManager.getInstance();
	  try {
	  	manager.deleteDataset("testDataset");
	  } catch (Exception e) {
	  	// the dataset did not exist.
	  }
	  manager.clearDBall();
	  int datasetId = manager.createDataset("testDataset");
	  manager.newRelation(datasetId, "occ_test", 2, true, true, true).createTable();
	  manager.finalizeDataset(datasetId, 1);
	  
	  // load dataset
    dataset = manager.getDataset("testDataset");
    
    // insert tuples
    Time.getInstance().setParsingOptions("-", "X", false, "time_begin", "time_end");
    addTuple("arg1", "arg2", "2000-01-01", "2000-02-01");
    addTuple("arg1", "arg2", "2000-02-01", "2000-03-01");
    addTuple("arg1", "arg2", "2000-03-01", "2000-04-01");
	}
	
	@After
	public void cleanUp() throws Exception {
		DatabaseManager.getInstance().deleteDataset(dataset.name);
    lineageArray.clearAll();
	}
	
	@Test
  public void learn() throws Exception {
		new WeightLearningConfig(null);
		QueryParser queryParser = new QueryParser(dataset);
		WeightLearningGrounder grounder = new WeightLearningGrounder();
		Query query = queryParser.parseQueryPlan(pathToCode + "/deductionRules.txt");
		
		WeightLearningParser labelParser = new WeightLearningParser();
		ParsedLabels labels = labelParser.parseLabels(pathToCode + "/labels.txt", dataset);
		Set<Relation> relationsToLearn = labelParser.parseRelationsToLearn(pathToCode + "/relationsToLearn.txt", dataset);
		grounder.createLineage(query);
		ProblemInstance problem = grounder.getProblemInstance("TestProblem", relationsToLearn, labels);
		assertTrue(problem.tuples.length == 3);
		assertTrue(problem.tuples[0].labeledFormulas.size() == 1);
		assertTrue(problem.tuples[1].labeledFormulas.size() == 1);
		assertTrue(problem.tuples[2].labeledFormulas.size() == 0);
		assertTrue(problem.formulas.length == 2);

		WeightLearningExecuter executer = new WeightLearningExecuter(null, null);
		executer.runCurrentConfig(problem, 0);
		assertTrue(Math.abs(problem.tuples[0].leaf.conf - 1.0) < 0.1);
		assertTrue(Math.abs(problem.tuples[1].leaf.conf - 0.5) < 0.1);
		assertTrue(problem.tuples[2].leaf.conf == 0.5);  // unlabeled
 
		problem.clear();
		problem.delete();
		DatabaseManager.getInstance().clearDBintensional();
	}
	
	private void addTuple(String arg1, String arg2, String timeBegin, String timeEnd) throws Exception {
		Relation relation = dataset.relations.get("occ_test");
    int lineageId = lineageArray.getNextArrayId();
		lineageArray.addLineage(LineageProvider.getInstance().getLineageLeaf(lineageId, 0.5));
		int begin = time.string2int(timeBegin)[0];
		int end = time.string2int(timeEnd)[0];
		Interval interval = IntervalProvider.getInstance().get(begin, end, lineageId);
		String[] arguments = {arg1, arg2};
    relation.addFactToBatch(arguments, 0.5, interval);
    relation.executeBatch();
	}
}
