package util.provider;

import java.util.Collection;


/**
 * This class keeps all objects of type T in memory, such that
 * for a warm cache no objects will be created. 
 * 
 * If a new instance is necessary, it is retrieved from the array.
 * Old instances are cleared and returned to the array.
 * 
 * @author mdylla@mpi-inf.mpg.de (Maximilian Dylla)
 *
 * @param T the class to be provided
 * 
 */
public abstract class AbstractProvider<T extends Providable> {
  
  private final int initialSize;
  private final String name;
  
  protected Providable[] availableObjects;
  private int arrayPointer;
  
  protected AbstractProvider(String name, int initialSize) throws Exception {
    this.initialSize = initialSize;
    availableObjects = new Providable[initialSize];
    for (int arrayId = 0; arrayId < initialSize; arrayId++) {
      availableObjects[arrayId] = getNew();
    }
    arrayPointer = initialSize - 1;
    this.name = name;
  }

  @SuppressWarnings("unchecked")
  protected T getCleanObject() throws Exception {
    if (arrayPointer == -1) {
      System.out.println("creating more objects of " + name + " " + availableObjects.length);
      enlargeArray();
    }
    T object = (T)availableObjects[arrayPointer--];
    object.isInArray = false;
    return object;
  }

  public void handBack(T toReturn) throws Exception {
    if (!toReturn.isInArray) {
      toReturn.clear();
      availableObjects[++arrayPointer] = toReturn;
      toReturn.isInArray = true;
    }
  }
  
  public void handBack(T[] toReturn) throws Exception {
    for (T t : toReturn) {
      handBack(t);
    }
  }
  
  public void handBack(Collection<T> toReturn) throws Exception {
    for (T t : toReturn) {
      handBack(t);
    }
  }
  
  public boolean allObjectsReturned() {
    if (arrayPointer != availableObjects.length - 1) {
      System.err.println((availableObjects.length - arrayPointer + 1)+" objects of "+name+" were not returned!");
      return false;
    }
    return true;
  }
  
  protected abstract T getNew() throws Exception;
  
  private void enlargeArray() throws Exception {
    int originalSize = availableObjects.length;
    availableObjects = new Providable[originalSize * 2];
    for (int arrayId = 0; arrayId < originalSize; arrayId++) {
      availableObjects[arrayId] = getNew();
    }
    // the last @originalSize spots remain empty.
    // So at all times all objects fit into the array.
    arrayPointer = originalSize - 1;
  }
  
  // Required for testing
  public int getInitialSize() {
    return initialSize;
  }
}
