package util.time;

import java.util.HashMap;

/**
 * A singleton class for keeping timers.
 * A timer is identified by a string.
 * A timer can be running, stopped and continue to run.
 */
public class Timer {

	private HashMap<String, Long> runningTimers;  // name -> startTime
	private HashMap<String, Long> elapsedTimes;  // name -> elapsed time
	private static Timer instance;
	
	private Timer() {
		runningTimers = new HashMap<String, Long>();
		elapsedTimes = new HashMap<String, Long>();
	}
	
	public static Timer getInstance() {
		if (instance == null) {
			instance = new Timer();
		}
		return instance;
	}
	
	public void initTimer(String name) throws Exception {
		if (runningTimers.containsKey(name)) {
			throw new Exception("Timer existed already: " + name);
		}
		if (elapsedTimes.containsKey(name) && elapsedTimes.get(name) != 0L) {
			// the timer can be initiated more than once
			throw new Exception("Timer existed already: " + name);
		}
		elapsedTimes.put(name, 0L);
	}
	
	public void startTimer(String name) throws Exception {
		if (runningTimers.containsKey(name)) {
			throw new Exception("Timer is already running: " + name);
		}
		runningTimers.put(name, System.currentTimeMillis());
	}
	
	public long getElapsedTime(String name) throws Exception {
		boolean unknownTimer = true;
		long elapsedTime = 0;
		if (runningTimers.containsKey(name)) {
			elapsedTime = System.currentTimeMillis() - runningTimers.get(name);
			unknownTimer = false;
		}
		if (elapsedTimes.containsKey(name)) {
			elapsedTime += elapsedTimes.get(name);
			unknownTimer = false;
		}
		if (unknownTimer) {
			throw new Exception("Unknown timer: " + name);
		}
		return elapsedTime;
	}
	
	public void stopTimer(String name) throws Exception {
		if (!runningTimers.containsKey(name)) {
			throw new Exception("Timer was not running: " + name);
		}
		long time = System.currentTimeMillis() - runningTimers.get(name);
		runningTimers.remove(name);
		if (!elapsedTimes.containsKey(name)) {
			elapsedTimes.put(name, 0L);
		}
		time += elapsedTimes.get(name);
		elapsedTimes.put(name, time);
	}
	
	public long elapsedAndRemoveTimer(String name) throws Exception {
		long time = getElapsedTime(name);
		remove(name);
		return time;
	}
	
	public void remove(String name) throws Exception {
		runningTimers.remove(name);
		elapsedTimes.remove(name);
	}
	
	public void reset(String name) throws Exception {
		remove(name);
		initTimer(name);
	}
	
	public void clear() {
		runningTimers.clear();
		elapsedTimes.clear();
	}
}
